/* asn1_helper.h
*
* Routine to augment previously existing ASN1 routines
*
* $Id: asn1_helper.c,v 1.0 2004/012/12  Exp $
*
*
* Herbert Falk <herb@sisconet.com>
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>

#include <stdio.h>
#include <string.h>


#include "asn1.h"
#include "packet-frame.h"
#include "prefs.h"

#include <epan/strutil.h>

#define ASN1H_CONSTRUCT	0x20000000
#define ASN1H_UNI		0x00000000
#define ASN1H_APP		0x40000000
#define ASN1H_CTX		0x80000000
#define ASN1H_PRIV		0xc0000000	
#define ASN1H_END_OF_TABLE			0xffffffff

/* the following structure is used to drive a table driven ASN1 decode */
struct asn1h_tbl_entry{
	   guint tag;						/*the tags can have a maximum value of 8191 	*/
								/*which should be enough for most protocols		*/
   	   int    *ett;						/*this is the pointer that will be used for building a new subtree */
	   int    *hf;						/*this is the pointer that will be used for display filtering */
	   guchar *tag_string; 					/*this is the pointer that will be used for display purposes */


/****************************************************
The following is the way the action mask is allocated

ayyy   zzzz    qxxx    xxxx
 \		\		\	  	\
 \		\		+-------+---- Action Options
 \      \
 \


a = 0 no mapping to string
    1 map to string

q = 1 to map to multipe octet bistrings

yy = Protocol selection

zzzz = table entry index for integer to string mapping
*****************************************************/


/* then following are the action Option definitions */

#define ASN1H_MAP_INT_TO_STRING 		0x80000000
#define ASN1H_MAP_TBL_ID_MASK   		0x70000000
#define ASN1H_MAP_TBL_ENTRY_MASK		0x0ff00000
#define ASN1H_TBL_SEQ					0x000c0000			/*acts as a sequence	*/
#define ASN1H_TBL_SET					0x00000000		    /*acts as a set	*/
#define ASN1H_BITSTRING_PAIR_DISPLAY	0x00008000
#define ASN1H_DISPLAY_ACTION_MASK		0x00003ffff


#define ASN1H_OCTETSTRING_DISPLAY  	1
#define ASN1H_INTEGER_DISPLAY		2
#define ASN1H_OBJID_DISPLAY			3
#define ASN1H_BITSTRING_DISPLAY	    4
#define ASN1H_VSTRING_DISPLAY		5
#define ASN1H_UNSIGNED_DISPLAY		6
#define ASN1H_BOOLEAN_DISPLAY		7
#define ASN1H_DONOT_DISPLAY			8
#define ASN1H_SKIP_TAG_AND_LENGTH	9
#define ASN1H_INDENT				10
#define ASN1H_DO_NOT_PROCESS		11

/* the following are special displays for MMS		*/
#define ASN1H_MMS_BTIME_DISPLAY			12
#define ASN1H_MMS_UTCTIME_DISPLAY		13
#define ASN1H_MMS_FLOAT_DISPLAY			14
#define ASN1H_INTEGER_ENUM_DISPLAY		15
#define ASN1H_NULL_DISPLAY				16
#define ASN1H_BITSTRING_ENUM_DISPLAY	17
#define ASN1H_MMS_TYPE_LENGTH			18
#define ASN1H_GENERALTIME_DISPLAY		19
#define ASN1H_MMS_UNICODE_DISPLAY		20
#define ASN1H_NONASN1_DECODE			21


	   guint conversion_type;   		/*this will be the type of data that needs to be converted  */
	   struct asn1h_tbl_entry *next;	/*pointer to the next table							*/ 
		};

typedef struct asn1h_tbl_entry ASN1_TBL_ENTRY;

typedef struct enum_lookup{
   guint value;					/*value or bit number to translate		*/
   char *stringvalue;					/*string to use for enumerated value	*/
}ENUM_LOOKUP;

typedef struct enum_xlat {				/*structure that can be passed so that values can be converted */
	guint num_entries;					/*to strings												   */
	ENUM_LOOKUP *lookup;
}ENUM_XLAT;

typedef struct spec_decode {
   guchar *display_offset;
   guint32 length;
   guchar  *stuff_to_decode;
   guint32   init_offset;
   proto_item *subtree;
   tvbuff_t *tvb;
}SPEC_DECODE;

extern int asn1h_tag_get (ASN1_SCK *asn1, guint *tag);
extern int asn1h_length_decode (ASN1_SCK *asn1, gboolean *def, guint *len);
extern int asn1h_do_parse (ASN1_SCK *asn1, ASN1_TBL_ENTRY *tbl_array, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, tvbuff_t *tvb, guint type_of_tbl, gboolean nodes_as_trees);

typedef void specDecodeFunc ( guint32 length, guchar *dptr, guchar *disp_string, guint32 offset, proto_item *subtree, tvbuff_t *tvb);




