/* packet-clisoacse.c
*
* Routine to dissect connectionless ISO 8823-1 OSI Presentation Protocol packets
*
* $Id: packet-isopres.c,v 1.0 2004/012/12  Exp $
*
* Herbert Falk <herb@sisconet.com>
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <glib.h>
#include <epan/packet.h>

#include <stdio.h>
#include <string.h>

#include "asn1.h"
#include "asn1_helper.h"
#include "packet-frame.h"
#include "prefs.h"

#include <epan/strutil.h>


/* find the dissector for data */
static dissector_handle_t adata_handle;
static dissector_handle_t acse_handle;


static int proto_isoacse          = -1;
static int proto_clisoacse          = -1;
static gint ett_isoacse           = -1;
proto_item *data_item;
static dissector_handle_t gsse_handle;
static dissector_handle_t mms_handle;


#define ISO_ACSE_CO				0			/*Connection Oriented Presentation	*/
#define ISO_ACSE_CL				1			/*Connectionless Presentation	*/
static int hf_isoacse_type        = -1;


#define PROTO_STRING_ACSE	"ISO 8850 ACSE"
#define PROTO_STRING_CLACSE  "ITU X.237 CL ACSE"
#define CL_ACSE_SEQ		0x60		/*sequence code for Connectionless Presentation */

#define ASN1H_SEQ_TAG	(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ)
#define ASN1H_INTEGER_ID (ASN1H_UNI | ASN1_INT)		/*ASN1_INT is defined in ASN1.h*/
#define ASN1H_OBJECT_ID (ASN1H_UNI | ASN1_OJI)		/*ASN1_OJI is defined in ASN1.h*/

#define NOET NULL
#define NOHF NULL
static ASN1_TBL_ENTRY isoclacse_external[]={
		{ (ASN1H_CONSTRUCT | 8), NOET, NOHF,  "External",NULL, isoclacse_external},  
		{ ASN1H_OBJECT_ID, NOET, NOHF, "Trasfer Syntax: ", ASN1H_OBJID_DISPLAY, NULL},
		{ ASN1H_INTEGER_ID,  NOET, NOHF,"ID: ", ASN1H_INTEGER_DISPLAY, NULL},
		{ (ASN1H_CONSTRUCT | ASN1H_CTX | 0), NOET, NOHF, NULL ,ASN1H_DO_NOT_PROCESS, NULL}, 
		{ASN1H_END_OF_TABLE	, NOET, NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY isoclacse_main[]={
		{ (ASN1H_CTX | 0 ),  NOET, NOHF,"Protocol Version:", ASN1H_BITSTRING_DISPLAY ,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 1 ), NOET, NOHF, "APP CTXT_NAME:"	,ASN1H_OBJID_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |2 ), NOET, NOHF, "Called AP-Title:",ASN1H_OBJID_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |3 ), NOET, NOHF, "Called AE-Qualifier:",ASN1H_INTEGER_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |4 ), NOET, NOHF, "Called AP-InvokeID:",ASN1H_INTEGER_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |5 ) , NOET, NOHF, "Called AE-InvokeID:",ASN1H_INTEGER_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |6 ), NOET, NOHF, "Calling AP-Title:",ASN1H_OBJID_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |7 ), NOET, NOHF, "Calling AE-Qualifier:",ASN1H_INTEGER_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |8 ), NOET, NOHF, "Calling AP-InvokeID:",ASN1H_INTEGER_DISPLAY,isoclacse_main},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |9 ), NOET, NOHF, "Calling AE-InvokeID:",ASN1H_INTEGER_DISPLAY,isoclacse_main},
		{ ASN1H_OBJECT_ID, NOET, NOHF, "", ASN1H_OBJID_DISPLAY, NULL},
		{ ASN1H_INTEGER_ID, NOET, NOHF, "", ASN1H_INTEGER_DISPLAY, NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 30), NOET, NOHF, "User Data",NULL,isoclacse_external},
		{ASN1H_END_OF_TABLE	, NOET, NOHF,NULL,NULL,NULL}
		};		 

static void
dissect_cl_acse(ASN1_SCK *asn1, guchar *display_string, guint32 *total_len, guint32 init_offset, proto_item *subtree, tvbuff_t *tvb)
{
guint tag;
guint total_length;
gboolean not_end = FALSE;
guchar *mysel;
guint i;
guchar display_text[32];
gint ret;
gboolean def;
guint init_offset1;

		 /*call to parse AUNIT-DATA */
        asn1h_do_parse (asn1, isoclacse_main, "", total_len, asn1->offset, subtree, tvb,ASN1H_TBL_SET, FALSE);
}

static void
dissect_isoacse(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset = 0;
	guint8 type;
	proto_item *ti;
    proto_tree  *isoacse_tree;
	guint tag;
	int ret;
	gboolean def;
	ASN1_SCK asn1;
	guint total_length;
	tvbuff_t *next_tvb;


	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "ACSE");
  	if (check_col(pinfo->cinfo, COL_INFO))
  		col_clear(pinfo->cinfo, COL_INFO);

	if (tree) 
	  {
	  /*open the ASN1 tree, offset is tvb->offset */
      asn1_open(&asn1, tvb, offset);

   	  ret = asn1_tag_decode (&asn1, &tag);

      ret = asn1_length_decode (&asn1, &def, &total_length);
	  if (tag ==CL_ACSE_SEQ)
	    {
		ti = proto_tree_add_item(tree, proto_clisoacse, tvb, 0, -1, FALSE);
  		data_item = proto_item_add_subtree(ti, ett_isoacse); 

		 dissect_cl_acse (&asn1, tree, &total_length, offset, data_item, tvb);

		asn1.offset = asn1.offset;
		if (tvb_reported_length_remaining(tvb, asn1.offset) > 0) 
			next_tvb = tvb_new_subset(tvb, asn1.offset, -1, -1);

		 if(gsse_handle)
		     call_dissector(gsse_handle, next_tvb, pinfo, tree);

		} 
	  else
		ti = proto_tree_add_item(tree, proto_isoacse, tvb, 0, -1, FALSE);
	  }
}

static void
dissect_clisoacse(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset = 0;
	guint8 type;
	proto_item *ti;
    proto_tree  *isoacse_tree;
	guint tag;
	int ret;
	gboolean def;
	ASN1_SCK asn1;
	guint total_length;
	tvbuff_t *next_tvb;


	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "ACCL");
  	if (check_col(pinfo->cinfo, COL_INFO))
  		col_clear(pinfo->cinfo, COL_INFO);

	if(!tree)
	{
		if(gsse_handle)
		     call_dissector(gsse_handle, tvb, pinfo, tree);

	}


	if (tree) 
	  {
	  /*open the ASN1 tree, offset is tvb->offset */
      asn1_open(&asn1, tvb, offset);

   	  ret = asn1_tag_decode (&asn1, &tag);

      ret = asn1_length_decode (&asn1, &def, &total_length);
	  if (tag ==CL_ACSE_SEQ)
	    {
		ti = proto_tree_add_item(tree, proto_clisoacse, tvb, 0, -1, FALSE);
  		data_item = proto_item_add_subtree(ti, ett_isoacse); 

		 dissect_cl_acse (&asn1, tree, &total_length, offset, data_item, tvb);

		asn1.offset = asn1.offset;
		if (tvb_reported_length_remaining(tvb, asn1.offset) > 0) 
			next_tvb = tvb_new_subset(tvb, asn1.offset, -1, -1);

		 if(gsse_handle)
		     call_dissector(gsse_handle, next_tvb, pinfo, tree);

		} 
	  }
}


void
proto_register_isoacse(void)
{
	static hf_register_info hf[] =
	{

		{
			&hf_isoacse_type,
			{
				"SPDU Type",
				"ses.type",
				FT_UINT8,
				BASE_DEC,
				NULL,
				0x0,
				"", HFILL
			}
		},
	};

	static gint *ett[] =
	{
		&ett_isoacse,
	};
	module_t *acse_module;


	proto_clisoacse = proto_register_protocol(PROTO_STRING_CLACSE, "ACCL", "adata");
	proto_register_field_array(proto_isoacse, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	register_dissector("adata", dissect_clisoacse, proto_clisoacse);
}

#if 0
static gboolean
dissect_isoacse_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree)
{
	/* must check that this really is a ses packet */
	int offset = 0;
	guint8 type;
	int len_len;
	guint16 len;

	/* first, check do we have at least 4 bytes (type+length) */
	if (!tvb_bytes_exist(tvb, 0, 4))
		return FALSE;	/* no */

	/* can we regognize session PDU ? Return FALSE if  not */
	/*   get SPDU type */
	type = tvb_get_guint8(tvb, offset);
	/* check SPDU type */

#if 0
	if (match_strval(type, ses_vals) == NULL)
	{
		return FALSE;  /* no, it isn't a session PDU */
	}
#endif

#if 0
	/*  OK,let's check SPDU length  */
	/*  get length of SPDU */

	len = get_item_len(tvb, offset, &len_len);

#endif

	/* do we have enough bytes ? */
	if (!tvb_bytes_exist(tvb, 0, len))
		return FALSE;	/* no */

	dissect_isopres(tvb, pinfo, parent_tree);
	return TRUE;
}

#endif


void
proto_reg_handoff_isoacse(void)
{
guchar myname[32];
#if 0
	/*   find data dissector  */
    strcpy (myname, "mms");
	mms_handle = find_dissector(myname);
#endif

	/* define sub dissector */
	strcpy(myname, "gsse");
	gsse_handle = find_dissector(myname);

}
