/* packet-IECGOOSE.c
 * Routines for IEC 61850 IECGOOSE dissection
 * 2003, Herbert Falk <herb@sisconet.com>
 *
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * Copied from WHATEVER_FILE_YOU_USED (where "WHATEVER_FILE_YOU_USED"
 * is a dissector file; if you just copied this from README.developer,
 * don't bother with the "Copied from" - you don't even need to put
 * in a "Copied from" if you copied an existing dissector, especially
 * if the bulk of the code in the new dissector is your code)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "packet-ber.h"

#include "asn1.h"
#include "asn1_helper.h"
#include "packet-frame.h"
#include "prefs.h"
#include "etypes.h"				/*for the Ethertype IDs, may no longer be needed*/

#include "packet-IECGOOSE.h"

#define NOETT NULL
#define NOHF  NULL

/* Initialize the protocol and registered fields */
static int proto_iecgoose = -1;
static int proto_IECGOOSE_GOOSE = -1;
static int proto_cltp = -1;

static int hf_IECGOOSE_GOOSE  = -1;
static gint ett_IECGOOSE = -1;
static gint ett_goosepdu = -1;

static int hf_iecgoose_appid_type = -1;

static int hf_iecgoose_length_type = -1;
static int hf_iecgoose_resv1_type = -1;
static int hf_iecgoose_resv2_type = -1;
static int hf_iecgoose_data_type = -1;
static int hf_iecgoose_control_block_reference = -1;
static gint ett_gocbRef = -1;
static gint ett_statenumber = -1;
static gint ett_data = -1;
static int hf_statenumber = -1;
static int hf_dataset = -1;
static int ett_dataset = -1;
static int hf_goid = -1;
static int ett_goid = -1;
static int hf_test = -1;
static int hf_ndscomm = -1;
static int hf_confrev = -1;

/*static int ett_test = -1; */

static int ett_datatbl=-1;
static int ett_array=-1;
static int ett_structure=-1;
static int ett_bitstring=-1;
static int ett_octettstring=-1;
static int ett_string=-1;
static int ett_gentime=-1;
static int ett_utctime=-1;
static int ett_bintime=-1;
static int ett_unicode=-1;
static int ett_bcd=-1;
static int ett_oid=-1;


#if 0
static int hf_iecgoose_hard_size = -1;
static int hf_iecgoose_proto_size = -1;
static int hf_iecgoose_opcode = -1;
static int hf_iecgoose_src_hw = -1;
static int hf_iecgoose_src_hw_mac = -1;
static int hf_iecgoose_src_proto = -1;
static int hf_iecgoose_src_proto_id = -1;
static int hf_iecgoose_dst_hw = -1;
static int hf_iecgoose_dst_hw_mac = -1;
static int hf_iecgoose_dst_proto = -1;
static int hf_iecgoose_dst_proto_id = -1;
#endif

static int displayed_header = FALSE;		/*use to display SISCO information one time for advertisement*/


void proto_reg_handoff_IECGOOSE(void);
static gboolean global_iecgoose_asn1_prefs_nodes_as_subtrees = FALSE;


/*#define VLAN_TPID	    0
#define VLAN_TCI		2
#define	IEC_ETHERTYPE	4 */
#define IEC_APPID		0
#define IEC_LENGTH      2
#define IEC_RESV1		4
#define IEC_RESV2		6
#define IEC_GOOSE_PDU	8



static ASN1_TBL_ENTRY data_tbl[]={
        {(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_array,NOHF,"ARRAY", 0,NULL},
        {(ASN1H_CTX | ASN1H_CONSTRUCT | 2),	&ett_structure,NOHF,"STRUCTURE", 0,NULL},
        {(ASN1H_CTX |  3),			NOETT,NOHF,"BOOLEAN: ",ASN1H_BOOLEAN_DISPLAY,NULL},
        {(ASN1H_CTX |  4),			&ett_bitstring,NOHF,"BITSTRING: ",ASN1H_BITSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  5),			NOETT,NOHF,"INTEGER: ", ASN1H_INTEGER_DISPLAY,NULL},
        {(ASN1H_CTX |  6),			NOETT,NOHF,"UNSIGNED: ", ASN1H_UNSIGNED_DISPLAY,NULL},
        {(ASN1H_CTX |  7),			NOETT,NOHF,"FLOAT: ", ASN1H_MMS_FLOAT_DISPLAY,NULL},
        {(ASN1H_CTX |  9),			&ett_octettstring,NOHF,"OSTRING: ", ASN1H_OCTETSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  10),			&ett_string,NOHF,"VSTRING: ", ASN1H_VSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  11),			&ett_gentime,NOHF,"GTIME ", ASN1H_MMS_BTIME_DISPLAY,NULL},
        {(ASN1H_CTX |  12),			&ett_bintime,NOHF,"BTIME ", ASN1H_MMS_BTIME_DISPLAY,NULL},
        {(ASN1H_CTX |  13),			&ett_bcd,NOHF,"BCD ", ASN1H_UNSIGNED_DISPLAY,NULL},
        {(ASN1H_CTX |  14),			&ett_octettstring,NOHF,"BOOLEAN Array ", ASN1H_OCTETSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  15),			&ett_oid,NOHF,"OBJID ", ASN1H_OBJID_DISPLAY,NULL},
	{(ASN1H_CTX |  16),			&ett_unicode,NOHF,"UNICODE",ASN1H_MMS_UNICODE_DISPLAY,NULL},	   
	{(ASN1H_CTX |  17), 			&ett_utctime,NOHF,"UTC",ASN1H_MMS_UTCTIME_DISPLAY,NULL},
	{(ASN1H_CTX |  18), 			&ett_unicode,NOHF,"UNICODE(CDV)",ASN1H_MMS_UNICODE_DISPLAY,NULL}, 
	{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY goose_start[]={
		{(ASN1H_CTX | 0),NOETT,&hf_iecgoose_control_block_reference,"",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOETT,NOHF,"Time Allowed to Live (msec): ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 2),NOETT,&hf_dataset,"",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 3),NOETT,&hf_goid,"",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 4),NOETT,NOHF,"Event Timestamp: ",ASN1H_MMS_UTCTIME_DISPLAY, NULL},
		{(ASN1H_CTX | 5),NOETT,&hf_statenumber,"",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 6),NOETT,NOHF,"Sequence Number: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 7),NOETT,&hf_test,"",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 8),NOETT,&hf_confrev,"",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 9),NOETT,&hf_ndscomm,"",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 10),NOETT,NOHF,"Number Dataset Entries: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 11),&ett_data,NOHF,"Data ",ASN1H_TBL_SET, data_tbl},
   		{ASN1H_END_OF_TABLE	,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY goose_message[]={
		{(ASN1H_APP | ASN1H_CONSTRUCT | 1),NOETT, NOHF,"IEC GOOSE ",ASN1H_TBL_SEQ, goose_start},
   		{ASN1H_END_OF_TABLE	,NULL,NULL,NULL}
};


static void
dissect_IECGOOSE( tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  guint16	  iecgoose_APPID;
  guint16	  iecgoose_length;
  guint16	  iecgoose_resv1;
  guint16	  iecgoose_resv2;
  proto_tree  *iecgoose_tree;
  proto_tree  *iecgoose_asn1tree;
  ASN1_SCK asn1;
  proto_item *ti;
  guint message_length=0;

/* Make entries in Protocol column and Info column on summary display */
	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "IECGOOSE");
    
	if (check_col(pinfo->cinfo, COL_INFO)) 
	{
	  if (displayed_header == TRUE)
		col_set_str(pinfo->cinfo, COL_INFO, "GOOSE Request");
	  else
	    {
		col_set_str(pinfo->cinfo, COL_INFO, "GOOSE Request Decoder, Courtesy of Herbert Falk of SISCO");
		displayed_header = TRUE;
		}
	}

/*get information from the TVB structure */
/*  vlan_tpid= tvb_get_ntohs(tvb, VLAN_TPID);
  vlan_tci= tvb_get_ntohs(tvb, VLAN_TCI);
  iecgoose_ethertype = tvb_get_ntohs(tvb, IEC_ETHERTYPE);  not needed */

  iecgoose_APPID =  tvb_get_ntohs(tvb, IEC_APPID);
  iecgoose_length = tvb_get_ntohs(tvb, IEC_LENGTH);
  iecgoose_resv1= tvb_get_ntohs(tvb, IEC_RESV1);
  iecgoose_resv2=  tvb_get_ntohs(tvb, IEC_RESV2);

	if (tree) {

/* NOTE: The offset and length values in the call to
   "proto_tree_add_item()" define what data bytes to highlight in the hex
   display window when the line in the protocol tree display
   corresponding to that item is selected.

   Supplying a length of -1 is the way to highlight all data from the
   offset to the end of the packet. */

/* create display subtree for the protocol */
		ti = proto_tree_add_item(tree, proto_iecgoose, tvb, 0, -1, FALSE);

/*		IECGOOSE_tree = proto_item_add_subtree(ti, ett_IECGOOSE); */

/* add an item to the subtree, see section 1.6 for more information */
/* this needs to be put back in $$$$ Herbert Falk */
/*	proto_tree_add_item(IECGOOSE_tree,
		    hf_IECGOOSE_GOOSE, tvb, offset, len, FALSE); */

    iecgoose_tree = proto_item_add_subtree(ti, ett_goosepdu);

   proto_tree_add_uint(iecgoose_tree ,hf_iecgoose_appid_type, tvb, IEC_APPID, 2,
			       iecgoose_APPID);

   proto_tree_add_uint(iecgoose_tree ,hf_iecgoose_length_type, tvb, IEC_LENGTH, 2,
			       iecgoose_length);

   proto_tree_add_uint(iecgoose_tree ,hf_iecgoose_resv1_type, tvb, IEC_RESV1, 2,
			       iecgoose_resv1);

   proto_tree_add_uint(iecgoose_tree ,hf_iecgoose_resv2_type, tvb, IEC_RESV2, 2,
			       iecgoose_resv2);

   ti = proto_tree_add_text(iecgoose_tree, tvb, IEC_GOOSE_PDU, iecgoose_length-IEC_GOOSE_PDU, "PDU");

 /* ready to start decoding the PDU */

   /*add the subtree  */
 /*  iecgoose_asn1tree = proto_item_add_subtree(ti, ett_asn1data); */

    iecgoose_asn1tree = proto_item_add_subtree(ti, ett_goosepdu); 
   /*initialize ASN! decoder  */


   asn1_open(&asn1, tvb, IEC_GOOSE_PDU);
	message_length = iecgoose_length - 8;

   asn1h_do_parse (&asn1, goose_message, "", &message_length, asn1.offset, iecgoose_asn1tree, tvb, ASN1H_TBL_SEQ,global_iecgoose_asn1_prefs_nodes_as_subtrees);
  }
}
#if 0
static gboolean
dissect_cltp_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *parent_tree)
{
	/* must check that this really is a smb packet */
	if (!tvb_bytes_exist(tvb, 0, 4))
		return FALSE;

	if( (tvb_get_guint8(tvb, 0) != 0xff)
	    || (tvb_get_guint8(tvb, 1) != 'S')
	    || (tvb_get_guint8(tvb, 2) != 'M')
	    || (tvb_get_guint8(tvb, 3) != 'B') ){
		return FALSE;
	}

 /*	dissect_smb(tvb, pinfo, parent_tree);*/
	return TRUE;
}
#endif

/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/


void
proto_register_IECGOOSE(void)
{                 

/* Setup list of header fields  See Section 1.6.1 for details*/
  static hf_register_info hf[] = {
    { &hf_iecgoose_appid_type, 
      { "AppID*","iecgoose.APPID", 	FT_UINT16,	BASE_DEC,	NULL,	0x0, "", HFILL }
    },
    { &hf_iecgoose_length_type, 
      { "PDU Length*","iecgoose.length",	FT_UINT16,	BASE_DEC, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecgoose_resv1_type,
      { "Reserved1*","iecgoose.resv1", FT_UINT16,BASE_HEX,NULL,0x0,"", HFILL }
    },
    { &hf_iecgoose_resv2_type,
      { "Reserved2*","iecgoose.resv2",FT_UINT16,	BASE_HEX, 	NULL,	0x0, "", HFILL }
    },
    { &hf_statenumber,
      { "  StateNumber*", "iecgoose.stateNumber", FT_UINT32,	BASE_DEC,NULL, 0x0,	"", HFILL }
    },
    { &hf_iecgoose_control_block_reference,
      { "  Control Block Reference*","iecgoose.gocbref", FT_STRINGZ,	BASE_NONE, 	NULL,0x0,"", HFILL }
    },
    { &hf_dataset,
      { "  DataSetReference*","iecgoose.datSetref", FT_STRINGZ,	BASE_NONE, 	NULL,0x0,"", HFILL }
    },
    { &hf_goid,
      { "  GOOSEID*","iecgoose.goid", FT_STRINGZ,	BASE_NONE, 	NULL,0x0,"", HFILL }
    },
    { &hf_test,
      { "  Test*","iecgoose.test", FT_STRINGZ,BASE_NONE,	NULL,0x0,"", HFILL }
    },
    { &hf_ndscomm,
      { "  Needs Commissioning*","iecgoose.ndscomm", FT_STRINGZ,BASE_NONE, 	NULL,0x0,"", HFILL }
    },
    { &hf_confrev,
      { "  Config Revision*","iecgoose.confrev", FT_UINT32,BASE_DEC, 	NULL,0x0,"", HFILL }
    },

  };

/* Setup protocol subtree array */
/* Initialize the subtree pointers */
static gint ett_asn1data = -1;
static gint ett_data = -1;
static gint ett_dataentry = -1;
static gint ett_statenumber = -1;

#if 0
	static gint *ett[] = {
		&ett_IECGOOSE,

	};
#endif

	static gint *ett[] = {
		&ett_IECGOOSE,
		&ett_goosepdu,
		&ett_gocbRef,
		&ett_statenumber,
		&ett_dataset,
		&ett_goid,
		&ett_data,
		&ett_datatbl,
		&ett_array,
		&ett_structure,
		&ett_bitstring,
		&ett_octettstring,
		&ett_string,
		&ett_gentime,
		&ett_utctime,
		&ett_bintime,
		&ett_unicode,
		&ett_bcd,
		&ett_oid,
	};

	module_t   *iecgoose_module;

/* Register the protocol name and description */
	proto_iecgoose = proto_register_protocol("IEC 61850 GOOSE",
	    "IECGOOSE", "iecgoose");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_iecgoose, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	iecgoose_module = prefs_register_protocol(proto_iecgoose, proto_reg_handoff_IECGOOSE);

	prefs_register_bool_preference(iecgoose_module, "asn1_nodes_as_subtrees", 
                                 "Show ASN1 nodes as subtrees", 
                                 "Show ASN1 nodes as subtrees",
                                 &global_iecgoose_asn1_prefs_nodes_as_subtrees);

/*	proto_register_subtree_array(ett1, array_length(ett1));
	proto_register_subtree_array(ett1, array_length(ett2));
	proto_register_subtree_array(ett1, array_length(ett3));
*/
}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_IECGOOSE(void)
{
	dissector_handle_t IECGOOSE_handle;


	IECGOOSE_handle = create_dissector_handle(dissect_IECGOOSE,
	    proto_iecgoose);
	dissector_add("ethertype", 0x88b8, IECGOOSE_handle);  


/*	heur_dissector_add("cltp", dissect_cltp_heur, proto_cltp); */


/*    dissector_add("chdlctype", ETHERTYPE_VLAN, IECGOOSE_handle); */

}

