/* packet-IECGSE.c
 * Routines for IEC 61850 IECGSE dissection
 * 2003, Herbert Falk <herb@sisconet.com>
 *
 * $Id: README.developer,v 1.86 2003/11/14 19:20:24 guy Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * Copied from WHATEVER_FILE_YOU_USED (where "WHATEVER_FILE_YOU_USED"
 * is a dissector file; if you just copied this from README.developer,
 * don't bother with the "Copied from" - you don't even need to put
 * in  "Copied from" if you copied an existing dissector, especially
 * if the bulk of the code in the new dissector is your code)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "etypes.h"				/*for the Ethertype IDs, may no longer be needed*/
#include "asn1.h"				/*for the ASN1 decoder functions */
#include "asn1_helper.h"

/* Initialize the protocol and registered fields */
static int proto_iecsmv = -1;

/* Initialize the subtree pointers */
static gint ett_IECGSE = -1;


static int hf_IECGSE_GSE  = -1;

static int hf_iecsmv_appid_type = -1;
static int hf_iecsmv_length_type = -1;
static int hf_iecsmv_resv1_type = -1;
static int hf_iecsmv_resv2_type = -1;
static int hf_iecsmv_data_type = -1;


static int hf_iecsmv_hard_size = -1;
static int hf_iecsmv_proto_size = -1;
static int hf_iecsmv_opcode = -1;
static int hf_iecsmv_src_hw = -1;
static int hf_iecsmv_src_hw_mac = -1;
static int hf_iecsmv_src_proto = -1;
static int hf_iecsmv_src_proto_id = -1;
static int hf_iecsmv_dst_hw = -1;
static int hf_iecsmv_dst_hw_mac = -1;
static int hf_iecsmv_dst_proto = -1;
static int hf_iecsmv_dst_proto_id = -1;
static int displayed_header = FALSE;		/*use to display SISCO information one time for advertisement*/

#ifndef AARP_REQUEST
#define AARP_REQUEST 	0x0001
#endif
#ifndef AARP_REPLY
#define AARP_REPLY	0x0002
#endif
#ifndef AARP_PROBE
#define AARP_PROBE	0x0003
#endif

/* The following is screwed up shit to deal with the fact that
   the linux kernel edits the packet inline. */
#define AARP_REQUEST_SWAPPED    0x0100
#define AARP_REPLY_SWAPPED  0x0200
#define AARP_PROBE_SWAPPED  0x0300

static const value_string op_vals[] = {
  {AARP_REQUEST,  "request" },
  {AARP_REPLY,    "reply"   },
  {AARP_PROBE,    "probe"   },
  {AARP_REQUEST_SWAPPED,  "request" },
  {AARP_REPLY_SWAPPED,    "reply"   },
  {AARP_PROBE_SWAPPED,    "probe"   },
  {0,             NULL           } };

/* AARP protocol HARDWARE identifiers. */
#define AARPHRD_ETHER 	1		/* Ethernet 10Mbps		*/
#define	AARPHRD_TR	2		/* Token Ring			*/


static const value_string hrd_vals[] = {
  {AARPHRD_ETHER,   "Ethernet"       },
  {AARPHRD_TR,      "Token Ring"     },
  {0,               NULL             } };


/* Initialize the subtree pointers */
static gint ett_asn1data = -1;
static gint ett_data = -1;
static gint ett_dataentry = -1;

#define IEC_APPID		0
#define IEC_LENGTH		2		
#define IEC_RESV1		4
#define IEC_RESV2		6
#define IEC_GSE_PDU		8


/* this function recurses through the data structures */

/*this function is used for the GSE PDU DATA production, there may be
  multiple data results, therefore it loops  
*/
#define ASN1H_SEQ_TAG	(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ)

static ENUM_LOOKUP error_reason1[2]= {
   {0,"Other"},
   {1, "Not Found"}};

static ENUM_XLAT error_reason = {2, error_reason1};


static ENUM_LOOKUP neg_error1[4]= {
   {0,"Other"},
   {1, "unkownControlBlock"},
   {2, "responseTooLarge"},
   {3, "controlBlockConfigurationError"}};

static ENUM_XLAT neg_error = {4, neg_error1};

#define NOET NULL
#define NOHF NULL

static ASN1_TBL_ENTRY results_tbl[]={
		{(ASN1H_CTX | 0),NOET,NOHF,"Offset: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOET,NOHF,"Reference: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 2),NOET,NOHF,"Error: ",ASN1H_INTEGER_ENUM_DISPLAY,(struct asn1h_tbl_entry*)&error_reason},
   		{ASN1H_END_OF_TABLE	,NOET,NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY gse_mngt_resp_pos[]={
		{(ASN1H_CTX | 0),NOET,NOHF,"DataSet: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),NOET,NOHF,"Results: ",ASN1H_TBL_SET, results_tbl},
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY gse_mngt_resp[]={
		{(ASN1H_CTX | 0),NOET,NOHF,"Ident: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOET,NOHF,"Config Rev: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),NOET,NOHF,"Response Positive",ASN1H_TBL_SET, gse_mngt_resp_pos},
		{(ASN1H_CTX | 3),NOET,NOHF,"Response Negative: ",ASN1H_INTEGER_ENUM_DISPLAY,(struct asn1h_tbl_entry*)&neg_error },
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY gse_responses[]={
		{(ASN1H_CTX | 0),NOET,NOHF,"GSE MNGT Not Supported",NULL, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),NOET,NOHF,"Get GOOSE Reference",ASN1H_TBL_SEQ, gse_mngt_resp},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),NOET,NOHF,"Get GOOSE Element Number",ASN1H_TBL_SEQ, gse_mngt_resp},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 3),NOET,NOHF,"Get GSSE Reference",ASN1H_TBL_SEQ, gse_mngt_resp},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 4),NOET,NOHF,"Get GSSE Data Offset",ASN1H_TBL_SEQ, gse_mngt_resp},
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};


static ASN1_TBL_ENTRY gse_getelem_request[]={
		{(ASN1H_CTX | 0),NOET,NOHF,"Ident: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),NOET,NOHF,"List of References",NULL, NULL},
		{(ASN1H_UNI | ASN1_VISSTR),NOET,NOHF,"Reference: ",ASN1H_VSTRING_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};


static ASN1_TBL_ENTRY gse_getref_request[]={
		{(ASN1H_CTX | 0),NOET,NOHF,"Ident: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),NOET,NOHF,"List of Offsets",NULL, NULL},
		{(ASN1H_UNI | ASN1_INT),NOET,NOHF,"Offset: ",ASN1H_UNSIGNED_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};


static ASN1_TBL_ENTRY gse_requests[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),NOET,NOHF,"Get GOOSE Reference: ",ASN1H_TBL_SET, gse_getref_request},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),NOET,NOHF,"Get GOOSE Element Number",ASN1H_TBL_SET, gse_getelem_request},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 3),NOET,NOHF,"Get GSSE Reference: ",ASN1H_TBL_SET, gse_getref_request},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 4),NOET,NOHF,"Get GSSE Data Offset",ASN1H_TBL_SET, gse_getelem_request},
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY gse_main[]={
		{(ASN1H_APP |ASN1H_CONSTRUCT | 0),NOET,NOHF,"GSE Management PDU: ",NULL, NULL},
		{(ASN1H_CTX | 0),NOET,NOHF,"State ID: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 3),NOET,NOHF,"Security: ",ASN1H_OCTETSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),NOET,NOHF,"GSE Request ",ASN1H_TBL_SEQ, gse_requests},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),NOET,NOHF,"GSE Response ",ASN1H_TBL_SEQ, gse_responses},
   		{ASN1H_END_OF_TABLE,NOET,NOHF,NULL,NULL,NULL}
};


/* Code to actually dissect the packets */
static void
dissect_IECGSE(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  guint16	  iecsmv_APPID;
  guint16	  iecsmv_resv1;
  guint16	  iecsmv_resv2;
  proto_tree  *iecsmv_tree;
  proto_tree  *iecsmv_asn1tree;
  ASN1_SCK asn1;
  guint message_length;
  guint iecsmv_length;




/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti;

/* Make entries in Protocol column and Info column on summary display */
	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "IECGSE");
    
/* This field shows up as the "Info" column in the display; you should make
   it, if possible, summarize what's in the packet, so that a user looking
   at the list of packets can tell what type of packet it is. See section 1.5
   for more information.

   If you are setting it to a constant string, use "col_set_str()", as
   it's more efficient than the other "col_set_XXX()" calls.

   If you're setting it to a string you've constructed, or will be
   appending to the column later, use "col_add_str()".

   "col_add_fstr()" can be used instead of "col_add_str()"; it takes
   "printf()"-like arguments.  Don't use "col_add_fstr()" with a format
   string of "%s" - just use "col_add_str()" or "col_set_str()", as it's
   more efficient than "col_add_fstr()".

   If you will be fetching any data from the packet before filling in
   the Info column, clear that column first, in case the calls to fetch
   data from the packet throw an exception because they're fetching data
   past the end of the packet, so that the Info column doesn't have data
   left over from the previous dissector; do

	if (check_col(pinfo->cinfo, COL_INFO)) 
		col_clear(pinfo->cinfo, COL_INFO);

   */
    if (check_col(pinfo->cinfo, COL_INFO)) 
    {
     if (displayed_header != TRUE)
      {
	  col_set_str(pinfo->cinfo, COL_INFO, "GSE Decoder, Courtesy of Herbert Falk of SISCO");
	  displayed_header = TRUE;
	  }
	}


  iecsmv_APPID =  tvb_get_ntohs(tvb, IEC_APPID);
  iecsmv_length = tvb_get_ntohs(tvb, IEC_LENGTH);
  iecsmv_resv1= tvb_get_ntohs(tvb, IEC_RESV1);
  iecsmv_resv2=  tvb_get_ntohs(tvb, IEC_RESV2);



/* In the interest of speed, if "tree" is NULL, avoid building a
   protocol tree and adding stuff to it if possible.  Note,
   however, that you must call subdissectors regardless of whether
   "tree" is NULL or not. */
	if (tree) {

/* NOTE: The offset and length values in the call to
   "proto_tree_add_item()" define what data bytes to highlight in the hex
   display window when the line in the protocol tree display
   corresponding to that item is selected.

   Supplying a length of -1 is the way to highlight all data from the
   offset to the end of the packet. */

/* create display subtree for the protocol */
		ti = proto_tree_add_item(tree, proto_iecsmv, tvb, 0, -1, FALSE);


/* add an item to the subtree, see section 1.6 for more information */
/* this needs to be put back in $$$$ Herbert Falk */
/*	proto_tree_add_item(IECGSE_tree,
		    hf_IECGSE_GSE, tvb, offset, len, FALSE); */

   iecsmv_tree = proto_item_add_subtree(ti, ett_IECGSE);

   proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_appid_type, tvb, IEC_APPID, 2,
			       iecsmv_APPID);

   proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_length_type, tvb, IEC_LENGTH, 2,
			       iecsmv_length);

   proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_resv1_type, tvb, IEC_RESV1, 2,
			       iecsmv_resv1);

   proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_resv2_type, tvb, IEC_RESV2, 2,
			       iecsmv_resv2);
   ti = proto_tree_add_text(iecsmv_tree, tvb, IEC_GSE_PDU, iecsmv_length-IEC_GSE_PDU, "PDU");

 /* ready to start decoding the PDU */

   /*add the subtree  */
 /*  iecsmv_asn1tree = proto_item_add_subtree(ti, ett_asn1data); */

    iecsmv_asn1tree = proto_item_add_subtree(ti, ett_IECGSE); /*_tree was _subtree */
   /*initialize ASN! decoder  */


    asn1_open(&asn1, tvb, IEC_GSE_PDU);
	message_length = iecsmv_length - 8;
    asn1h_do_parse (&asn1, gse_main, "", &message_length, asn1.offset, ti, tvb, ASN1H_TBL_SEQ,FALSE);
  }

}




/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_IECGSE(void)
{                 

/* Setup list of header fields  See Section 1.6.1 for details*/
  static hf_register_info hf[] = {
    { &hf_iecsmv_appid_type,
      { "AppID",		"iecgse.APPID",
	FT_UINT16,	BASE_HEX,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_length_type,
      { "PDU Length",		"iecgse.length",
	FT_UINT16,	BASE_DEC, 	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_resv1_type,
      { "Reserved1",		"iecgse.resv1",
	FT_UINT16,	BASE_HEX, 	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_resv2_type,
      { "Reserved2",		"iecgse.resv2",
	FT_UINT16,	BASE_HEX, 	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_data_type,
      { "Reserved2",		"iecgse.resv2",
	FT_NONE,	BASE_HEX, 	NULL,	0x0,
      	"", HFILL }},

/*
    { &hf_iecsmv_hard_size,
      { "Hardware size",		"iecsmv.hard.size",
	FT_UINT8,	BASE_DEC, 	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_proto_size,
      { "Protocol size",		"iecsmv.proto.size",
	FT_UINT8,	BASE_DEC, 	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_opcode,
      { "Opcode",			"iecsmv.opcode",
	FT_UINT16,	BASE_DEC,	VALS(op_vals),	0x0,
      	"", HFILL }},

    { &hf_iecsmv_src_hw,
      { "Sender hardware address",	"iecsmv.src.hw",
	FT_BYTES,	BASE_NONE,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_src_hw_mac,
      { "Sender MAC address",		"iecsmv.src.hw_mac",
	FT_ETHER,	BASE_NONE,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_src_proto,
      { "Sender protocol address",	"iecsmv.src.proto",
	FT_BYTES,	BASE_NONE,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_src_proto_id,
      { "Sender ID",			"iecsmv.src.proto_id",
	FT_BYTES,	BASE_HEX,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_dst_hw,
      { "Target hardware address",	"iecsmv.dst.hw",
	FT_BYTES,	BASE_NONE,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_dst_hw_mac,
      { "Target MAC address",		"iecsmv.dst.hw_mac",
	FT_ETHER,	BASE_NONE,	NULL,	0x0,
      	"", HFILL }},

    { &hf_iecsmv_dst_proto,
      { "Target protocol address",	"iecsmv.dst.proto",
	FT_BYTES,	BASE_NONE,	NULL,	0x0,
      "", HFILL }},

    { &hf_iecsmv_dst_proto_id,
      { "Target ID",			"iecsmv.dst.proto_id",
	FT_BYTES,	BASE_HEX,	NULL,	0x0,
      	"", HFILL }},										*/
  };

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_IECGSE,

	};

	static gint *ett1[] = {
		&ett_asn1data,
		&ett_data,
		&ett_dataentry,
	};

	static gint *ett2[] = {
		&ett_data,

	};

	static gint *ett3[] = {
		&ett_dataentry,	
	};

/* Register the protocol name and description */
	proto_iecsmv = proto_register_protocol("IEC 61850 GSE",
	    "IECGSE", "iecgse");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_iecsmv, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

/*	proto_register_subtree_array(ett1, array_length(ett1));
	proto_register_subtree_array(ett1, array_length(ett2));
	proto_register_subtree_array(ett1, array_length(ett3));
*/

}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_IECGSE(void)
{
	dissector_handle_t IECGSE_handle;


	IECGSE_handle = create_dissector_handle(dissect_IECGSE,
	    proto_iecsmv);
	dissector_add("ethertype", 0x88b9, IECGSE_handle);  

}

