/* packet-IECGOOSE.c
 * Routines for IEC 61850 IECGOOSE dissection
 * 2003, Herbert Falk <herb@sisconet.com>
 *
 * $Id: README.developer,v 1.86 2003/11/14 19:20:24 guy Exp $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 * Copied from WHATEVER_FILE_YOU_USED (where "WHATEVER_FILE_YOU_USED"
 * is a dissector file; if you just copied this from README.developer,
 * don't bother with the "Copied from" - you don't even need to put
 * in  "Copied from" if you copied an existing dissector, especially
 * if the bulk of the code in the new dissector is your code)
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <glib.h>

#ifdef NEED_SNPRINTF_H
# include "snprintf.h"
#endif

#include <epan/packet.h>
#include "etypes.h"				/*for the Ethertype IDs, may no longer be needed*/
#include "asn1.h"				/*for the ASN1 decoder functions */
#include "asn1_helper.h"

void proto_reg_handoff_IECSMV(void);

/* Initialize the protocol and registered fields */
static int proto_iecsmv = -1;

/* Initialize the subtree pointers */
static gint ett_IECSMV = -1;
static gint ett_ASDU = -1;
static gint ett_startASDU = -1;
static gint ett_92 = -1;

#define NOETT NULL
#define NOHF  NULL

static int hf_IECSMV_SMV  = -1;

static int hf_iecsmv_appid_type = -1;
static int hf_iecsmv_length_type = -1;
static int hf_iecsmv_resv1_type = -1;
static int hf_iecsmv_resv2_type = -1;
static int hf_iecsmv_data_type = -1;
static int hf_iecsmv_id = -1;
static int hf_iecsmv_dataset = -1;
static int hf_iecsmv_confrev = -1;
static int hf_iecsmv_samplesync = -1;

static gboolean	iecsmv_profile = FALSE;
static gboolean smv_typefrommib = TRUE;
static gboolean smv_nodes_as_subtrees = FALSE;

static int hf_iecsmv_hard_size = -1;
static int hf_iecsmv_proto_size = -1;
static int hf_iecsmv_opcode = -1;
static int hf_iecsmv_src_hw = -1;
static int hf_iecsmv_src_hw_mac = -1;
static int hf_iecsmv_src_proto = -1;
static int hf_iecsmv_src_proto_id = -1;
static int hf_iecsmv_dst_hw = -1;
static int hf_iecsmv_dst_hw_mac = -1;
static int hf_iecsmv_dst_proto = -1;
static int hf_iecsmv_dst_proto_id = -1;
static int displayed_header = FALSE;		/*use to display SISCO information one time for advertisement*/

/* Initialize the subtree pointers */
static gint ett_asn1data = -1;
static gint ett_data = -1;
static gint ett_dataentry = -1;

#define IEC_APPID		0
#define IEC_LENGTH		2		
#define IEC_RESV1		4
#define IEC_RESV2		6
#define IEC_SMV_PDU		8


/* this function recurses through the data structures */

/*this function is used for the SMV PDU DATA production, there may be
  multiple data results, therefore it loops  
*/
#define ASN1H_SEQ_TAG	(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ)
static display_quality (guchar *disp_offset, gchar *disp_string, guchar *int_ptr, guint32 offset, proto_item *subtree, tvbuff_t *tvb)
{
guint qual_value = 0;
gchar *disp_text;	    
guint  myint;

#define QUAL_DER_MASK 0x4000 
#define QUAL_OPB_MASK 0x2000
#define QUAL_TEST_MASK 0x1000
#define QUAL_SRC_MASK  0x0C00
#define QUAL_OVERFLOW_MASK 0x0200
#define QUAL_OUTOFRANGE_MASK 0x0100
#define QUAL_BADREF_MASK 0x0080
#define QUAL_OSC_MASK 0x0040
#define QUAL_FAIL_MASK 0x0020
#define QUAL_OLD_MASK 0x0010
#define QUAL_INCONSISTENT_MASK 0x0008
#define QUAL_INACCURATE_MASK 0x0004
#define QUAL_VAL_MASK 0x0003


   qual_value = qual_value + int_ptr[2];
   qual_value = qual_value <<8;
   qual_value = qual_value + int_ptr[3];
   disp_text = g_malloc(strlen(disp_offset) + strlen(disp_string) + 200);
   sprintf(disp_text,"%s  %s:",disp_offset, disp_string);

   if(qual_value & QUAL_DER_MASK)		
     sprintf(&disp_text[strlen(disp_text)], " der");

   if(qual_value & QUAL_OPB_MASK)		
     sprintf(&disp_text[strlen(disp_text)], " OpB");

   if(qual_value & QUAL_TEST_MASK)		
     sprintf(&disp_text[strlen(disp_text)], " Test");

   myint = (qual_value & QUAL_SRC_MASK) >> 2;
   sprintf(&disp_text[strlen(disp_text)]," Source= ");
   switch (myint)
     {
	 case 0:
	     sprintf(&disp_text[strlen(disp_text)],"process");
	   break;

     case 1:
 	     sprintf(&disp_text[strlen(disp_text)],"substituted");
	   break;

     default:
	     sprintf(&disp_text[strlen(disp_text)],"unknown");
     break;
	 }

   if (qual_value & QUAL_OVERFLOW_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," Overflow");

   if (qual_value & QUAL_OUTOFRANGE_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," OutofRange");

   if (qual_value & QUAL_BADREF_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," BadRef");

   if (qual_value & QUAL_OSC_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," Oscillatory");

   if (qual_value & QUAL_FAIL_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," Failure");

   if (qual_value & QUAL_OLD_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," OldData");

   if (qual_value & QUAL_INCONSISTENT_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," Inconsistent");

   if (qual_value & QUAL_INACCURATE_MASK)
	     sprintf(&disp_text[strlen(disp_text)]," Inaccurate");

   myint = qual_value & QUAL_VAL_MASK;
   switch(myint)
   {
    case 0: 
	     sprintf(&disp_text[strlen(disp_text)]," validity = good");
		 break;

    case 1: 
	     sprintf(&disp_text[strlen(disp_text)]," validity = invalid");
		 break;

    case 2: 
	     sprintf(&disp_text[strlen(disp_text)]," validity = reserved");
		 break;

    case 3: 
	     sprintf(&disp_text[strlen(disp_text)]," validity =questionable");
		 break;

   }
        
   proto_tree_add_text (subtree,tvb,offset,4,disp_text);
   g_free(disp_text);  
}


/* function for special decode of sample data */
static display_int (guchar *disp_offset, gchar *disp_string, guchar *int_ptr, guint32 offset, proto_item *subtree, tvbuff_t *tvb)
{
gint32 int_value = 0;
gchar *disp_text;

   int_value = (gint32) int_ptr[0] & 0xff;
   int_value = int_value <<8;
   int_value = int_value + int_ptr[1];
   int_value = int_value <<8;
   int_value = int_value + int_ptr[2];
   int_value = int_value <<8;
   int_value = int_value + int_ptr[3];
   disp_text = g_malloc(strlen(disp_offset) + strlen(disp_string) + 40);
   sprintf(disp_text,"%s  %s: %ld",disp_offset, disp_string, int_value);
   proto_tree_add_text (subtree,tvb,offset,4,disp_text);
   g_free(disp_text);  
}

static 
void spec_decode_func (guint32 length, guchar *dptr, guchar	*disp_string, guint32 offset, proto_item *subtree, tvbuff_t *tvb)
{
  if ((iecsmv_profile == TRUE) && (length == 64)) 	/*do the initial profile */
    {
	  display_int (disp_string, "ImATCTRL1.Amp.instMag.i",dptr, offset, subtree, tvb);
	  display_quality (disp_string, "ImATCTRL1.Amp.q",&dptr[4], offset+4, subtree, tvb);
	  display_int (disp_string, "ImBTCTRL2.Amp.instMag.i", &dptr[8], offset +8, subtree, tvb);
	  display_quality (disp_string, "ImBTCTRL2.Amp.q", &dptr[12], offset +12, subtree, tvb);
	  display_int (disp_string, "ImCTCTRL3.Amp.instMag.i", &dptr[16], offset +16, subtree, tvb);
	  display_quality (disp_string, "ImCTCTRL3.Amp.q", &dptr[20], offset +20, subtree, tvb);   
	  display_int (disp_string, "ImNmTCTRL4.Amp.instMag.i", &dptr[24], offset +24, subtree, tvb);
	  display_quality (disp_string, "ImNmTCTRL4.Amp.q", &dptr[28], offset +28, subtree, tvb);
	  display_int (disp_string, "UmnATVTR1.Vol.instMag.i",&dptr[32], offset + 32, subtree, tvb);
	  display_quality (disp_string, "UmnATVTR1.Vol.q",&dptr[36], offset + 36, subtree, tvb);
	  display_int (disp_string, "UmnBTVTR2.Vol.instMag.i",&dptr[40], offset + 40, subtree, tvb);
	  display_quality (disp_string, "UmnBTVTR2.Vol.q",&dptr[44], offset + 44, subtree, tvb);
	  display_int (disp_string, "UmnCTVTR3.Vol.instMag.i",&dptr[48], offset + 48, subtree, tvb);
	  display_quality (disp_string, "UmnCTVTR3.Vol.q",&dptr[52], offset + 52, subtree, tvb);
	  display_int (disp_string, "UmnATVTR4.Vol.instMag.i",&dptr[56], offset + 56, subtree, tvb);
	  display_quality (disp_string, "UmnATVTR4.Vol.q",&dptr[60], offset + 60, subtree, tvb);

	}						  
} 
static ASN1_TBL_ENTRY smv_asdu_start[]={
		{(ASN1H_CTX | 0),NOETT,&hf_iecsmv_id,"",ASN1H_VSTRING_DISPLAY, NULL},  /*need to check this tag */
		{(ASN1H_CTX | 1),NOETT,&hf_iecsmv_dataset,"",ASN1H_VSTRING_DISPLAY, NULL}, /*need to check this tag*/
		{(ASN1H_CTX | 2),NOETT,NOHF,"Sample Count: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 3),NOETT,&hf_iecsmv_confrev,"",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 4),NOETT,NOHF,"RefrTm: ",ASN1H_MMS_UTCTIME_DISPLAY, NULL},
		{(ASN1H_CTX | 5),NOETT,&hf_iecsmv_samplesync,"",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 6),NOETT,NOHF,"Sample Rate: ",ASN1H_UNSIGNED_DISPLAY, NULL},
 	    {(ASN1H_CTX |  7),NOETT,NOHF,"Samples ",ASN1H_NONASN1_DECODE, (void *)spec_decode_func},
   		{ASN1H_END_OF_TABLE	,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY smv_asdu[]={
		{ASN1H_SEQ_TAG, &ett_ASDU, NOHF, "ASDU",ASN1H_TBL_SEQ, smv_asdu_start},
   		{ASN1H_END_OF_TABLE	,NULL,NULL,NULL}
};


static ASN1_TBL_ENTRY smv_start[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"Number of ASDUs: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),&ett_startASDU,NOHF,"Start of ASDUs",ASN1H_TBL_SET, smv_asdu},
   		{ASN1H_END_OF_TABLE	,NULL,NULL,NULL}
};

static ASN1_TBL_ENTRY smv_publish[]={
		{(ASN1H_APP | 0),NOETT,NOHF,"SMV 9-1 not decoded: ",NULL, NULL},
		{(ASN1H_APP | ASN1H_CONSTRUCT | 0),&ett_92, NOHF,"SMV 9-2 ",ASN1H_TBL_SEQ, smv_start},
   		{ASN1H_END_OF_TABLE	,NULL,NULL,NULL}
};


/* Code to actually dissect the packets */
static void
dissect_IECSMV(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  guint16	  iecsmv_APPID;
  guint16	  iecsmv_length;
  guint16	  iecsmv_resv1;
  guint16	  iecsmv_resv2;
  guint8	  iecsmv_APPLID;
  proto_tree  *iecsmv_tree;
  proto_tree  *iecsmv_asn1tree;
  ASN1_SCK asn1;
  guint message_length;




/* Set up structures needed to add the protocol subtree and manage it */
	proto_item *ti;

/* Make entries in Protocol column and Info column on summary display */
	if (check_col(pinfo->cinfo, COL_PROTOCOL)) 
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "IECSMV");
    
/* This field shows up as the "Info" column in the display; you should make
   it, if possible, summarize what's in the packet, so that a user looking
   at the list of packets can tell what type of packet it is. See section 1.5
   for more information.

   If you are setting it to a constant string, use "col_set_str()", as
   it's more efficient than the other "col_set_XXX()" calls.

   If you're setting it to a string you've constructed, or will be
   appending to the column later, use "col_add_str()".

   "col_add_fstr()" can be used instead of "col_add_str()"; it takes
   "printf()"-like arguments.  Don't use "col_add_fstr()" with a format
   string of "%s" - just use "col_add_str()" or "col_set_str()", as it's
   more efficient than "col_add_fstr()".

   If you will be fetching any data from the packet before filling in
   the Info column, clear that column first, in case the calls to fetch
   data from the packet throw an exception because they're fetching data
   past the end of the packet, so that the Info column doesn't have data
   left over from the previous dissector; do

	if (check_col(pinfo->cinfo, COL_INFO)) 
		col_clear(pinfo->cinfo, COL_INFO);

   */
	iecsmv_APPLID= tvb_get_guint8(tvb, IEC_SMV_PDU);
	if(iecsmv_APPLID == 0x60)			/*then we have a 9-2 SMV */
	{
	  if (check_col(pinfo->cinfo, COL_INFO)) 
	  {
	    if (displayed_header == TRUE)
		  col_set_str(pinfo->cinfo, COL_INFO, "SMV 9-2 Publish");
	    else
		{
		col_set_str(pinfo->cinfo, COL_INFO, "SMV 9-2 Publish Decoder, Courtesy of Herbert Falk of SISCO");
		displayed_header = TRUE;
		}
	  }
	}
	else
	{
	  if (check_col(pinfo->cinfo, COL_INFO)) 
	  {
	    if (displayed_header == TRUE)
		  col_set_str(pinfo->cinfo, COL_INFO, "SMV 9-1 Publish");
	    else
		{
		col_set_str(pinfo->cinfo, COL_INFO, "SMV 9-1 Publish Decoder, Courtesy of Herbert Falk of SISCO");
		displayed_header = TRUE;
		}
	  }
	}

/*get information from the TVB structure */
/*  vlan_tpid= tvb_get_ntohs(tvb, VLAN_TPID);
  vlan_tci= tvb_get_ntohs(tvb, VLAN_TCI);
  iecsmv_ethertype = tvb_get_ntohs(tvb, IEC_ETHERTYPE);  not needed */

  iecsmv_APPID =  tvb_get_ntohs(tvb, IEC_APPID);
  iecsmv_length = tvb_get_ntohs(tvb, IEC_LENGTH);
  iecsmv_resv1= tvb_get_ntohs(tvb, IEC_RESV1);
  iecsmv_resv2=  tvb_get_ntohs(tvb, IEC_RESV2);

/*  iecsmv_pro = tvb_get_ntohs(tvb, AR_PRO);
  iecsmv_hln = tvb_get_guint8(tvb, AR_HLN);
  iecsmv_pln = tvb_get_guint8(tvb, AR_PLN);
  iecsmv_op  = tvb_get_ntohs(tvb, AR_OP);   */


/* In the interest of speed, if "tree" is NULL, avoid building a
   protocol tree and adding stuff to it if possible.  Note,
   however, that you must call subdissectors regardless of whether
   "tree" is NULL or not. */
	if (tree) 
	{


/* create display subtree for the protocol */
		ti = proto_tree_add_item(tree, proto_iecsmv, tvb, 0, -1, FALSE);

/*		IECSMV_tree = proto_item_add_subtree(ti, ett_IECSMV); */

/* add an item to the subtree, see section 1.6 for more information */
/* this needs to be put back in $$$$ Herbert Falk */
/*	proto_tree_add_item(IECSMV_tree,
		    hf_IECSMV_SMV, tvb, offset, len, FALSE); */

       iecsmv_tree = proto_item_add_subtree(ti, ett_IECSMV);

       proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_appid_type, tvb, IEC_APPID, 2,
			       iecsmv_APPID);

       proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_length_type, tvb, IEC_LENGTH, 2,
			       iecsmv_length);

       proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_resv1_type, tvb, IEC_RESV1, 2,
			       iecsmv_resv1);

       proto_tree_add_uint(iecsmv_tree ,hf_iecsmv_resv2_type, tvb, IEC_RESV2, 2,
			       iecsmv_resv2);
       ti = proto_tree_add_text(iecsmv_tree, tvb, IEC_SMV_PDU, iecsmv_length-IEC_SMV_PDU, "PDU");

 /* ready to start decoding the PDU */

   /*add the subtree  */
 /*  iecsmv_asn1tree = proto_item_add_subtree(ti, ett_asn1data); */

       iecsmv_asn1tree = proto_item_add_subtree(ti, ett_IECSMV); /*_tree was _subtree */
   /*initialize ASN! decoder  */


       asn1_open(&asn1, tvb, IEC_SMV_PDU);
/*       asn1h_do_parse (&asn1, smv_publish, "", &message_length, asn1.offset, ti, tvb, ASN1H_TBL_SEQ); */
       asn1h_do_parse (&asn1, smv_publish, "", &message_length, asn1.offset, iecsmv_asn1tree, tvb, ASN1H_TBL_SEQ,smv_nodes_as_subtrees);

    }

}




/* Register the protocol with Ethereal */

/* this format is require because a script is used to build the C function
   that calls all the protocol registration.
*/

void
proto_register_IECSMV(void)
{                 

/* Setup list of header fields  See Section 1.6.1 for details*/
  static hf_register_info hf[] = {
    { &hf_iecsmv_appid_type, 
      { "AppID*","iecsmv.APPID",FT_UINT16,	BASE_HEX,	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_length_type,
      { "PDU Length*", "iecsmv.length",	FT_UINT16,	BASE_DEC, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_resv1_type,
      { "Reserved1*","iecsmv.resv1", FT_UINT16,	BASE_HEX, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_resv2_type,
      { "Reserved2*","iecsmv.resv2",FT_UINT16,	BASE_HEX, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_id,
      { "ID*","iecsmv.id",FT_STRING,	BASE_NONE, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_dataset,
      { "DataSet*","iecsmv.dataset",FT_STRING,	BASE_NONE, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_confrev,
      { "Config Rev*","iecsmv.confrev",FT_UINT32,	BASE_DEC, 	NULL,	0x0,"", HFILL }
    },
    { &hf_iecsmv_samplesync,
      { "Sample Synched*","iecsmv.samplesync",FT_STRING,	BASE_NONE, 	NULL,	0x0,"", HFILL }
    },


  };

/* Setup protocol subtree array */
	static gint *ett[] = {
		&ett_IECSMV,

	};

	static gint *ett1[] = {
		&ett_asn1data,
		&ett_data,
		&ett_dataentry,
	};

	static gint *ett2[] = {
		&ett_data,

	};

	static gint *ett3[] = {
		&ett_dataentry,	
	};

  module_t* smv_module;

/* Register the protocol name and description */
	proto_iecsmv = proto_register_protocol("IEC 61850 SMV",
	    "IECSMV", "iecsmv");

/* Required function calls to register the header fields and subtrees used */
	proto_register_field_array(proto_iecsmv, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

/*	proto_register_subtree_array(ett1, array_length(ett1));
	proto_register_subtree_array(ett1, array_length(ett2));
	proto_register_subtree_array(ett1, array_length(ett3));
*/
  smv_module = prefs_register_protocol(proto_iecsmv, proto_reg_handoff_IECSMV);
  prefs_register_bool_preference(smv_module, "profile",
                                 "Interpret Sampled Data per 9-2 Implementation Guide R2-0a",
                                 "Whether to decode the Samples per a pre-defined DataSet as defined in the implementation guide",
                                 &iecsmv_profile);
  prefs_register_bool_preference(smv_module, "asn1_nodes_as_subtrees", 
                                 "Show ASN1 nodes as subtrees", 
                                 "Show ASN1 nodes as subtrees",
                                 &smv_nodes_as_subtrees);


}


/* If this dissector uses sub-dissector registration add a registration routine.
   This format is required because a script is used to find these routines and
   create the code that calls these routines.
*/
void
proto_reg_handoff_IECSMV(void)
{
  static int smv_prefs_initialized = FALSE;
  dissector_handle_t IECSMV_handle;

  
  if (!smv_prefs_initialized) {
	IECSMV_handle = create_dissector_handle(dissect_IECSMV, proto_iecsmv);
    smv_prefs_initialized = TRUE;
  } else
	IECSMV_handle = create_dissector_handle(dissect_IECSMV, proto_iecsmv);

	dissector_add("ethertype", 0x88bA, IECSMV_handle);  

}

