/* packet-ieeec37-118.c
 * Routines for IEEE C37.118 Synchrophaspr
 *  packet dissection;
 * Copyright 2005, Herbert Falk <herb@sisconet.com>
 *
 * $Id: packet-ieeec37-118.c 15361 2005-10-30 20:28:04Z falk $
 *
 * Ethereal - Network traffic analyzer
 * By Gerald Combs <gerald@ethereal.com>
 * Copyright 1998 Gerald Combs
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
/* Telnet authentication options as per     RFC2941
 * Kerberos v5 telnet authentication as per RFC2942
 */
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>

#include <string.h>
#include <glib.h>
#include <epan/filesystem.h>
#include <epan/packet.h>
#include <epan/strutil.h>
#include <epan/packet.h>
#include <epan/strutil.h>
#include <epan/emem.h>
#include "packet-tcp.h"
#include <epan/filesystem.h>
#include <epan/xmlstub.h>
#include "prefs.h"




static gchar *quality_strings[] = {
  "Normal operation, clock locked",
  "Clock unlocked, time within 10^-9 sec",
  "Clock unlocked, time within 10^-8 sec",
  "Clock unlocked, time within 10^-7 sec",
  "Clock unlocked, time within 10^-6 sec",
  "Clock unlocked, time within 10^-5 sec",
  "Clock unlocked, time within 10^-4 sec",
  "Clock unlocked, time within 10^-3 sec",
  "Clock unlocked, time within 10^-2 sec",
  "Clock unlocked, time within 10^-1 sec",
  "Clock unlocked, time within 1 sec",
  "Clock unlocked, time within 10 sec",
  "Unkown",
  "Unkown",
  "Unkown",
  "Fault - clock failure, time not reliable"   
  };
static gint saved_framesize =0;		/*used to store encountered framesize */
static int proto_c37_118 = -1;
static int hf_saved_PMU_ID = -1;

#define CONFIG1_REQ "Config1Req"
#define CONFIG2_REQ "Config2Req"  
#define ENABLE_REQ "Enable"
#define DISABLE_REQ "Disable"
#define EXTENDED_REQ "Extended"
#define UNKNOWN_REQ "UnknownReq"
#define HEADER_REQ "HeaderReq"
#define CONFIG1_RESP "Config1Resp"
#define CONFIG2_RESP "Config2Resp"
#define HEADER_RESP "HeaderResp"
#define DATA_MSG "Data"
#define UNKNOWN_RESP "UnkownResp"
static int hf_saved_pdu_type = -1;
static int hf_saved_config_station_name = -1;
static int hf_saved_config_channel_name = -1;
static guint16 myPMU;


static gint ett_c37 = -1;
static gint ett_c37_subopt1 = -1;
static gint ett_c37_subopt2 = -1;
static gint ett_c37_subopt3 = -1;
static gint ett_c37_timestamp = -1;
static gint ett_c37_status = -1;
static gint ett_c37_chan_info = -1;
static gint ett_c37_indv_chan_info = -1;

void proto_reg_handoff_c37_118(void);

/* Some defines for c37-118 */

#define TCP_PORT_C37_118	4712
#define UDP_PORT_C37_118	4713

#define C37_SYNC   170
#define C37_DATA_PACKET 0
#define C37_HEADER_PACKET 1
#define C37_CONFIG1_PACKET 2
#define C37_CONFIG2_PACKET 3
#define C37_COMMAND_PACKET 4
#define C37_EXTENDED_PACKET 4

typedef struct {
#define PHASOR_CHANNEL 0			/*list of the type of channel*/
#define ANALOG_CHANNEL 1
#define DIGITAL_CHANNEL 2
   guint16 type_of_channel;
   guchar *channame;	   
   guint16 channame_offset;
   guint32 unit_information;
   guint32 unit_offset;
   guint16 normal_state;
   guint16 valid_channel;
   } TRACK_INFO;


/*preference support section */
typedef enum {
  DATA_DECODE_NONE,
  DATA_DECODE_SINGLE,
  DATA_DECODE_FILE_BASED
} Version_Type;

static const enum_val_t options[] = {
  { "None", "Do not perform binary data decode",  DATA_DECODE_NONE  },
  { "Single Set", "Based upon preferences ", DATA_DECODE_SINGLE },
  { "Configuration Files", "Decode binary data based upon configuration files in specified directory ",
  		 DATA_DECODE_FILE_BASED },
  { NULL, NULL, 0 }
};
static Version_Type gbl_type = DATA_DECODE_NONE;

typedef enum {
  DATA_DECODED_AS_INTEGER,
  DATA_DECODED_AS_FLOAT
} Data_Type;   

static const enum_val_t type_options[] = {
  { "INTEGER", "Convert to Integer Values",  DATA_DECODED_AS_INTEGER  },
  { "FLOAT", "Convert to Floating Point Value ", DATA_DECODED_AS_FLOAT },
  { NULL, NULL, 0 }
};

static Version_Type analog_decode_type = DATA_DECODED_AS_INTEGER;
static Version_Type phasor_decode_type = DATA_DECODED_AS_INTEGER;
static Version_Type frequency_decode_type = DATA_DECODED_AS_INTEGER;

#define DICT_FN  "\.\\pmu_config"
static gboolean pmu_configuration_directory_usage = FALSE;
static const gchar *gbl_diameterDictionary;
static guint glb_num_phasors=3;
static guint glb_num_analogs=0;
static guint glb_num_discretes=0;

typedef struct{
  gchar *name;
  double  conversion_factor;
  }CHANNEL_CONFIG;
  
typedef struct{
 guint num_phasors;
 guint type_of_phasors;
 CHANNEL_CONFIG *phasor_config;	  /*to to allocated as needed */
 guint num_analogs;
 guint type_of_analogs;
 CHANNEL_CONFIG *analog_config;	 /*to be allocated as needed */
 guint num_discretes;
 CHANNEL_CONFIG *discrete_config; /*to be allocated as needed */
 guint type_of_frequency;
 }PMU_CONFIG;

static PMU_CONFIG pmu_config_info;

static gboolean c37_118_desegment = TRUE;
static xmlDocPtr doc=NULL;
static xmlNodePtr cur=NULL;

/* XML TAG Definitions */
#define PMU_CONFIG_ROOT "CONFIG"

static 
CHANNEL_CONFIG *init_channels(CHANNEL_CONFIG **inptr, guint num)
{
CHANNEL_CONFIG *ptr;
guint i;

  if (*inptr)		/*if there is something already there, get rid of it */
    {
    g_free(*inptr);
	*inptr = NULL;
	}

  ptr = (CHANNEL_CONFIG *)g_malloc(num * sizeof(CHANNEL_CONFIG));
  for (i=0;i<num;++i)		/*now need to fill in defaults */
    {
	ptr[i].name = "N/C";
	ptr[i].conversion_factor = 1.0;
	}
  
  *inptr = ptr;
  return ptr;
}

static
void decode_channel_info(xmlNodePtr cur, CHANNEL_CONFIG **config_ptr, guint num_channels)
{
CHANNEL_CONFIG *myconfig;
gchar *temp;
guint chan_num;
						   
  if(num_channels == 0) /*then have a config issue */
    return;

  if(myconfig = init_channels (config_ptr, num_channels))
    {
	while(cur)
	  {
	  if(!strcmpi(cur->name,"CHANINFO"))
  		{
  		temp = XmlStub.xmlGetProp(cur,"num");
  		if(temp)
  		  {
  		  chan_num = atoi(temp);
		  if (chan_num < num_channels)
		    {
			/*get the conversion factor */
			temp = XmlStub.xmlGetProp(cur,"conv");
			if(temp)
			  myconfig[chan_num].conversion_factor = atof(temp);
					 
			temp = XmlStub.xmlGetProp(cur,"name");
			if(temp)
			  myconfig[chan_num].name = temp;
			}
  		  } 
		}
	  cur = cur->next;
	  }
	}
}

static
gchar *parse_phasor_xml(xmlNodePtr cur1)
{
gchar *temp = NULL;
xmlNodePtr cur2;

  pmu_config_info.num_phasors = 0;
  temp = XmlStub.xmlGetProp(cur1,"num"); 
  if (temp)
    {
    pmu_config_info.num_phasors = atoi(temp);
	init_channels(&pmu_config_info.phasor_config,pmu_config_info.num_phasors);
    temp = XmlStub.xmlGetProp(cur1,"type");
    if (temp)
      {
      if(!strcmpi(temp,"Float"))
   	     pmu_config_info.type_of_phasors = DATA_DECODED_AS_FLOAT;
      else
	     pmu_config_info.type_of_phasors = DATA_DECODED_AS_INTEGER;
		 
	  if(cur1->children)			/*then somebody is trying to define channel names */
	    {
	    cur2 = cur1->children;
	    decode_channel_info(cur2, &pmu_config_info.phasor_config, pmu_config_info.num_phasors);
	    }
   	  }
    }
  return temp;
}

static
gchar *parse_analogs_xml(xmlNodePtr cur1)
{
gchar *temp = NULL;
xmlNodePtr cur2;

  pmu_config_info.num_analogs = 0;
  temp = XmlStub.xmlGetProp(cur1,"num"); 
  if (temp)
    {
    pmu_config_info.num_analogs = atoi(temp);
	init_channels(&pmu_config_info.analog_config,pmu_config_info.num_analogs);
    temp = XmlStub.xmlGetProp(cur1,"type");
    if (temp)
      {
      if(!strcmpi(temp,"Float"))
   	     pmu_config_info.type_of_analogs = DATA_DECODED_AS_FLOAT;
      else
   	     pmu_config_info.type_of_analogs = DATA_DECODED_AS_INTEGER;

	   if(cur1->children)			/*then somebody is trying to define channel names */
	     {
	     cur2 = cur1->children;
	     decode_channel_info(cur2, &pmu_config_info.analog_config, pmu_config_info.num_analogs);
	     }

   	   }
    }
  return temp;
}


static
gchar *parse_frequency_xml(xmlNodePtr cur1)
{
gchar *temp = NULL;

  temp = XmlStub.xmlGetProp(cur1,"type");	   /*get the type */

  if(!strcmpi(temp,"Float"))
     pmu_config_info.type_of_frequency = DATA_DECODED_AS_FLOAT;
  else
     pmu_config_info.type_of_frequency = DATA_DECODED_AS_INTEGER;
  return temp;
}

static
gchar *parse_discretes_xml(xmlNodePtr cur1)
{
gchar *temp = NULL;
xmlNodePtr cur2;

  temp = XmlStub.xmlGetProp(cur1,"num");	/*get the number of discretes*/
  if(temp)
    {
    pmu_config_info.num_discretes = atoi(temp);
	init_channels(&pmu_config_info.discrete_config,pmu_config_info.num_discretes);
	if(cur1->children)			/*then somebody is trying to define channel names */
	  {
	  cur2 = cur1->children;
	  decode_channel_info(cur2, &pmu_config_info.discrete_config, pmu_config_info.num_discretes);
	  }
	}
  return temp;
}

/* looks for possible out-of-sequence XML */
static
gchar *decide_to_parse_xml(xmlNodePtr cur)
{
gchar *temp=NULL;

  if(!strcmpi(cur->name,"PHASORS"))
    temp = parse_phasor_xml(cur);

  if(!strcmpi(cur->name,"FREQ"))
    temp = parse_frequency_xml(cur);

  if(!strcmpi(cur->name,"DISCRETES"))
    temp = parse_discretes_xml(cur);

  if(!strcmpi(cur->name,"Analogs"))
    temp = parse_analogs_xml(cur);


  return temp;
  
}

static
int xmlParseFilePush( const char *filename, int checkValid
#ifndef ETHEREAL_XML_DO_VALIDITY_CHECKING
                 _U_
#endif
) {
  FILE *f;
  xmlDocPtr doc=NULL;
#ifdef ETHEREAL_XML_DO_VALIDITY_CHECKING
  int valid=0;
#endif
  int res, size = 1024;
  char chars[1024];
  xmlParserCtxtPtr ctxt;

#ifdef ETHEREAL_XML_DO_VALIDITY_CHECKING
  /* I wonder what kind of a performance hit this is? */
  *XmlStub.xmlDoValidityCheckingDefaultValue = checkValid;
#endif

  f = fopen(filename, "r");
  if (f == NULL) {
	return NULL;
  }

  res = fread(chars, 1, 4, f);
  if (res > 0) {
	ctxt = XmlStub.xmlCreatePushParserCtxt(NULL, NULL,
										   chars, res, filename);
	while ((res = fread(chars, 1, size-1, f)) > 0) {
	  XmlStub.xmlParseChunk(ctxt, chars, res, 0);
	}
	XmlStub.xmlParseChunk(ctxt, chars, 0, 1);
	doc = ctxt->myDoc;
#ifdef ETHEREAL_XML_DO_VALIDITY_CHECKING
  valid=ctxt->valid;
#endif
	XmlStub.xmlFreeParserCtxt(ctxt);
  }
  fclose(f);

#ifdef ETHEREAL_XML_DO_VALIDITY_CHECKING
  /* Check valid */
  if (!valid) {
	return NULL;
  }
#endif

  return doc;
} /* xmlParseFilePush */

/*
 * This routine will call libxml to parse in the dictionary.
 */
static int
loadXMLDictionary(gchar *myfilename)
{
#if 0
  xmlDocPtr doc;
  xmlNodePtr cur;
#endif

  /*
   * build an XML tree from a the file;
   */
  XmlStub.xmlKeepBlanksDefault(0);                    /* Strip leading and trailing blanks */
  XmlStub.xmlSubstituteEntitiesDefault(1);            /* Substitute entities automagically */
  doc = xmlParseFilePush(myfilename, 1);  /* Parse the XML (do validity checks)*/

  /* Check for invalid xml */
  if (doc == NULL) 
  {
	return -1;
  }

  /*
   * Check the document is of the right kind
   */
  cur = XmlStub.xmlDocGetRootElement(doc);
  if (cur == NULL) {
	XmlStub.xmlFreeDoc(doc);
	return -1;
  }
  if (XmlStub.xmlStrcmp(cur->name, (const xmlChar *) PMU_CONFIG_ROOT)) {
	XmlStub.xmlFreeDoc(doc);
	return -1;
  }

  /*
   * Ok, the dictionary has been parsed by libxml, and is valid.
   * All we have to do now is read in our information.
   */
#if 0
  if (xmlDictionaryParse(cur->xmlChildrenNode) != 0) 
  {
	/* Error has already been printed */
	return -1;
  }

  /* Once we're done parsing, free up the xml memory */
  XmlStub.xmlFreeDoc(doc);
#endif

  return 0;

} /* loadXMLDictionary */

/*****************************************************************************/
static
int decode_float_value(proto_tree *root_tree, tvbuff_t *tvb, int *offset, gchar *string)
{
float value;
gchar mybuffer[100];


 value = tvb_get_ntohieee_float(tvb, *offset);
 sprintf(mybuffer,"%s%e",string, value);
 proto_tree_add_text(root_tree, tvb,*offset, 4, mybuffer);
 *offset +=4;

}

static
int decode_discrete_value(proto_tree *root_tree, tvbuff_t *tvb, int *offset, gchar *string)
{
gint value;
gchar mybuffer[100];
gint i;
gint len;


 value = tvb_get_ntohs(tvb, *offset);
 sprintf(mybuffer,"%s",string);
 for(i=0;i<16;++i)
 {
 len = strlen(mybuffer);
 if (value & 0x8000)
   sprintf(mybuffer+len,"1 ");
 else
   sprintf(mybuffer+len,"0 ");
 value = value <<1;
 }
 proto_tree_add_text(root_tree, tvb,*offset, 2, mybuffer);
 *offset +=2;
 return 0;
}

static
int decode_integer_value(proto_tree *root_tree, tvbuff_t *tvb, int *offset, gchar *string)
{
gint16 value;
gchar mybuffer[100];




 value = tvb_get_ntohs(tvb, *offset);
 sprintf(mybuffer,"%s%d",string,value);
 proto_tree_add_text(root_tree, tvb,*offset, 2, mybuffer);
 *offset +=2;
 return 0;
}

static
int decode_integer_value_to_float(proto_tree *root_tree, tvbuff_t *tvb, int *offset, gchar *string, double conv)
{
gint16 value;
gchar mybuffer[100];
double fvalue;

 value = tvb_get_ntohs(tvb, *offset);
 fvalue = value *conv;
 sprintf(mybuffer,"%s%e",string,fvalue);
 proto_tree_add_text(root_tree, tvb,*offset, 2, mybuffer);
 *offset +=2;
 return 0;
}


static gdouble
convert_fractionsOfSeconds (guint32 initialvalue)
{
  guint32 maskedvalue;
  int i;
  gdouble init = 0.5;
  gdouble outputVal = 0.0;

  maskedvalue = initialvalue & 0x00ffffff;		/*mask off the first byte (time quality) */
  for (i = 0; i<24;++i)
  {
	 if (maskedvalue & 0x00800000)
		 outputVal += init;

	 maskedvalue  = maskedvalue <<1;
	 init = init/2;
  }
  return(outputVal);
}

static
int decode_c37_framesize(proto_tree *root_tree, tvbuff_t *tvb, int *offset)
{
gchar string_buffer[100];

   saved_framesize = tvb_get_ntohs(tvb ,*offset);
   sprintf (string_buffer,"Framsize: %d",saved_framesize);
   proto_tree_add_text(root_tree, tvb, *offset, 2,
			string_buffer);
   *offset +=2;
   return(0);
}

static
int decode_c37_pmuid(proto_tree *root_tree, tvbuff_t *tvb, int *offset)
{


   myPMU= tvb_get_ntohs(tvb,*offset);

   proto_tree_add_uint_format (root_tree, hf_saved_PMU_ID, tvb, *offset, 2, myPMU,
						"PMU ID: (%u)",myPMU);     
#if 0
   hf_saved_PMU_ID = tvb_get_ntohs(tvb,*offset);
   sprintf (string_buffer,"PMU ID: %d",hf_saved_PMU_ID);
   proto_tree_add_text(root_tree, tvb, *offset, 2,
			string_buffer);
#endif
   *offset +=2;
   return(0);
}

static
int decode_c37_timestamp(proto_tree *root_tree, tvbuff_t *tvb, int *offset)
{
int offset1 = *offset;
gdouble fractions;
time_t  ts;
struct tm *t;
gboolean need_offset= FALSE;
guint32 intstorage;
gchar time_quality;
proto_tree *t2, *option_tree;
gchar string_buffer[100];

	ts = tvb_get_ntohl(tvb, offset1);

/* checking for compatibility or overflow in number of seconds */
	if (ts & 0x80000000)
	  {
	  need_offset = TRUE;
	  ts -= 0x7fe81780;
	  }


	offset1+=4;
	intstorage = tvb_get_ntohl(tvb, offset1);
	offset1 +=4;									   

#if 0
    /* the compatible version to SMTP Time Stamp */
 	fractions = convert_fractionsOfSeconds ( intstorage & 0xffffff00); 
    time_quality = intstorage & 0xff;
#else
 	fractions = convert_fractionsOfSeconds (intstorage & 0x00ffffff); 
    time_quality = (intstorage & 0xff000000) >>24;
#endif


	t = gmtime(&ts); 

    if (t)
	{
	  fractions += t->tm_sec;
	  if (!need_offset)
 	    sprintf(string_buffer,"Timestamp: %04d-%02d-%02d %02d:%02d.%f", 
					1900+t->tm_year,
					t->tm_mon+1, 
					t->tm_mday, 
					t->tm_hour, 
					t->tm_min,
                     fractions);
	  else
 	    sprintf(string_buffer,"Timestamp: %04d-%02d-%02d %02d:%02d.%f", 
					2038+t->tm_year,
					t->tm_mon+1, 
					t->tm_mday, 
					t->tm_hour, 
					t->tm_min,
                     fractions);


      t2 = proto_tree_add_text (root_tree, tvb, offset1-8,8,string_buffer); 
	  option_tree = proto_item_add_subtree(t2,ett_c37_timestamp);
	  }
	 else
	  {
      t2 = proto_tree_add_text (root_tree, tvb, offset1-8,8,"Timestamp: Unable to Covert for Display"); 
	  option_tree = proto_item_add_subtree(t2,ett_c37_timestamp);
      proto_tree_add_text (option_tree, tvb, offset1-8,4,"Seconds of Century"); 
#if 0
      proto_tree_add_text (option_tree, tvb, offset1-3,3,"Fractions of Second"); 
#else
      proto_tree_add_text (option_tree, tvb, offset1-4,4,"Fractions of Second"); 
#endif
 	  }

#if 0
	  if (time_quality & 0x40)
	    proto_tree_add_text(option_tree,tvb,offset1-1,1,"Leap Second Direction: Subtract");
	  else
	    proto_tree_add_text(option_tree,tvb,offset1-1,1,"Leap Second Direction: Add");

	  if (time_quality & 0x20)
	    proto_tree_add_text(option_tree,tvb,offset1-1,1,"Leap Second Occurred");

	  if (time_quality & 0x10)
	    proto_tree_add_text(option_tree,tvb,offset1-1,1,"Leap Second Pending");

	  proto_tree_add_text(option_tree,tvb,offset1-1,1,quality_strings[time_quality & 0x0f]);
#else
	  if (time_quality & 0x40)
	    proto_tree_add_text(option_tree,tvb,offset1-4,1,"Leap Second Direction: Subtract");
	  else
	    proto_tree_add_text(option_tree,tvb,offset1-4,1,"Leap Second Direction: Add");

	  if (time_quality & 0x20)
	    proto_tree_add_text(option_tree,tvb,offset1-4,1,"Leap Second Occurred");

	  if (time_quality & 0x10)
	    proto_tree_add_text(option_tree,tvb,offset1-4,1,"Leap Second Pending");

	  proto_tree_add_text(option_tree,tvb,offset1-4,1,quality_strings[time_quality & 0x0f]);
#endif

	 *offset = offset1;
	 return(0);
}

static
void decode_c37_stat(proto_tree *root_tree, tvbuff_t *tvb, int *offset)
{
int offset1 = *offset;
guint16 stat;
guchar intstorage;
proto_tree *t2, *option_tree, *trigger_tree, *option1;
gchar string_buffer[100];

	stat = tvb_get_ntohs(tvb, offset1);

    if (stat & 0x8000)
	  sprintf(string_buffer, "Status: Invalid or Test Mode");
	else
	  sprintf(string_buffer, "Status: Valid");

    t2 = proto_tree_add_text (root_tree, tvb, offset1,2,string_buffer); 
	option_tree = proto_item_add_subtree(t2,ett_c37_status);

    /* now can decode the rest of the status */
	if (stat & 0x4000)
	    proto_tree_add_text(option_tree,tvb,offset1,1,"PMU error including configuration error");

	if (stat & 0x2000)
	    proto_tree_add_text(option_tree,tvb,offset1,1,"PMU not in Sync");

	if (stat & 0x1000)
	    proto_tree_add_text(option_tree,tvb,offset1,1,"Data Sorting by arrival");
	else
	    proto_tree_add_text(option_tree, tvb, offset1,1,"Data Sorting by timestamp");

	if (stat & 0x0800)
	  {
	  trigger_tree = proto_tree_add_text(option_tree,tvb,offset1,1,"PMU trigger occurred");
	  option1 = proto_item_add_subtree(trigger_tree,ett_c37_timestamp);

	  intstorage = stat & 0x000f;
	  switch (intstorage) {
	    case 7:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Digital Trigger");
		  break;

	    case 6:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Reserved");
		  break;

	    case 5:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"df/dt High");
		  break;

	    case 4:														 
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Frequency high/low");
		  break;

	    case 3:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Phase Angle Difference");
		  break;

	    case 2:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Magnitude High");
		  break;

	    case 1:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Magnitude Low");
		  break;

	    case 0:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"Manual");
		  break;

		default:
		  proto_tree_add_text(option1,tvb,offset1+1,1,"User defined reason: %d",intstorage);
	    }
	  }		 

	if (stat & 0x0400)
	    proto_tree_add_text(option_tree,tvb,offset1,1,"Configuration changed");

	intstorage = (stat & 0x01c00) >>6;
	proto_tree_add_text(option_tree,tvb,offset1,2,"Security Parameter: %d",intstorage);

	intstorage = (stat & 0x0030) >> 4;
	switch (intstorage)
	  {
	  case 0:
	    proto_tree_add_text(option_tree,tvb,offset1+1,1,"Sync locked, best quality");
		break;

	  case 1:
	    proto_tree_add_text(option_tree,tvb,offset1+1,1,"Unlocked for 10 sec");
		break;

	  case 2:
	    proto_tree_add_text(option_tree,tvb,offset1+1,1,"Unlocked for 100 sec");
		break;

	  case 3:
	    proto_tree_add_text(option_tree,tvb,offset1+1,1,"Unlocked for 1000 sec");
		break;
	  }

	 *offset = offset1 +2;
}

static
gboolean iterate_sub_process(proto_tree *root_tree, tvbuff_t *tvb, int *offset,
		 guint16 *length_of_data,  gchar *mytype_prefix)
{
gboolean return_value=TRUE;
guint i;
gchar mybuffer[100];
int offset1 = *offset;
guint num_iteration = pmu_config_info.num_analogs;
guint type_of_conversion = pmu_config_info.type_of_analogs;
CHANNEL_CONFIG *myptr=pmu_config_info.analog_config;


  for (i=0;i<num_iteration;++i)
  {
     offset1 = *offset;
	 sprintf(mybuffer,"%s_%d (%s): ",mytype_prefix,i,myptr[i].name);
	 if (type_of_conversion == DATA_DECODED_AS_FLOAT)
	   decode_float_value(root_tree, tvb, offset, mybuffer);
	 else
	   decode_integer_value_to_float(root_tree, tvb, offset, mybuffer,myptr[i].conversion_factor);
  }
  return(return_value);	
}

static
gboolean iterate_sub_phasors(proto_tree *root_tree, tvbuff_t *tvb, int *offset,	
		 		 gchar *mytype_prefix)
{
gboolean return_value=TRUE;
guint i;
gchar mybuffer[100];
int offset1 = *offset;
proto_tree *t2,*t3;
guint num_iteration = pmu_config_info.num_phasors;
guint type_of_conversion = pmu_config_info.type_of_phasors;
CHANNEL_CONFIG *myptr=pmu_config_info.phasor_config;

  for (i=0;i<num_iteration;++i)
  {
     offset1 = *offset;
	 sprintf(mybuffer,"PHASOR_%d (%s):  ",i,myptr[i].name);
	 if (type_of_conversion == DATA_DECODED_AS_FLOAT)
       t2 =  proto_tree_add_text (root_tree, tvb, offset1,8,mybuffer); 
	 else
       t2 =  proto_tree_add_text (root_tree, tvb, offset1,4,mybuffer); 
	 
     t3 = proto_item_add_subtree(t2,ett_c37_chan_info);

     if (type_of_conversion == DATA_DECODED_AS_FLOAT)
	   {
       decode_float_value(t3, tvb, offset, "a:");
	   decode_float_value(t3, tvb, offset, "b:");
	   }
     else
	   {
       decode_integer_value_to_float(t3, tvb, offset, "a:", myptr[i].conversion_factor);
	   decode_integer_value_to_float(t3, tvb, offset, "b:",myptr[i].conversion_factor);
	   }
  }
  return(return_value);	
}

static
gboolean decode_frequency(proto_tree *root_tree, tvbuff_t *tvb, int *offset,
		 guint16 *length_of_data, guint num_iteration, guint type_of_conversion,
		 gchar *mytype_prefix)
{
gboolean return_value=TRUE;
guint i;
gchar mybuffer[100];
int offset1 = *offset;
proto_tree *t2,*t3;


     offset1 = *offset;
	 sprintf(mybuffer,"%s  ",mytype_prefix,i+1);
	 if (pmu_config_info.type_of_frequency == DATA_DECODED_AS_FLOAT)
       t2 =  proto_tree_add_text (root_tree, tvb, offset1,4,mybuffer); 
	 else
       t2 =  proto_tree_add_text (root_tree, tvb, offset1,2,mybuffer); 

     t3 = proto_item_add_subtree(t2,ett_c37_chan_info);
	 if (pmu_config_info.type_of_frequency == DATA_DECODED_AS_FLOAT)
	   decode_float_value(t3, tvb, offset, "");
	 else
	   decode_integer_value(t3, tvb, offset, "");

  return(return_value);	
}

static
void iterate_through_data(proto_tree *root_tree, tvbuff_t *tvb, int *offset, 
						  guint16 *length_of_data)
													
{
gboolean ret_value = TRUE;
guint i;
gchar mybuffer[100];
int offset1 = *offset;
CHANNEL_CONFIG *myptr=pmu_config_info.discrete_config;


/* phasors */
  ret_value =  iterate_sub_phasors(root_tree, tvb, offset, "PHASOR");

/* Frequency */
  ret_value =  decode_frequency(root_tree, tvb, offset, length_of_data,
						1, frequency_decode_type,"FREQ");

/* DFREQ */
  ret_value =  decode_frequency(root_tree, tvb, offset, length_of_data,
						1, frequency_decode_type,"DFREQ");

/*analogs */
  ret_value =  iterate_sub_process(root_tree, tvb, offset, length_of_data,"ANALOG");

/*discretes */
  for (i=0;i<pmu_config_info.num_discretes;++i)
    {
    offset1 = *offset;

    sprintf(mybuffer,"%s_%d (%s):  ","DISCRETE",i+1,myptr[i].name);  
/*    sprintf(mybuffer,"%s_%d :  ","DISCRETE",i+1);   */
 
    decode_discrete_value(root_tree, tvb, offset, mybuffer);
    }
}

static
void decode_c37_data(proto_tree *root_tree, tvbuff_t *tvb, int *offset,guint16 length_of_data,
					guint decode_method)
{
int offset1 = *offset;
proto_tree *t2,*t3;
guint16 remaining_length = length_of_data;

/* mark all of the data just to be safe */
   t2 =  proto_tree_add_text (root_tree, tvb, offset1,length_of_data,"Data"); 

  switch (decode_method)
	 {
	case DATA_DECODE_SINGLE:
	case DATA_DECODE_FILE_BASED:
	  t3 = proto_item_add_subtree(t2,ett_c37_subopt3);
	  iterate_through_data(t3,tvb,offset,&remaining_length);
	break;

    default:
	  break;
	}

/*once finished, need to go to the end */
   *offset = offset1 + length_of_data;

}

static
void decode_c37_data_PMU(proto_tree *root_tree, tvbuff_t *tvb, int *offset,guint16 length_of_data,
						guint decode_method)
{
int offset1 = *offset;
guint16 remaining_size_of_data;
guint internal_decode_method = decode_method;
char *temp;
xmlNodePtr cur1 = NULL;
xmlNodePtr cur2 = NULL;




/* parse the appropriate informaiton */
  if(internal_decode_method == DATA_DECODE_FILE_BASED)
   {
	 if(!strcmpi(cur->name,"PMU"))
	   cur1=cur->children;

     /*start of phasor decode */
	 while(cur1)
	   {
	   temp = decide_to_parse_xml(cur1);
	   if(temp)
	      cur1 = cur1->next;
	   else
	     {
		 internal_decode_method = DATA_DECODE_NONE;
	     cur1 = NULL;
		 XmlStub.xmlFreeDoc(doc);
		 doc = NULL;
		 }
	   }
  }	/*end of XML decode */

if (internal_decode_method ==  DATA_DECODE_SINGLE)
  {
   pmu_config_info.num_phasors = glb_num_phasors;
   pmu_config_info.num_analogs = glb_num_analogs;
   pmu_config_info.num_discretes = glb_num_discretes;
   pmu_config_info.type_of_analogs = analog_decode_type;
   pmu_config_info.type_of_phasors = phasor_decode_type;
   pmu_config_info.type_of_frequency =  frequency_decode_type;
   init_channels ( &pmu_config_info.phasor_config, glb_num_phasors);
   init_channels ( &pmu_config_info.analog_config, glb_num_analogs);
   init_channels ( &pmu_config_info.discrete_config, glb_num_discretes);
   doc=NULL;
  }

/* get the status */
  decode_c37_stat(root_tree,tvb,offset);

/* actually get the data, but can't really decode it since it is anonymous*/
  remaining_size_of_data = 	length_of_data - (*offset - offset1);  /*the additional -2 is for CRC*/
  decode_c37_data(root_tree, tvb, offset,remaining_size_of_data, internal_decode_method);
  
  if (doc!=NULL)
    {
    XmlStub.xmlFreeDoc(doc);
	doc = NULL;
    } 
}


/* function performs the decode of command messages */
static
void decode_c37_command (packet_info *pinfo, proto_tree *c37_tree, tvbuff_t *tvb, int start_offset)
{
gchar string_buffer[100];
int offset = start_offset;
proto_tree *ti;
guint16 cmd;

   ti = proto_item_add_subtree(c37_tree,ett_c37_subopt1);

/* get the framesize  */
  decode_c37_framesize(ti, tvb, &offset);

/*get the PMU ID */
  decode_c37_pmuid(ti, tvb, &offset);

/*get the timestamp */
 decode_c37_timestamp(ti, tvb, &offset);

/* actually get the command */
    cmd = tvb_get_ntohs(tvb,offset);
	switch (cmd){
	 case 2:
	   sprintf(string_buffer,"%s",ENABLE_REQ);
	   break;

	 case 1:
	   sprintf(string_buffer,"%s",DISABLE_REQ);
	   break;
			   
     case 3:
	   sprintf(string_buffer,"%s", HEADER_REQ);
	   break;
								
     case 4:
	   sprintf(string_buffer,"%s", CONFIG1_REQ);
	   break;

     case 5:
	   sprintf(string_buffer,"%s", CONFIG2_REQ);
	   break;

     case 8:
	   sprintf(string_buffer,"%s", EXTENDED_REQ);
	   break;

     default:
	   sprintf(string_buffer,"%s", UNKNOWN_REQ);
	   break;
    }
/*use this one so we can do a display filter */
	proto_tree_add_string(ti, hf_saved_pdu_type, tvb,offset, 2, string_buffer);
#if 0
     proto_tree_add_text (ti, tvb, offset, 2,string_buffer);
#endif
     offset +=2; 
  
	/*point to the CRC */
	proto_tree_add_text (ti, tvb, offset,2, "CRC");
	offset +=2;			 
}

/* function performs the decode of command messages */
static
void decode_c37_data_message (packet_info *pinfo, proto_tree *c37_tree, tvbuff_t *tvb, int start_offset)
{

int offset = start_offset;
proto_tree *ti;
guint16 remaining_size_of_data;
gchar myfilename[200];
guint actual_decode_method = gbl_type; 


   ti = proto_item_add_subtree(c37_tree,ett_c37_subopt1);

/* get the framesize  */
  decode_c37_framesize(ti, tvb, &offset);

/*get the PMU ID */
  decode_c37_pmuid(ti, tvb, &offset);

/*get the timestamp */
 decode_c37_timestamp(ti, tvb, &offset);

  remaining_size_of_data = 	saved_framesize - (offset - start_offset) -4;  /*the additional -2 is for CRC*/

  remaining_size_of_data = 	saved_framesize - (offset - start_offset) -4;  /*the additional -2 is for CRC*/

  if(actual_decode_method == DATA_DECODE_FILE_BASED)
    {
    /*generate the name of the file including directory */
    sprintf(myfilename,"%s\\PMU_%05u\.xml",gbl_diameterDictionary,myPMU);
	loadLibXML();
    if (loadXMLDictionary(myfilename)!=0)
       actual_decode_method = DATA_DECODE_NONE;
	else
	  {
	  cur = cur->children;
	  }
    }

  while(remaining_size_of_data >0)
    {
    decode_c37_data_PMU(ti, tvb, &offset,remaining_size_of_data, actual_decode_method); 
    remaining_size_of_data = 	saved_framesize - (offset - start_offset) -4;  /*the additional -2 is for CRC*/
	}

	/*point to the CRC */
	proto_tree_add_text (ti, tvb, offset,2, "CRC");
	offset +=2;			 

				 
}

/* function performs the decode of command messages */
static
void decode_c37_header_message (packet_info *pinfo, proto_tree *c37_tree, tvbuff_t *tvb, int start_offset)
{

int offset = start_offset;
proto_tree *ti;
guint16 remaining_size_of_data;
guchar *myHeaderInfo;
proto_tree *t2,*t3;




  ti = proto_item_add_subtree(c37_tree,ett_c37_subopt1);

/* get the framesize  */
  decode_c37_framesize(ti, tvb, &offset);

/*get the PMU ID */
  decode_c37_pmuid(ti, tvb, &offset);

/*get the timestamp */
 decode_c37_timestamp(ti, tvb, &offset);


/* actually get the header */
  remaining_size_of_data = 	saved_framesize - (offset - start_offset) -4;  /*the additional -2 is for CRC*/
  t2 = proto_tree_add_text(ti, tvb,offset,remaining_size_of_data,"Header String");
  t3 = proto_item_add_subtree(t2,ett_c37_subopt2);   
   myHeaderInfo =  tvb_get_string(tvb,  offset, remaining_size_of_data);
   t2 = proto_tree_add_text(t3, tvb, offset, remaining_size_of_data,myHeaderInfo);
   offset += remaining_size_of_data;
   g_free(myHeaderInfo);	 

	/*point to the CRC */
	proto_tree_add_text (ti, tvb, offset,2, "CRC");
	offset +=2;			 

				 
}


static
void decode_c37_config_message (packet_info *pinfo, proto_tree *c37_tree, tvbuff_t *tvb, int start_offset)
{
gchar string_buffer[100];
int offset = start_offset;
proto_tree *ti;
guint32 timebase;
guint16 numPMUs;
guint16 i,j,z;			/*used to iterate through the packet  based upon numPMUs */
guint8 *StationName;
proto_tree *t2, *pmu_tree, *internal_pmu_tree, *internals, *format_branch,*f1;
proto_tree *chan_branch, *chan_info, *indv_chan_info, *units;
guint16 format_info;
guint16 phnmr;
guint16 annmr;
guint16 dgnmr;
guint16 numTotChans;
guint16 chan_info_org_offset;
TRACK_INFO *myinfo;
TRACK_INFO *myptr;
guint32 temp,temp1;


   ti = proto_item_add_subtree(c37_tree,ett_c37_subopt1);

/* get the framesize  */
  decode_c37_framesize(ti, tvb, &offset);

/*get the PMU ID */
  decode_c37_pmuid(ti, tvb, &offset);

/*get the timestamp */
 decode_c37_timestamp(ti, tvb, &offset);
					   
/* get the time base */	
   timebase = tvb_get_ntohl(tvb,offset);
   proto_tree_add_text(ti, tvb, offset, 2,"Timebase: %d",timebase);	 
   offset +=4;

/* get the number of PMUs */
   numPMUs = tvb_get_ntohs(tvb,offset);
   t2 = proto_tree_add_text(ti, tvb, offset, 2,"Number of PMUs: %d",numPMUs);	 
   offset +=2;

   pmu_tree = proto_item_add_subtree(t2,ett_c37_subopt2);
    
  for (i= 0; i<numPMUs; i++)
    {  
    /* get the station name */


    StationName =  tvb_get_string(tvb,  offset, 16);
	internal_pmu_tree = proto_tree_add_string_format(pmu_tree, hf_saved_config_station_name, tvb,offset, 16,StationName,"Station Name: %s", StationName);
#if 0
    internal_pmu_tree= proto_tree_add_text(pmu_tree, tvb, offset, 16,"Station Name: %s",StationName);	 
#endif
    offset +=16;
    g_free(StationName);

  	internals =proto_item_add_subtree(internal_pmu_tree,ett_c37_subopt3);
	/*get the PMU ID */
	decode_c37_pmuid (internals, tvb, &offset);

	/*get the format information */
	format_info = (tvb_get_ntohs(tvb, offset)) & 0x000F;
	format_branch = proto_tree_add_text(internals,tvb,offset,2,"Format Information");
	f1 = proto_item_add_subtree(format_branch,ett_c37_status);
	if (format_info & 0x8)
	   proto_tree_add_text(f1,tvb,offset,2,"FREQ/DFREQ in Floating Point");
	else
	   proto_tree_add_text(f1,tvb,offset,2,"FREQ/DFREQ in 16-bit Integer");

	if (format_info & 0x4)
	   proto_tree_add_text(f1,tvb,offset,2,"Analogs in Floating Point");
	else
	   proto_tree_add_text(f1,tvb,offset,2,"Analogs in 16-bit Integer");

	if (format_info & 0x2)
	   proto_tree_add_text(f1,tvb,offset,2,"Phasors in Floating Point");
	else
	   proto_tree_add_text(f1,tvb,offset,2,"Phasors in 16-bit Integer");

	if (format_info & 0x1)
	   proto_tree_add_text(f1,tvb,offset,2,"Phasor = magnitude/angle");
	else
	   proto_tree_add_text(f1,tvb,offset,2,"Phasor = real/imaginary (rectangular)");

	offset +=2;

    /*get the number of phasors */
	phnmr = tvb_get_ntohs(tvb, offset);
	proto_tree_add_text(internals,tvb,offset,2,"Number of Phasors: %d",phnmr);
	offset +=2;

	/*get the number of analogs */
	annmr = tvb_get_ntohs(tvb, offset);
	proto_tree_add_text(internals,tvb,offset,2,"Number of Analogs: %d",annmr);
	offset +=2;

	dgnmr = tvb_get_ntohs(tvb, offset);
	proto_tree_add_text(internals,tvb,offset,2,"Number of Digitals: %d",dgnmr);
	offset +=2;

	numTotChans = phnmr + annmr +(dgnmr * 16);
	myinfo = (TRACK_INFO *)g_malloc(numTotChans * sizeof(TRACK_INFO));
	myptr = myinfo;
    chan_info_org_offset = offset;		/*save this value for future use */

	/*get all the channel names*/
	for(j=0;j<numTotChans;j++)
	  {
 	   /* then we have a phasor channel */
	   myptr->channame_offset =offset;
	   myptr->channame =tvb_get_string(tvb,  offset, 16);
	   ++myptr;
	   offset +=16;  
	  }

    myptr = myinfo;
	for(j=0;j<phnmr;j++)
	  {
	  myptr->unit_information = tvb_get_ntohl(tvb, offset);
 	  myptr->unit_offset = offset;
	  offset +=4;
	  myptr->type_of_channel = PHASOR_CHANNEL;
	  ++myptr;
	  }

	for(j=0;j<annmr;j++)
	  {
	  myptr->unit_information = tvb_get_ntohl(tvb, offset);
 	  myptr->unit_offset = offset;
	  offset +=4;
	  myptr->type_of_channel = ANALOG_CHANNEL;
	  ++myptr;
	  }

	for(j=0;j<dgnmr;j++)
	  {
	  temp = tvb_get_ntohs(tvb,offset);		/*this should get the normal state */
	  offset +=2;
	  temp1 = tvb_get_ntohs(tvb,offset);	/*valid or invalid channel mask    */
	  offset +=2;
	  for (z=0; z<16;++z)
	    {
 	    myptr->type_of_channel = DIGITAL_CHANNEL;

		myptr->unit_offset = offset-4;
		myptr->normal_state = temp & 0x8000;
		myptr->valid_channel = temp1 & 0x8000;
		temp = temp <<1;
		temp1 = temp1<<1;
	    ++myptr;
	  	}
	  }

    chan_branch = proto_tree_add_text(internals,tvb,chan_info_org_offset,offset-chan_info_org_offset-1,"Channel Configuration Information");
    myptr = myinfo;
    for(j=0;j<numTotChans;j++)
	  {
	  /*now present the information to the display */
	  chan_info = proto_item_add_subtree(chan_branch,ett_c37_chan_info); 
	  if (myptr->type_of_channel ==PHASOR_CHANNEL)
	    {
		sprintf(string_buffer,"Channel (%d): Phasor Channel",j);
		}

	  if (myptr->type_of_channel ==ANALOG_CHANNEL)
	    {
		sprintf(string_buffer,"Channel (%d): Analog Channel",j);
		}

	  if (myptr->type_of_channel ==DIGITAL_CHANNEL)
	    {
		sprintf(string_buffer,"Channel (%d): Digital Channel",j);
		}
#if 0	  
       indv_chan_info = proto_tree_add_text(chan_info,tvb,myptr->channame_offset,16,
      					"%s",myptr->channame);
#endif
       indv_chan_info = proto_tree_add_string(chan_info,hf_saved_config_channel_name,tvb,myptr->channame_offset,
       					16,
      					myptr->channame);


	   units = proto_item_add_subtree(indv_chan_info,ett_c37_chan_info);

       if(myptr->type_of_channel == PHASOR_CHANNEL)
	     {
 		  proto_tree_add_text(units,tvb,myptr->channame_offset,16,"Phasor Channel");

		 if (myptr->unit_information & 0xff000000)
		   	proto_tree_add_text(units,tvb,myptr->unit_offset,1,"Current phasor");
		 else
		   	proto_tree_add_text(units,tvb,myptr->unit_offset,1,"Voltage phasor");

         proto_tree_add_text(units,tvb,(myptr->unit_offset)+1,3,"Scale Factor: %ld",
					myptr->unit_information & 0x00ffffff);
         }

	   if(myptr->type_of_channel == ANALOG_CHANNEL)
	     {
 		  proto_tree_add_text(units,tvb,myptr->channame_offset,16,"Analog Channel");

		  temp = (myptr->unit_information & 0xff000000)	>>24;
		  switch(temp)
		    {
			case 0:
				proto_tree_add_text(units,tvb,myptr->unit_offset,1,"single point-on-wave");
				break;

			case 1:
				proto_tree_add_text(units,tvb,myptr->unit_offset,1,"rms");
				break;

			case 2:
				proto_tree_add_text(units,tvb,myptr->unit_offset,1,"peak");
				break;

			default:
				proto_tree_add_text(units,tvb,myptr->unit_offset,1,"Unknown value (%d)",
						temp);
				break;
			}
		  temp = (myptr->unit_information) & 0x00ffffff;
		  proto_tree_add_text(units,tvb,myptr->unit_offset+1,3,"Scale factor: %ld",
						temp);

		 }

	   if(myptr->type_of_channel == DIGITAL_CHANNEL)
	     {
 		  proto_tree_add_text(units,tvb,myptr->channame_offset,16,"Digital Channel");
		 if (myptr->normal_state)
		   proto_tree_add_text(units,tvb,myptr->unit_offset,2,"Normal State: True");	 
		 else
		   proto_tree_add_text(units,tvb,myptr->unit_offset,2,"Normal State: False");   

		 if (myptr->valid_channel)
		   proto_tree_add_text(units,tvb,(myptr->unit_offset)+2,2,"Valid Channel");	  
		 else
		   proto_tree_add_text(units,tvb,(myptr->unit_offset)+2,2,"Channel not in use");
		 }

	  ++myptr;	 
	  }		/*end of j loop */

	  /*need to free up what has been allocated */
    myptr = myinfo;
    for(j=0;j<numTotChans;j++)
	  {
	  g_free(myptr->channame);
	  ++myptr;
	  }
	g_free(myinfo);



	/*now get the frequency information */
	j = (tvb_get_ntohs(tvb ,offset)) & 1;
    if (j)
	  proto_tree_add_text(internals,tvb,offset,2,"50 Hertz Frequency");	 
	else
	  proto_tree_add_text(internals,tvb,offset,2,"60 Hertz Frequency");
	offset +=2;


	/*now get the configuration count */
	j = tvb_get_ntohs(tvb ,offset);
     proto_tree_add_text(internals,tvb,offset,2,"Configuration Revision Count: %u",j);	 
	offset +=2;
    }

	/*now get the data rate */
	j = tvb_get_ntohs(tvb ,offset);
    if (j & 0x8000)
	  proto_tree_add_text(ti,tvb,offset,2,"DataRate: 1/%d seconds/frame",(j&0x7fff));	 
	else
	  proto_tree_add_text(ti,tvb,offset,2,"DataRate: %d frames/second",(j&0x7fff));
	offset +=2;

	/*point to the CRC */
	proto_tree_add_text (ti, tvb, offset,2, "CRC");
	offset +=2;			 

}
						   


static int
c37_command(packet_info *pinfo, proto_tree *c37_tree, tvbuff_t *tvb, int start_offset)
{
  int offset = start_offset;
  guchar optcode;
  guchar command;
  proto_tree *ti;

  offset += 1;	/* skip sync */
  optcode = tvb_get_guint8(tvb, offset);
  command = (optcode & 0x70)>>4;
  offset++;
  switch(command) {

  case C37_DATA_PACKET:
	ti = proto_tree_add_string(c37_tree, hf_saved_pdu_type, tvb, start_offset, 2, DATA_MSG);
    decode_c37_data_message(pinfo, ti, tvb, offset);
    break;

  case C37_HEADER_PACKET:
	ti = proto_tree_add_string(c37_tree, hf_saved_pdu_type,tvb, start_offset, 2, HEADER_RESP);
    decode_c37_header_message(pinfo, ti, tvb, offset);
    break;

  case C37_CONFIG1_PACKET:
	ti = proto_tree_add_string(c37_tree, hf_saved_pdu_type, tvb,start_offset, 2, CONFIG1_RESP);
    decode_c37_config_message(pinfo, ti, tvb, offset);
    break;

  case C37_CONFIG2_PACKET:
	ti = proto_tree_add_string(c37_tree, hf_saved_pdu_type, tvb,start_offset, 2, CONFIG2_RESP);
    decode_c37_config_message(pinfo, ti, tvb, offset);
    break;

  case C37_COMMAND_PACKET:
    ti= proto_tree_add_text(c37_tree, tvb, start_offset, 2,
			"Command Message");
	decode_c37_command (pinfo, ti, tvb, offset);
    break;

  default:
	ti = proto_tree_add_string(c37_tree, hf_saved_pdu_type, tvb,start_offset, 2, UNKNOWN_RESP);
    break;
  }

/* next need to get the framesize */

  return offset;
}


static void
dissect_c37_118_common(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
        proto_tree      *c37_tree, *ti;


	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "c37-118");

	if (check_col(pinfo->cinfo, COL_INFO))
		col_add_fstr(pinfo->cinfo, COL_INFO, "c37-118");

	if (tree) {
	  gint offset = 0;
	  guint len;
	  gint sync_offset;

	  ti = proto_tree_add_item(tree, proto_c37_118, tvb, offset, -1, FALSE);
	  c37_tree = proto_item_add_subtree(ti, ett_c37);

	  /*
	   * Scan through the buffer looking for an IAC byte.
	   */
/*	  while ((len = tvb_length_remaining(tvb, offset)) > 0) */
		if ((len = tvb_length_remaining(tvb,offset))>0)
		{	
	    sync_offset = tvb_find_guint8(tvb, offset, len, C37_SYNC);
	    if (sync_offset != -1) {
	      /*
	       * We found a SYNC byte.
	       * If there's any data before it, add that data to the
	       * tree, a line at a time.  */
	      offset = c37_command(pinfo, c37_tree, tvb, sync_offset);
	    }
	  }
	}
}

static guint get_c37_pdu_len(tvbuff_t *tvb, int offset)
{
  guint16 plen;
  plen = tvb_get_ntohs(tvb, offset +2);
  return plen; 
}


static int
dissect_c37_118_tcp(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  if(tvb_bytes_exist(tvb,0,1) && (tvb_get_guint8(tvb,0) !=0xAA))
  {
  return 0; 	/*not a C37.118 packet */
  }

  tcp_dissect_pdus(tvb,pinfo,tree,c37_118_desegment, 4, get_c37_pdu_len, dissect_c37_118_common);
    return tvb_length(tvb);
}


void
proto_register_c37_118(void)
{
	gchar *default_diameterDictionary;

	static hf_register_info hf[] =
	{
		/* hfields for packet-ieeec37-118.c */	
		{ &hf_saved_PMU_ID,
			{ "PMUID", "c37-118.PMUID", FT_UINT16, BASE_DEC, NULL, 0x0,"", HFILL}
		},
		{ &hf_saved_pdu_type,
			{ "MessageType", "c37-118.MessageType", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},
		{ &hf_saved_config_station_name,
			{ "StationName", "c37-118.StationName", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},
		{ &hf_saved_config_channel_name,
			{ "ChannelName", "c37-118.ChannelName", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},

	};

 
	static gint *ett[] = {
		&ett_c37,
		&ett_c37_subopt1,
		&ett_c37_subopt2,
		&ett_c37_subopt3,
		&ett_c37_timestamp,
		&ett_c37_status,
		&ett_c37_chan_info,
		&ett_c37_indv_chan_info,
	};
	module_t   *diameter_module;

        proto_c37_118 = proto_register_protocol("IEEE c37-118", "c37-118", "c37-118");
        proto_register_field_array(proto_c37_118, hf, array_length(hf));
		proto_register_subtree_array(ett, array_length(ett));


	/* Register a configuration option for port */
	diameter_module = prefs_register_protocol(proto_c37_118,
											  proto_reg_handoff_c37_118);

    prefs_register_enum_preference(diameter_module, "type_of_conversion",
     "Type of Data Conversion", "Select None for no conversion", (gint *)&gbl_type, options, FALSE);

    prefs_register_enum_preference(diameter_module, "phasor_type_of_conversion",
     "Phasor Data Type", "Select INTEGER or FLOAT", (gint *)&phasor_decode_type, 
     	type_options, TRUE);

    prefs_register_enum_preference(diameter_module, "analog_type_of_conversion",
     "Analog Data Type", "Select INTEGER or FLOAT", (gint *)&analog_decode_type, 
     	type_options, TRUE);

    prefs_register_enum_preference(diameter_module, "frequency_type_of_conversion",
     "Frequency Data Type", "Select INTEGER or FLOAT", (gint *)&frequency_decode_type, 
     	type_options, TRUE);

	prefs_register_uint_preference(diameter_module, "num.phasors",
								   "Number of phasors",
								   "Set the number of phasors for Based Upon Preference decode ",
								   10,
								   &glb_num_phasors);

	prefs_register_uint_preference(diameter_module, "num.analogs",
								   "Number of analogs",
								   "Set the number of analogs for Based Upon Preference decode ",
								   10,
								   &glb_num_analogs);

	prefs_register_uint_preference(diameter_module, "num.discretes",
								   "Number of discretes",
								   "Set the number of discretes for Based Upon Preference decode ",
								   10,				   
								   &glb_num_discretes);


	/*
	 * Build our default dictionary filename
	 */
	default_diameterDictionary = get_datafile_path(DICT_FN);

	/*
	 * Now register the dictionary filename as a preference,
	 * so it can be changed.
	 */
	gbl_diameterDictionary = default_diameterDictionary;
	prefs_register_string_preference(diameter_module, "pmudirectory.name",
									 "PMU Definition Directory",
									 "Set the directory of where to find PMU Configuration Files",
									 &gbl_diameterDictionary);

	/*
	 * We don't need the default dictionary, so free it (a copy was made
	 * of it in "gbl_diameterDictionary" by
	 * "prefs_register_string_preference()").
	 */
	g_free(default_diameterDictionary);


}

void
proto_reg_handoff_c37_118(void)
{
	dissector_handle_t c37_handle_tcp;
	dissector_handle_t c37_handle_udp;

	c37_handle_udp = create_dissector_handle(dissect_c37_118_common, proto_c37_118);
    c37_handle_tcp = create_dissector_handle(dissect_c37_118_tcp, proto_c37_118);

	dissector_add("tcp.port", TCP_PORT_C37_118, c37_handle_tcp);
	dissector_add("udp.port", UDP_PORT_C37_118, c37_handle_tcp);

}

