/* packet-iso9506.c
*
* Routine to dissect ISO/IEC 9506 MMS Protocol packets
*
* $Id: packet-iso9506.c,v 1.0 2004/012/12  Exp $
*
* Herbert Falk <herb@sisconet.com>
*
* Ethereal - Network traffic analyzer
* By Gerald Combs <gerald@ethereal.com>
* Copyright 1998 Gerald Combs
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


/* Notes to finish up HSF

need to fix the FileOpen Response to properly decode Gentime


*/

#ifdef HAVE_CONFIG_H
# include "config.h"
#endif


#include <glib.h>
#include <epan/packet.h>

#include <stdio.h>
#include <string.h>

#include "packet-ber.h"

#include "asn1.h"
#include "asn1_helper.h"
#include "packet-frame.h"
#include "prefs.h"

#include <epan/strutil.h>

#include "packet-mms.h"

/* proto_iso9506 must be public for plugin init code in packet-mms.c */
int proto_iso9506 	= -1;
int proto_iso9506_abbplc 	= -1;
static dissector_handle_t generic_handle;
void proto_reg_handoff_iso9506(void);



/* prefs must be public for asn1_helper.c */
static gboolean global_mms_asn1_prefs_nodes_as_subtrees = TRUE;
static gboolean global_mms_prefs_use_generic_decoder=FALSE;

/* the following data should be static */

static int ett_iso9506	= -1;
static int ett_service	= -1;
static int ett_varlist	= -1;
static int ett_vardesc	= -1;	  
static int ett_varspec	= -1;
static int ett_addrspec	= -1;
static int ett_typespec	= -1;
static int ett_altspec	= -1;
static int ett_obj	= -1;
static int ett_item	= -1;
static int ett_filetab	= -1;
static int ett_filetab2	= -1;
static int ett_fileattrtab	= -1;
static int ett_domainname	= -1;
static int ett_domainspec	= -1;
static int ett_datatbl		= -1;
static int ett_init_detail 	= -1;
static int ett_init_cbb 	= -1;
static int ett_init_services 	= -1;
static int ett_service_error	= -1;

static int ett_array	= -1;
static int ett_structure	= -1;
static int ett_bitstring	= -1;
static int ett_octettstring	= -1;
static int ett_string	= -1;
static int ett_gentime	= -1;
static int ett_utctime	= -1;
static int ett_bintime	= -1;
static int ett_unicode	= -1;
static int ett_bcd	= -1;
static int ett_oid	= -1;
static int ett_components	= -1;


static int hf_mms_pdu_type  = -1;
static int hf_mms_invoke_id  = -1;
static int hf_mms_service_id  = -1;
static int hf_mms_version	= -1;

static int hf_mms_filename  = -1;
static int hf_mms_domainname  = -1;
static int hf_mms_itemname  = -1;
static int hf_mms_pin  = -1;

static int displayed_header = FALSE;		/*used to display SISCO ad once*/

#define PROTO_STRING_ISO9506	"ISO/IEC 9506 MMS"
#define PROTO_STRING_ISO9506_ABBPLC	"ABBPLC MMS"

#define ASN1H_SEQ_TAG	 (ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ)
#define ASN1H_INTEGER_ID (ASN1H_UNI | ASN1_INT)		/*ASN1_INT is defined in ASN1.h*/
#define ASN1H_OBJECT_ID  (ASN1H_UNI | ASN1_OJI)		/*ASN1_OJI is defined in ASN1.h*/

#define NOETT NULL
#define NOHF  NULL

/********************************************************************/
/*                  File OPEN Request								*/
/********************************************************************/
/*
FileOpen-Request ::= SEQUENCE {
	fileName	[0] IMPLICIT FileName,
	initialPosition	[1] IMPLICIT Unsigned32
	}
*/

static ASN1_TBL_ENTRY filename_tbl[]={
		{(ASN1H_UNI | 25), 			NOETT,NOHF,"",ASN1H_VSTRING_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY file_open_req[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_filetab,&hf_mms_filename,"File Name: ",ASN1H_TBL_SET, filename_tbl},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Initial Position:  ",ASN1H_INTEGER_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/********************************************************************/
/*                  File OPEN Response								*/
/********************************************************************/

/*
FileOpen-Response ::= SEQUENCE {
	frsmID		[0] IMPLICIT Integer32,
	fileAttributes	[1] IMPLICIT FileAttributes
	}

FileAttributes ::= SEQUENCE {
	sizeOfFile	[0] IMPLICIT Unsigned32,
	lastModified	[1] IMPLICIT GeneralizedTime OPTIONAL
	}

*/

static ASN1_TBL_ENTRY file_attr_tbl[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"Size of File: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Last Modified:  ",ASN1H_GENERALTIME_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};



static ASN1_TBL_ENTRY file_open_resp[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"File Resource ID: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_fileattrtab,NOHF,"File Attributes ",ASN1H_TBL_SEQ, file_attr_tbl},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*                  File Read Response								*/
/********************************************************************/

/*
 FileRead-Response ::= SEQUENCE {
	fileData	[0] IMPLICIT OCTET STRING,
	moreFollows	[1] IMPLICIT BOOLEAN DEFAULT TRUE
	}
*/
static ASN1_TBL_ENTRY file_read_resp[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"File Data: ",ASN1H_DONOT_DISPLAY, NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"More Data Follows ",ASN1H_BOOLEAN_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*					File Rename Request								*/
/********************************************************************/

/*
FileRename-Request ::= SEQUENCE {
	currentFileName	[0] IMPLICIT FileName,
	newfileName	[1] IMPLICIT FileName
	}
*/	 
 

static ASN1_TBL_ENTRY file_rename_req[]={
		{(ASN1H_CTX | 0),			&ett_filetab,NOHF,"Current File Name: ",ASN1H_TBL_SET, filename_tbl},
		{(ASN1H_CTX | 1),			&ett_filetab2,NOHF,"New File Name: ",ASN1H_TBL_SET, filename_tbl},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*					File Directory Request								*/
/********************************************************************/

/*
FileDirectory-Request ::= SEQUENCE {
	fileSpecification	[0] IMPLICIT FileName OPTIONAL,
	continueAfter		[1] IMPLICIT FileName OPTIONAL
	}
*/
static ASN1_TBL_ENTRY file_directory_req[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	&ett_filetab,NOHF,"FileSpecification: ",ASN1H_TBL_SET, filename_tbl},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_filetab2,NOHF,"ContinueAfter: ",ASN1H_TBL_SET, filename_tbl},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*					File Directory Response							*/
/********************************************************************/

/*
DirectoryEntry ::= SEQUENCE {
	filename		[0] IMPLICIT FileName,
	fileAttributes		[1] IMPLICIT FileAttributes
	}
*/
static ASN1_TBL_ENTRY filename_dirEntry[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	&ett_filetab,&hf_mms_filename,"File Name: ",ASN1H_TBL_SET, filename_tbl},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_fileattrtab,NOHF,"FileAttributes: ",ASN1H_TBL_SEQ, file_attr_tbl},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
FileDirectory-Response ::= SEQUENCE {
	listOfDirectoryEntry	[0] SEQUENCE OF DirectoryEntry,
	moreFollows		[1] IMPLICIT BOOLEAN DEFAULT FALSE
	}
*/

static ASN1_TBL_ENTRY file_directory_resp1[]={
		{ASN1H_SEQ_TAG,				NOETT,NOHF,NULL,ASN1H_TBL_SET, filename_dirEntry},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY file_directory_resp[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	NOETT,NOHF,"ListOfDirectoryEntries: ",0,NULL},
		{ASN1H_SEQ_TAG,				NOETT,NOHF,NULL,ASN1H_TBL_SET, file_directory_resp1},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	NOETT,NOHF,"moreFollows: ",ASN1H_BOOLEAN_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/********************************************************************/
/*					ObtainFile Req									*/
/********************************************************************/


/*
ApplicationReference ::= SEQUENCE
	{
	ap-title		[0] ISO-8650-ACSE-1.AP-title 	        OPTIONAL,
	ap-invocation-id	[1] ISO-8650-ACSE-1.AP-invocation-identifier OPTIONAL,
	ae-qualifier		[2] ISO-8650-ACSE-1.AE-qualifier	        OPTIONAL,
	ae-invocation-id	[3] ISO-8650-ACSE-1.AE-invocation-identifier OPTIONAL
	}
*/
static ASN1_TBL_ENTRY applicationReference[]={
		{ASN1H_SEQ_TAG,				NOETT,NOHF,"", 0,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0 ),	NOETT,NOHF,"", 0,NULL},
		{(ASN1H_UNI | ASN1_OJI ),		NOETT,NOHF,"AP-title", ASN1H_OBJID_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1 ),	NOETT,NOHF,"", 0,NULL},
		{(ASN1H_UNI | ASN1_INT ),		NOETT,NOHF,"AP Invocation ID: ", ASN1H_INTEGER_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2 ),	NOETT,NOHF,"", 0,NULL},
		{(ASN1H_UNI | ASN1_INT ),		NOETT,NOHF,"AE Qualifier: ", ASN1H_INTEGER_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 3 ),	NOETT,NOHF,"", 0,NULL},
		{(ASN1H_UNI | ASN1_INT ),		NOETT,NOHF,"AE Invocation ID: ", ASN1H_INTEGER_DISPLAY,NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
ObtainFile-Request ::= SEQUENCE {
	sourceFileServer	[0] IMPLICIT ApplicationReference OPTIONAL,
	sourceFile		[1] IMPLICIT FileName,
	destinationFile		[2] IMPLICIT FileName
	}
*/

static ASN1_TBL_ENTRY obtainfile_req[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	NOETT,NOHF,"ApplicationReference ",ASN1H_TBL_SEQ,applicationReference},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_filetab,NOHF,"Source FileName: ",ASN1H_TBL_SET, filename_tbl},
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),	&ett_filetab2,NOHF,"Destination FileName: ",ASN1H_TBL_SET, filename_tbl},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*					CreateProgramInvocation Req						*/
/********************************************************************/

/*
CreateProgramInvocation-Request ::= SEQUENCE
	{
	programInvocationName 	[0] IMPLICIT Identifier,
	listOfDomainName		[1] IMPLICIT SEQUENCE OF Identifier,
	reusable			[2] IMPLICIT BOOLEAN DEFAULT TRUE,
	monitorType			[3] IMPLICIT BOOLEAN OPTIONAL
		-- TRUE indicates PERMANENT monitoring
		-- FALSE indicates CURRENT monitoring
	}

CreateProgramInvocation-Response ::= NULL
*/
static ASN1_TBL_ENTRY domain_names[]={
		{(ASN1H_UNI | 19),			NOETT,NOHF,"Name: ",ASN1H_VSTRING_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY createpi_req[]={
		{(ASN1H_CTX | 0),			NOETT,&hf_mms_pin,"ProgramInvocationName: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_domainname,NOHF,"DomainNames: ",ASN1H_TBL_SET, domain_names},
		{(ASN1H_CTX | 2),			NOETT,NOHF,"Reusable: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 3),			NOETT,NOHF,"MonitorType (TRUE=PERMANENT): ",ASN1H_BOOLEAN_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*				Start Request										*/
/********************************************************************/

/*
Start-Request ::= SEQUENCE	
	{
	programInvocationName 	[0] IMPLICIT Identifier,
	executionArgument	CHOICE {
	    simpleString	  [1] IMPLICIT VisibleString,
	    encodedString	  EXTERNAL
	    } OPTIONAL
	}
*/

static ASN1_TBL_ENTRY startpi_req[]={
		{(ASN1H_CTX | 0),			NOETT,&hf_mms_pin,"ProgramInvocationName: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Execution String: ",ASN1H_VSTRING_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/*
Stop-Request ::= SEQUENCE 
	{
	programInvocationName	[0] IMPLICIT Identifier
	}
*/
static ASN1_TBL_ENTRY stoppi_req[]={
		{(ASN1H_CTX | 0),			NOETT,&hf_mms_pin,"ProgramInvocationName: ",ASN1H_VSTRING_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*

ProgramInvocationState ::= INTEGER
	{
	non-existent		(0),
	unrunable		(1),
	idle			(2),
	running			(3),
	stopped			(4),
	starting		(5),
	stopping		(6),
	resuming		(7),
	resetting		(8)
	}
	-- Companion Standard may add additional values
*/
static ENUM_LOOKUP piState_tbl_text[9]= {
   {0,"non-existent"},
   {1, "unrunable"},
   {2, "idle"},
   {3, "running"},
   {4, "stopped"},
   {5, "starting"},
   {6, "stopping"},
   {7, "resuming"},
   {8, "resetting"}};

static ENUM_XLAT piState_tbl = {9, piState_tbl_text};

/*
GetProgramInvocationAttributes-Response ::= SEQUENCE
	{
	state			[0] IMPLICIT ProgramInvocationState,
	listOfDomainNames	[1] IMPLICIT SEQUENCE OF Identifier,				
	mmsDeletable		[2] IMPLICIT BOOLEAN,
	reusable		[3] IMPLICIT BOOLEAN,
	monitor			[4] IMPLICIT BOOLEAN,
	startArgument		[5] IMPLICIT VisibleString,
	executionArgument	CHOICE {
	    simpleString	  [1] IMPLICIT VisibleString,
	    encodedString	  EXTERNAL
	    } OPTIONAL
	}
*/

static ASN1_TBL_ENTRY getpi_attribs_resp[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"State: ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&piState_tbl},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_domainname,NOHF,"DomainNames: ",ASN1H_TBL_SET, domain_names},
		{(ASN1H_CTX | 2 ),			NOETT,NOHF,"MMS Deletable: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 3 ),			NOETT,NOHF,"Reusable: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 4 ),			NOETT,NOHF,"Monitor (Permanent=TRUE): ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 5 ),			NOETT,NOHF,"Start Arguement ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1 ),			NOETT,NOHF,"Execution Arguement ",ASN1H_VSTRING_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/********************************************************************/
/*					Reject											*/
/********************************************************************/

/*
RejectPDU ::= SEQUENCE
	{
	originalInvokeID		[0] IMPLICIT Unsigned32 OPTIONAL,
	rejectReason 			    CHOICE 
		{
		confirmed-requestPDU		[1] IMPLICIT INTEGER 
			{
			other					(0),
			unrecognized-service			(1),
			unrecognized-modifier			(2),
			invalid-invokeID			(3),
			invalid-argument			(4),
			invalid-modifier			(5),
			max-serv-outstanding-exceeded		(6),
			max-recursion-exceeded			(8),
			value-out-of-range			(9)
			},

		confirmed-responsePDU		[2] IMPLICIT INTEGER 
			{
			other					(0),
			unrecognized-service			(1),
			invalid-invokeID			(2),
			invalid-result				(3),
			max-recursion-exceeded			(5),
			value-out-of-range			(6) 	
			},	
		
		confirmed-errorPDU		[3] IMPLICIT INTEGER 
			{
			other					(0),
			unrecognized-service			(1),
			invalid-invokeID			(2),
			invalid-serviceError			(3),
			value-out-of-range			(4) 	 
			} ,			
			
		unconfirmedPDU			[4] IMPLICIT INTEGER
			{
			other					(0),
			unrecognized-service			(1),
			invalid-argument			(2),
			max-recursion-exceeded			(3),
			value-out-of-range			(4) 	
			},

		pdu-error			[5] IMPLICIT INTEGER
			{
			unknown-pdu-type			(0),
			invalid-pdu				(1),
			illegal-acse-mapping			(2)
			},

		cancel-requestPDU		[6] IMPLICIT INTEGER
			{
			other					(0),
			invalid-invokeID			(1)	
			},	

		cancel-responsePDU	        [7] IMPLICIT INTEGER
			{
			other					(0),
			invalid-invokeID			(1)	
			},

		cancel-errorPDU			[8] IMPLICIT INTEGER
			{
			other					(0),
			invalid-invokeID			(1),
			invalid-serviceError			(2),
			value-out-of-range			(3) 	
			},

		conclude-requestPDU		[9] IMPLICIT INTEGER
			{
			other					(0),
			invalid-argument			(1)	
			},	

		conclude-responsePDU	        [10] IMPLICIT INTEGER
			{
			other					(0),
			invalid-result				(1)	
			},

		conclude-errorPDU		[11] IMPLICIT INTEGER
			{
			other					(0),
			invalid-serviceError			(1),
			value-out-of-range			(2) 	
			}
		}
	}
*/

static ENUM_LOOKUP confirmedReject_text[9]= {
   {0,"Other"},
   {1, "Unrecognized Service"},
   {2, "Unrecognized Modifier"},
   {3, "Invalid InvokeID"},
   {4, "Invalid Argument"},
   {5, "Invalid Modifier"},
   {6, "Max Services Outstanding Exceeded"},
   {7, "Max Recursion Exceeded"},
   {8, "Value Out-of-range"}};

static ENUM_XLAT confirmedReject = {9, confirmedReject_text};


static ENUM_LOOKUP confirmedRespReject_text[7]= {
   {0,"Other"},
   {1, "Unrecognized Service"},
   {2, "Unrecognized Modifier"},
   {3, "Invalid InvokeID"},
   {4, "Invalid Result"},
   {5, "Max Recursion Exceeded"},
   {6, "Value Out-of-range"}};

static ENUM_XLAT confirmedRespReject = {7, confirmedRespReject_text};

static ENUM_LOOKUP confirmedErrReject_text[5]= {
   {0,"Other"},
   {1, "Unrecognized Service"},
   {2, "Invalid InvokeID"},
   {3, "Invalid Service Error"},
   {4, "Value Out-of-range"}};

static ENUM_XLAT confirmedErrReject = {5, confirmedErrReject_text};

static ENUM_LOOKUP UnConfReject_text[4]= {
   {0,"Other"},
   {1, "Unrecognized Service"},
   {2, "Invalid Argument"},
   {3, "Value Out-of-range"}};

static ENUM_XLAT UnConfReject = {4, UnConfReject_text};

static ENUM_LOOKUP PDUErrReject_text[3]= {
   {0,"Unknown PDU Type"},
   {1, "Invalid PDU"},
   {2, "Illegal ACSE Mapping"}};

static ENUM_XLAT PDUErrReject = {3, PDUErrReject_text};

static ENUM_LOOKUP CancelReqReject_text[2]= {
   {0,"Other"},
   {1, "Invalid InvokeID"}};

static ENUM_XLAT CancelReqReject = {2, CancelReqReject_text};

static ENUM_LOOKUP CancelErrReject_text[4]= {
   {0,"Other"},
   {1, "Invalid InvokeID"},
   {2, "Invalid Service Error"},
   {3, "Value Out-of-range"}};
					   
static ENUM_XLAT CancelErrReject = {4, CancelErrReject_text};

static ENUM_LOOKUP ConcReqReject_text[2]= {
   {0,"Other"},
   {1, "Invalid Arguement"}};
					   
static ENUM_XLAT ConcReqReject = {2, ConcReqReject_text};

static ENUM_LOOKUP ConcRespReject_text[2]= {
   {0,"Other"},
   {1, "Invalid Result"}};
					   
static ENUM_XLAT ConcRespReject = {2, ConcRespReject_text};

static ENUM_LOOKUP ConcErrReject_text[3]= {
   {0,"Other"},
   {1, "Invalid Service Error"},
   {2, "Value Out-of-range"}};
					   
static ENUM_XLAT ConcErrReject = {3, ConcErrReject_text};


static ASN1_TBL_ENTRY mms_reject[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"Original Invoke ID: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOETT,NOHF,"Confirmed Request Reject  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&confirmedReject},
		{(ASN1H_CTX | 2),NOETT,NOHF,"Confirmed Response Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&confirmedRespReject},
		{(ASN1H_CTX | 3),NOETT,NOHF,"Confirmed Error Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&confirmedErrReject},
		{(ASN1H_CTX | 4),NOETT,NOHF,"Unconfirmed Service Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&UnConfReject},
		{(ASN1H_CTX | 5),NOETT,NOHF,"PDU Error Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&PDUErrReject},
		{(ASN1H_CTX | 6),NOETT,NOHF,"Cancel Request Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&CancelReqReject},
		{(ASN1H_CTX | 7),NOETT,NOHF,"Cancel Response Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&CancelReqReject}, /*same mappings*/
		{(ASN1H_CTX | 8),NOETT,NOHF,"Cancel Error Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&CancelErrReject},
		{(ASN1H_CTX | 9),NOETT,NOHF,"Conclude Request Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&ConcReqReject},
		{(ASN1H_CTX | 10),NOETT,NOHF,"Conclude Response Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&ConcRespReject},
		{(ASN1H_CTX | 11),NOETT,NOHF,"Conclude Error Reject:  ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&ConcErrReject},
   		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};


/*****************************************************************/
/*   Initiate Request Decode									 */
/*****************************************************************/

/*
ServiceSupportOptions ::= BIT STRING 
	{
	status								(0),	
	getNameList							(1), 	
	identify							(2), 		
	rename								(3),	
	read								(4),	
	write								(5),	
	getVariableAccessAttributes			(6),	
	defineNamedVariable					(7),	
	defineScatteredAccess				(8),	
	getScatteredAccessAttributes		(9),	
	deleteVariableAccess				(10),		
	defineNamedVariableList				(11),	
	getNamedVariableListAttributes		(12),	
	deleteNamedVariableList				(13),	
	defineNamedType						(14),		
	getNamedTypeAttributes				(15),	
	deleteNamedType						(16),	
	input								(17),	
	output								(18),	
	takeControl							(19),	
	relinquishControl					(20),	
	defineSemaphore						(21),	
	deleteSemaphore						(22),	
	reportSemaphoreStatus				(23),	
	reportPoolSemaphoreStatus			(24),	
	reportSemaphoreEntryStatus			(25),	
	initiateDownloadSequence			(26),	
	downloadSegment						(27),	
	terminateDownloadSequence			(28),	
	initiateUploadSequence				(29),	
	uploadSegment						(30),	
	terminateUploadSequence				(31),	
	requestDomainDownload				(32),	
	requestDomainUpload					(33),	
	loadDomainContent					(34),	
	storeDomainContent					(35),	
	deleteDomain						(36),	
	getDomainAttributes					(37),	
	createProgramInvocation				(38),	
	deleteProgramInvocation				(39),	
	start								(40),	
	stop								(41),	
	resume								(42),	
	reset								(43),	
	kill								(44),	
	getProgramInvocationAttributes		(45),	
	obtainFile							(46),	
	defineEventCondition				(47),	
	deleteEventCondition				(48),	
	getEventConditionAttributes			(49),	
	reportEventConditionStatus			(50),	
	alterEventConditionMonitoring		(51),	
	triggerEvent						(52),	
	defineEventAction					(53),	
	deleteEventAction					(54),	
	getEventActionAttributes			(55),	
	reportActionStatus					(56),	
	defineEventEnrollment				(57),	
	deleteEventEnrollment				(58),	
	alterEventEnrollment				(59),	
	reportEventEnrollmentStatus			(60),	
	getEventEnrollmentAttributes		(61),	
	acknowledgeEventNotification		(62),	
	getAlarmSummary						(63),	
	getAlarmEnrollmentSummary			(64),	
	readJournal							(65),	
	writeJournal						(66),	
	initializeJournal					(67),	
	reportJournalStatus					(68),	
	createJournal						(69),	
	deleteJournal						(70),	
	getCapabilityList					(71),	
	fileOpen							(72),	
	fileRead							(73),	
	fileClose							(74),	
	fileRename							(75),	
	fileDelete							(76),	
	fileDirectory						(77),	
	unsolicitedStatus					(78),
	informationReport					(79),
	eventNotification					(80),
	attachToEventCondition				(81),
	attachToSemaphore					(82),
	conclude							(83),
	cancel								(84)
	}
*/
static ENUM_LOOKUP servicesSupported_text[85]= {
   {0,		"status"	 						},						 
   {1, 		"getNameList"						}, 
   {2, 		"identify"						}, 	
   {3,		"rename"							}, 
   {4,		"read"							}, 
   {5,		"write"							}, 
   {6,		"getVariableAccessAttributes"		}, 
   {7,		"defineNamedVariable"				}, 
   {8,		"defineScatteredAccess"			}, 
   {9,		"getScatteredAccessAttributes"	}, 
   {10,		"deleteVariableAccess"			}, 	
   {11,		"defineNamedVariableList"			}, 
   {12,		"getNamedVariableListAttributes"	}, 
   {13,		"deleteNamedVariableList"			}, 
   {14,		"defineNamedType"					}, 	
   {15,		"getNamedTypeAttributes"			}, 
   {16,		"deleteNamedType"					}, 
   {17,		"input"							}, 
   {18,		"output"							}, 
   {19,		"takeControl"						}, 
   {20,		"relinquishControl"				}, 
   {21,		"defineSemaphore"					}, 
   {22,		"deleteSemaphore"					}, 
   {23,		"reportSemaphoreStatus"			}, 
   {24,		"reportPoolSemaphoreStatus"		}, 
   {25,		"reportSemaphoreEntryStatus"		}, 
   {26,		"initiateDownloadSequence"		}, 
   {27,		"downloadSegment"					}, 
   {28,		"terminateDownloadSequence"		}, 
   {29,		"initiateUploadSequence"			}, 
   {30,		"uploadSegment"					}, 
   {31,		"terminateUploadSequence"			}, 
   {32,		"requestDomainDownload"			}, 
   {33,		"requestDomainUpload"				}, 
   {34,		"loadDomainContent"				}, 
   {35,		"storeDomainContent"				}, 
   {36,		"deleteDomain"					}, 
   {37,		"getDomainAttributes"				}, 
   {38,		"createProgramInvocation"			}, 
   {39,		"deleteProgramInvocation"			}, 
   {40,		"start"							}, 
   {41,		"stop"							}, 
   {42,		"resume"							}, 
   {43,		"reset"							}, 
   {44,		"kill"							}, 
   {45,		"getProgramInvocationAttributes"	}, 
   {46,		"obtainFile"						}, 
   {47,		"defineEventCondition"			}, 
   {48,		"deleteEventCondition"			}, 
   {49,		"getEventConditionAttributes"		}, 
   {50,		"reportEventConditionStatus"		}, 
   {51,		"alterEventConditionMonitoring"	}, 
   {52,		"triggerEvent"					}, 
   {53,		"defineEventAction"				}, 
   {54,		"deleteEventAction"				}, 
   {55,		"getEventActionAttributes"		}, 
   {56,		"reportActionStatus"				}, 
   {57,		"defineEventEnrollment"			}, 
   {58,		"deleteEventEnrollment"			}, 
   {59,		"alterEventEnrollment"			}, 
   {60,		"reportEventEnrollmentStatus"		}, 
   {61,		"getEventEnrollmentAttributes"	}, 
   {62,		"acknowledgeEventNotification"	}, 
   {63,		"getAlarmSummary"					}, 
   {64,		"getAlarmEnrollmentSummary"		}, 
   {65,		"readJournal"						}, 
   {66,		"writeJournal"					}, 
   {67,		"initializeJournal"				}, 
   {68,		"reportJournalStatus"				}, 
   {69,		"createJournal"					}, 
   {70,		"deleteJournal"					}, 
   {71,		"getCapabilityList"				}, 
   {72,		"fileOpen"						}, 
   {73,		"fileRead"						}, 
   {74,		"fileClose"						}, 
   {75,		"fileRename"						}, 
   {76,		"fileDelete"						}, 
   {77,		"fileDirectory"					}, 
   {78,	"unsolicitedStatus"				 	},
   {79,	"informationReport"				 	},
   {80,	"eventNotification"				 	},
   {81,	"attachToEventCondition"			 	},
   {82,	"attachToSemaphore"				 	},
   {83,	"conclude"						 	},
   {84,	"cancel"							 	}};

static ENUM_XLAT servicesSupported = {85, servicesSupported_text};


/*
ParameterSupportOptions ::= BIT STRING {
	str1 (0),
	str2 (1),
	vnam (2),
	valt (3),
	vadr (4),
	vsca (5),
	tpy  (6),
	vlis (7),
	real (8),
	cei  (10)
	}
*/
static ENUM_LOOKUP paramCBB_text[10]= {
   {0,"Array Support [STR1]"},
   {1, "Structure Support [STR2]"},
   {2, "Named Variable Support [VNAM]"},
   {3, "Alternate Access Support [VALT]"},
   {4, "Addressed Variable Support [VADR]"},
   {5, "Scattered Access (depricated) [VSCA]"},
   {6, "Third Pary Service Support [TPY]"},
   {7, "Named Variable List Support [VLIS]"},
   {8, "ASN.1 Floating Point Support (depricated) [real]"},
   {9, "Conditioned Event Support [CEI]"}};

static ENUM_XLAT paramCBBs = {10, paramCBB_text};

/*
InitRequestDetail ::= SEQUENCE
	{
	proposedVersionNumber			[0] IMPLICIT Integer16,
	proposedParameterCBB			[1] IMPLICIT ParameterSupportOptions,
	servicesSupportedCalling		[2] IMPLICIT ServiceSupportOptions
	}
*/

static ASN1_TBL_ENTRY mms_init_detail[]={
		{(ASN1H_CTX | 0),	NOETT,&hf_mms_version,"Proposed MMS Version Number:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),	&ett_init_cbb,NOHF,"Proposed Parameter CBBs:  ",ASN1H_BITSTRING_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&paramCBBs},
		{(ASN1H_CTX | 2),	&ett_init_services,NOHF,"Services Supported Calling:  ",ASN1H_BITSTRING_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&servicesSupported},
   		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
Initiate-RequestPDU ::= SEQUENCE
	{
	localDetailCalling	        	[0] IMPLICIT Integer32 OPTIONAL,
	proposedMaxServOutstandingCalling	[1] IMPLICIT Integer16,
	proposedMaxServOutstandingCalled	[2] IMPLICIT Integer16,
	proposedDataStructureNestingLevel	[3] IMPLICIT Integer8 OPTIONAL,
	mmsInitRequestDetail			[4] IMPLICIT InitRequestDetail
	}
*/

static ASN1_TBL_ENTRY mms_initiate_req[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"Proposed MMS PDU Size: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOETT,NOHF,"Proposed Outstanding Requests Calling:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 2),NOETT,NOHF,"Proposed Outstanding Requests Called:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 3),NOETT,NOHF,"Proposed Data Nesting Level: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 4),&ett_init_detail,NOHF,"Initiate Request Detail",ASN1H_TBL_SEQ,mms_init_detail},
   		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};


/****************************************************************************/
/*				InitiateResponse											*/
/****************************************************************************/

/*
Initiate-ResponsePDU ::= SEQUENCE
	{
	localDetailCalled			[0] IMPLICIT Integer32 OPTIONAL,
	negociatedMaxServOutstandingCalling	[1] IMPLICIT Integer16,
	negociatedMaxServOutstandingCalled	[2] IMPLICIT Integer16,
	negociatedDataStructureNestingLevel	[3] IMPLICIT Integer8 OPTIONAL,
	mmsInitResponseDetail			[4] IMPLICIT InitResponseDetail
	}
*/
static ASN1_TBL_ENTRY mms_init_resp_detail[]={
		{(ASN1H_CTX | 0),NOETT,&hf_mms_version,"Negotiated MMS Version Number:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),&ett_init_cbb,NOHF,"Negotiated Parameter CBBs:  ",ASN1H_BITSTRING_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&paramCBBs},
		{(ASN1H_CTX | 2),&ett_init_services,NOHF,"Services Supported Called:  ",ASN1H_BITSTRING_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&servicesSupported},
   		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY mms_initiate_resp[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"Negotiated MMS PDU Size: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Negotiated Max Outstatind Requests Calling:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 2),			NOETT,NOHF,"Negotiated Outstanding Requests Called:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 3),			NOETT,NOHF,"Negotiated Data Nesting Level: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 4),	&ett_init_detail,NOHF,"Initiate Response Detail",ASN1H_TBL_SEQ,mms_init_resp_detail},
   		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/****************************************************************************/
/*			Initiate Error	- See Service Error								*/
/****************************************************************************/


/****************************************************************************/
static ASN1_TBL_ENTRY mms_service_error2[]={
		{(ASN1H_CTX | 0),	NOETT,NOHF,"VMD State: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),	NOETT,NOHF,"Application Reference:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 2),	NOETT,NOHF,"Definition:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 3),	NOETT,NOHF,"Resource: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 4),	NOETT,NOHF,"Service: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 5),	NOETT,NOHF,"Service Pre-empt:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 6),	NOETT,NOHF,"Time Resolution: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 7),	NOETT,NOHF,"Access: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 8),	NOETT,NOHF,"Initiate:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 9),	NOETT,NOHF,"Conclude:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 10),	NOETT,NOHF,"Cancel: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 11),	NOETT,NOHF,"File:  ",ASN1H_INTEGER_DISPLAY, NULL},
   		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY mms_service_error1[]={
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 0),	NOETT,NOHF,"",0,NULL},			/*choice tag */
		{(ASN1H_SEQ_TAG),			NOETT,NOHF,"",ASN1H_TBL_SEQ,mms_service_error2},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Additional Code:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 2),			NOETT,NOHF,"Additional Description:  ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 3),	NOETT,NOHF,"ServiceSpecific: ",0, NULL},
		{(ASN1H_SEQ_TAG),			NOETT,NOHF,"",0,NULL},
		{(ASN1H_CTX | 0),			NOETT,NOHF,"ObtainFile:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Start:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 2),			NOETT,NOHF,"Stop:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 3),			NOETT,NOHF,"Resume:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 4),			NOETT,NOHF,"Reset:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 5),			NOETT,NOHF,"DeleteVariableAccess (num deleted):  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 6),			NOETT,NOHF,"DeleteNamedVariableList (num deleted):  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 7),			NOETT,NOHF,"DeleteNamedType (num deleted):  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 8),			NOETT,NOHF,"DefineEventEnrollment(not decoded)  ",0, NULL},
		{(ASN1H_CTX | 9),			NOETT,NOHF,"FileRename:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 10),			NOETT,NOHF,"Additional Code (not decoded)  ",0, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY mms_service_error[]={
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 0),	NOETT,NOHF,"",0,NULL},			/*choice tag */
		{(ASN1H_SEQ_TAG),			NOETT,NOHF,"",ASN1H_TBL_SEQ,mms_service_error1},
		{(ASN1H_CTX | 0),			NOETT,NOHF,"ObtainFile:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Start:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 2),			NOETT,NOHF,"Stop:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 3),			NOETT,NOHF,"Resume:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 4),			NOETT,NOHF,"Reset:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 5),			NOETT,NOHF,"DeleteVariableAccess (num deleted):  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 6),			NOETT,NOHF,"DeleteNamedVariableList (num deleted):  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 7),			NOETT,NOHF,"DeleteNamedType (num deleted):  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 8),			NOETT,NOHF,"DefineEventEnrollment(not decoded)  ",0, NULL},
		{(ASN1H_CTX | 9),			NOETT,NOHF,"FileRename:  ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 10),			NOETT,NOHF,"Additional Code (not decoded)  ",0, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY object_name[]={
		{(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ),	NOETT,NOHF,NULL, ASN1H_TBL_SEQ ,NULL},
		{(ASN1H_CTX | 0),				NOETT,NOHF,"VMD Specific ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),		&ett_domainspec,NOHF,"Domain Specific ",ASN1H_TBL_SEQ,NULL},
		{(ASN1H_UNI | ASN1_VISSTR),			&ett_domainname,&hf_mms_domainname,"Domain Name: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_UNI | ASN1_VISSTR),			&ett_item,&hf_mms_itemname,"Item Name: ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 2),				NOETT,NOHF,"AA-Specific ",ASN1H_VSTRING_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE				,NOETT,NOHF,NULL,0,NULL}
};

/*
Address ::= CHOICE
	{
	numericAddress		[0] IMPLICIT Unsigned32,
	symbolicAddress		[1] IMPLICIT VisibleString,
	unconstrainedAddress	[2] IMPLICIT OCTET STRING
	}
*/
static ASN1_TBL_ENTRY address_spec[]={
		{(ASN1H_CTX | 0), NOETT,NOHF,"NumericAddress: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{(ASN1H_CTX | 1), NOETT,NOHF,"symbolicAddress:  ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 2), NOETT,NOHF,"unconstrained:  ",ASN1H_OCTETSTRING_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};



/*
AlternateAccess ::= SEQUENCE OF CHOICE
	{
	unnamed	AlternateAccessSelection,
	named		[5] IMPLICIT SEQUENCE 
		{
		componentName	[0] IMPLICIT Identifier,
		accesst		AlternateAccessSelection
		}
	}
*/

static ASN1_TBL_ENTRY aa_index_range[]={
		{(ASN1H_CTX | 0), NOETT,NOHF,"LowIndex:  ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{(ASN1H_CTX | 1), NOETT,NOHF,"NumberOfElements:  ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
AlternateAccessSelection ::= CHOICE
	{
	selectAccess	CHOICE
		{
		component	[1] IMPLICIT Identifier,
		index		[2] IMPLICIT Unsigned32,
		indexRange	[3] IMPLICIT SEQUENCE
			{
			lowIndex			[0] IMPLICIT Unsigned32,
			numberOfElements		[1] IMPLICIT Unsigned32
			},
		allElements	[4] IMPLICIT NULL	-- all array elements
		}
	}
*/

/*
AlternateAccess ::= SEQUENCE OF CHOICE
	{
	unnamed	AlternateAccessSelection,
	named		[5] IMPLICIT SEQUENCE 
		{
		componentName	[0] IMPLICIT Identifier,
		accesst		AlternateAccessSelection
		}
	}
*/
extern ASN1_TBL_ENTRY alternate_spec[];

static ASN1_TBL_ENTRY select_alternate[]={
		{(ASN1H_CONSTRUCT | ASN1_SEQ), &ett_altspec,NOHF,NULL,ASN1H_TBL_SET,alternate_spec},
		{(ASN1H_CTX | 0), NOETT,NOHF,"Component:  ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 1), NOETT,NOHF,"Index:  ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT| 2), NOETT,NOHF,"indexRange"  ,ASN1H_TBL_SEQ,aa_index_range},
		{(ASN1H_CTX | 3), NOETT,NOHF,"allElements"  ,ASN1H_NULL_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY alternate_spec[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT), NOETT,NOHF,"selectAlternate",ASN1H_TBL_SET,select_alternate},
		{(ASN1H_CTX | 1), NOETT,NOHF,"Component:  ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 2), NOETT,NOHF,"Index:  ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT| 3), NOETT,NOHF,"indexRange"  ,ASN1H_TBL_SEQ,aa_index_range},
		{(ASN1H_CTX | 4), NOETT,NOHF,"allElements"  ,ASN1H_NULL_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY vardesc_tbl[]={
		{(ASN1H_CONSTRUCT | ASN1_SEQ), 		NOETT,NOHF,NULL,0,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0), 	NOETT,NOHF,NULL,0,NULL},	/*tag for Object Name */
		{(ASN1H_CONSTRUCT | ASN1_SEQ), 		&ett_obj,NOHF,"Object Name",ASN1H_TBL_SEQ, object_name},
 		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY list_of_var[]={
		{(ASN1H_CONSTRUCT | ASN1_SEQ),		NOETT,NOHF,NULL,0,NULL},
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 0),	&ett_obj,NOHF,"Object Name",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 1), 	&ett_addrspec,NOHF,"Address",ASN1H_TBL_SEQ, address_spec},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2 ),	&ett_vardesc,NOHF,"Variable Description",0,vardesc_tbl},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 5), 	&ett_altspec,NOHF,"AlternateAccess",ASN1H_TBL_SET,alternate_spec},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};




/*
VariableAccessSpecification ::= CHOICE
	{
	listOfVariable		[0] IMPLICIT SEQUENCE OF SEQUENCE
		{
		variableSpecification	VariableSpecification,
		alternateAccess		[5] IMPLICIT AlternateAccess OPTIONAL
		},
	variableListName	[1] ObjectName
	}
*/
static ASN1_TBL_ENTRY var_spec_tbl[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_varlist,NOHF,"List of Variable",0,NULL},
		{ASN1H_SEQ_TAG,				&ett_varlist,NOHF,"VariableSpecification",ASN1H_TBL_SEQ,list_of_var},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 5), 	&ett_altspec,NOHF,"AlternateAccess",ASN1H_TBL_SEQ,alternate_spec},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1), 	&ett_varlist,NOHF,"VariableList",ASN1H_TBL_SEQ,object_name},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/*
DataAccessError ::= INTEGER
	{
	object-invalidated		(0),
	hardware-fault			(1),
	temporarly-unavailable		(2),
	object-access-denied		(3),
	object-undefined		(4),
	invalid-address			(5),
	type-unsupported		(6),
	type-inconsistent		(7),
	object-attribute-inconsistent	(8),
	object-access-unsupported	(9),
	object-non-existent		(10)
	}
*/
static ENUM_LOOKUP dtaacess_error1[12]= {
   {0,"object-invalidated"},
   {1, "hardware-fault"},
   {2, "temporarly-unavailable"},
   {3, "object-access-denied"},
   {4, "object-undefined"},
   {5, "invalid-address"},
   {6, "type-unsupported"},
   {7, "type-inconsistent"},
   {8, "object-attribute-inconsistent"},
   {9, "object-access-unsupported"},
   {10, "ProgramInvocation"},
   {11, "object-non-existent"}};

static ENUM_XLAT dtaacess_error = {12, dtaacess_error1};

static ASN1_TBL_ENTRY data_tbl[]={
        {(ASN1H_CTX | 0), 			NOETT,NOHF,"DATA Access Error: ", ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&dtaacess_error},
        {(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_array,NOHF,"ARRAY", 0,NULL},
        {(ASN1H_CTX | ASN1H_CONSTRUCT | 2),	&ett_structure,NOHF,"STRUCTURE", 0,NULL},
        {(ASN1H_CTX |  3),			NOETT,NOHF,"BOOLEAN: ",ASN1H_BOOLEAN_DISPLAY,NULL},
        {(ASN1H_CTX |  4),			&ett_bitstring,NOHF,"BITSTRING: ",ASN1H_BITSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  5),			NOETT,NOHF,"INTEGER: ", ASN1H_INTEGER_DISPLAY,NULL},
        {(ASN1H_CTX |  6),			NOETT,NOHF,"UNSIGNED: ", ASN1H_UNSIGNED_DISPLAY,NULL},
        {(ASN1H_CTX |  7),			NOETT,NOHF,"FLOAT: ", ASN1H_MMS_FLOAT_DISPLAY,NULL},
        {(ASN1H_CTX |  9),			&ett_octettstring,NOHF,"OSTRING: ", ASN1H_OCTETSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  10),			&ett_string,NOHF,"VSTRING: ", ASN1H_VSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  11),			&ett_gentime,NOHF,"GTIME ", ASN1H_MMS_BTIME_DISPLAY,NULL},
        {(ASN1H_CTX |  12),			&ett_bintime,NOHF,"BTIME ", ASN1H_MMS_BTIME_DISPLAY,NULL},
        {(ASN1H_CTX |  13),			&ett_bcd,NOHF,"BCD ", ASN1H_UNSIGNED_DISPLAY,NULL},
        {(ASN1H_CTX |  14),			&ett_octettstring,NOHF,"BOOLEAN Array ", ASN1H_OCTETSTRING_DISPLAY,NULL},
        {(ASN1H_CTX |  15),			&ett_oid,NOHF,"OBJID ", ASN1H_OBJID_DISPLAY,NULL},
	{(ASN1H_CTX |  16),			&ett_unicode,NOHF,"UNICODE",ASN1H_MMS_UNICODE_DISPLAY,NULL},	   
	{(ASN1H_CTX |  17), 			&ett_utctime,NOHF,"UTC",ASN1H_MMS_UTCTIME_DISPLAY,NULL},
	{(ASN1H_CTX |  18), 			&ett_unicode,NOHF,"UNICODE(CDV)",ASN1H_MMS_UNICODE_DISPLAY,NULL}, 
	{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY read_resp[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),&ett_datatbl,NOHF,NULL,0,data_tbl},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
Read-Request ::= SEQUENCE
	{
	specificationWithResult		[0] IMPLICIT BOOLEAN DEFAULT FALSE,
	variableAccessSpecificatn	[1] VariableAccessSpecification
	}
*/
static ASN1_TBL_ENTRY read_req[]={
		{(ASN1H_CTX | 0), 			NOETT,NOHF,"Specification with Result: ",ASN1H_BOOLEAN_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1), 	&ett_varspec,NOHF,NULL,ASN1H_TBL_SET,var_spec_tbl},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 5), 	&ett_altspec,NOHF,"AlternateAccess",ASN1H_TBL_SEQ,alternate_spec},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY write_resp[]={
		{(ASN1H_CTX |  0),	NOETT,NOHF,"Data Write Failure; ",ASN1H_INTEGER_ENUM_DISPLAY,(struct asn1h_tbl_entry*)&dtaacess_error},
		{(ASN1H_CTX |  1), 	NOETT,NOHF,"Data Write Success",ASN1H_DONOT_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY write_req[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_varlist,NOHF,"List of Variable",ASN1H_TBL_SET,list_of_var},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1), 	&ett_varlist,NOHF,"VariableList",ASN1H_TBL_SEQ,object_name},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_datatbl,NOHF,"Data",ASN1H_TBL_SEQ,data_tbl},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*************************************************************************/
/*InformationReport ::= SEQUENCE
	{
	variableAccessSpecification	VariableAccessSpecification,
	listOfAccessResult		[0] IMPLICIT SEQUENCE OF AccessResult
	}

VariableAccessSpecification ::= CHOICE
	{
	listOfVariable		[0] IMPLICIT SEQUENCE OF SEQUENCE
		{
		variableSpecification	VariableSpecification,
		alternateAccess		[5] IMPLICIT AlternateAccess OPTIONAL
		},
	variableListName	[1] ObjectName
	}
*/



static ASN1_TBL_ENTRY info_rpt_dec[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_varlist,NOHF,"List of Variable",ASN1H_TBL_SET,list_of_var},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 5), 	&ett_altspec,NOHF,"AlternateAccess",ASN1H_TBL_SEQ,alternate_spec},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1), 	&ett_varlist,NOHF,"VariableList",ASN1H_TBL_SEQ,object_name},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_datatbl,NOHF,"AccessResults",ASN1H_TBL_SET,data_tbl},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/************************************************************************/
/*
Status-Response ::= SEQUENCE {
    vmdLogicalStatus	     [0] IMPLICIT INTEGER {
      state-changes-allowed	  (0),
      no-state-changes-allowed	  (1),
      limited-services-allowed	  (2),
      support-services-allowed	  (3)
      },
    vmdPhysicalStatus	     [1] IMPLICIT INTEGER {
      operational		  (0),
      partially-operational	  (1),
      inoperable		  (2),
      needs-commissioning	  (3)
      },
    localDetail		     [2] IMPLICIT BIT STRING(SIZE(0..128)) OPTIONAL

    }
*/

static ENUM_LOOKUP statusvalues_1[4]= {
   {0,"state-changes-allowed"},
   {1, "no-state-changes-allowed"},
   {2, "limited-services-allowed"},
   {3, "support-services-allowed"}};

static ENUM_XLAT stvmd_values = {4, statusvalues_1};

static ENUM_LOOKUP statusvalues_2[4]= {
   {0,"operational"},
   {1, "partially-operational"},
   {2, "inoperable"},
   {3, "needs-commissioning"}};

static ENUM_XLAT stphy_values = {4, statusvalues_2};

/* Status response decoding												*/
static ASN1_TBL_ENTRY status_resp[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"VMD Logical Status: ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&stvmd_values},
		{(ASN1H_CTX | 1),NOETT,NOHF,"VMD Physical Status: ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&stphy_values},
		{(ASN1H_CTX | 2),NOETT,NOHF,"localDetail: ",ASN1H_BITSTRING_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY mms_unconfirmed_pdu[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_service,NOHF,"InformationReport",ASN1H_TBL_SEQ, info_rpt_dec},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_service,NOHF,"UnsolicitedStatus",ASN1H_TBL_SEQ, status_resp},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),	&ett_service,NOHF,"EventNotification (not currently decoded)",0, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*************************************************************************/
/* Identify decoding													 */
/* no additional decoding is required for the request since it is NULL	*/

static ASN1_TBL_ENTRY identify_resp[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"VendorName: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOETT,NOHF,"ModelName:  ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 2),NOETT,NOHF,"Revision:  ",ASN1H_VSTRING_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};



/*************************************************************************/
/* getNameList decoding													 */

/*GetNameList-Response ::= SEQUENCE
	{
	listOfIdentifier	[0] IMPLICIT SEQUENCE OF Identifier,
	moreFollows		[1] IMPLICIT BOOLEAN DEFAULT TRUE
	}
*/

static ASN1_TBL_ENTRY getNameList_resp[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),NOETT,NOHF,"ListOfIdentifier",0, NULL},
		{ (ASN1H_UNI | ASN1_VISSTR ),NOETT,NOHF,"", ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1 ), NOETT,NOHF,"MoreFollows" ,ASN1H_BOOLEAN_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY getNameListreq_choice[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"VMD Specific ",0, NULL},
		{(ASN1H_CTX | 1 ),NOETT,NOHF,"Domain: ",ASN1H_VSTRING_DISPLAY, getNameListreq_choice},
		{(ASN1H_CTX | 3),NOETT,NOHF,"AA Specific",0, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ENUM_LOOKUP getNameListClassList[12]= {
   {0,"NamedVariable"},
   {1, "ScatteredAccess - Depricated use in 9506"},
   {2, "NamedVariableList"},
   {3, "NamedType"},
   {4, "Semaphore"},
   {5, "EventCondition"},
   {6, "EventAction"},
   {7, "EventEnrollment"},
   {8, "Journal"},
   {9, "Domain"},
   {10, "ProgramInvocation"},
   {11, "OperatorStation"}};

static ENUM_XLAT getNameListreq_choice1 = {12  , getNameListClassList};
/*
GetNameList-Request ::= SEQUENCE
	{
	extendedObjectClass [0] CHOICE
		{
		objectClass		[0] IMPLICIT INTEGER
			{
			nammedVariable		(0),
			scatteredAccess		(1),
			namedVariableList	(2),
			namedType		(3),
			semaphore		(4),
			eventCondition		(5),
			eventAction		(6),
			eventEnrollment		(7),
			journal			(8),
			domain			(9),
			programInvocation	(10),
			operatorStation		(11)
			}
		},
	objectScope 		[1] CHOICE
		{
		vmdSpecific		[0] IMPLICIT NULL,
		domainSpecific		[1] IMPLICIT Identifier,
		aaSpecific		[2] IMPLICIT NULL
		},
	continueAfter 	[2] IMPLICIT Identifier OPTIONAL
	}
*/
static ASN1_TBL_ENTRY getObjectScope_tbl[]={
		{(ASN1H_CTX | 0),	NOETT,NOHF,"VmdSpecific",ASN1H_NULL_DISPLAY, NULL},
		{(ASN1H_CTX | 1),	NOETT,NOHF,"DomainSpecific: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 2 ),	NOETT,NOHF,"aaSpecific",ASN1H_NULL_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY getNameList_req[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),NOETT,NOHF,"extendedObjectClass",ASN1H_TBL_SEQ, NULL},
		{(ASN1H_CTX | 0),NOETT,NOHF,"OBJECT Class: ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&getNameListreq_choice1},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),NOETT,NOHF,"objectScope",ASN1H_TBL_SEQ, getObjectScope_tbl},
		{(ASN1H_CTX | 2),NOETT,NOHF,"Continue After Name:  ",ASN1H_VSTRING_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};
/************************************************************************/
/* GetVariableAccess Attribute Decoding									*/
/*
GetVariableAccessAttributes-Request ::= CHOICE
	{
	name		[0] ObjectName,
	address		[1] Address
	}
*/

static ASN1_TBL_ENTRY getVAA_req[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_obj,NOHF,"Object Name",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_addrspec,NOHF,"Address",0,address_spec },
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
TypeSpecification ::= CHOICE 
	{
	typeName		[0] ObjectName,
	array			[1] IMPLICIT SEQUENCE
		{
		packed			[0] IMPLICIT BOOLEAN DEFAULT FALSE,
		numberOfElements	[1] IMPLICIT Unsigned32,
		elementType		[2] TypeSpecification,
		},
	structure		[2] IMPLICIT SEQUENCE
		{
		packed			[0] IMPLICIT BOOLEAN DEFAULT FALSE,
		components		[1] IMPLICIT SEQUENCE OF SEQUENCE
			{
			componentName		[0] IMPLICIT Identifier OPTIONAL,
			componentType		[1] TypeSpecification
			}
		},
	
	-- Simple Type --------------------- Size ---------
	boolean			[3] IMPLICIT NULL,
	bit-string		[4] IMPLICIT Integer32,
	integer			[5] IMPLICIT Unsigned8,
	unsigned		[6] IMPLICIT Unsigned8,
	floating-point		[7] IMPLICIT SEQUENCE {
		format-width		Unsigned8,
		exponent-width          Unsigned8
		},
	real			[8] IMPLICIT SEQUENCE {
		base			[0] IMPLICIT INTEGER (2|10),
		exponent		[1] IMPLICIT INTEGER OPTIONAL,
		mantissa		[2] IMPLICIT INTEGER OPTIONAL
		},
	octet-string		[9] IMPLICIT Integer32,
	visible-string		[10] IMPLICIT Integer32,
	generalized-time	[11] IMPLICIT NULL,
	binary-time		[12] IMPLICIT BOOLEAN,
	bcd			[13] IMPLICIT Unsigned8,
	objId			[15] IMPLICIT NULL
	}
*/


/*
	array			[1] IMPLICIT SEQUENCE
		{
		packed			[0] IMPLICIT BOOLEAN DEFAULT FALSE,
		numberOfElements	[1] IMPLICIT Unsigned32,
		elementType		[2] TypeSpecification,
		},
*/

extern ASN1_TBL_ENTRY type_spec[];
extern ASN1_TBL_ENTRY structure_type[];


static ASN1_TBL_ENTRY float_type1[]={
		{(ASN1H_UNI | ASN1_INT),NOETT,NOHF,"format-width: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_UNI | ASN1_INT),NOETT,NOHF,"exponent-width: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY array_type[]={
		{(ASN1H_CTX | 0),NOETT,NOHF,"packed: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 1),NOETT,NOHF,"numberOfElements: ",ASN1H_UNSIGNED_DISPLAY, NULL},
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),&ett_typespec,NOHF,"",ASN1H_TBL_SET, type_spec},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
	structure		[2] IMPLICIT SEQUENCE
		{
		packed			[0] IMPLICIT BOOLEAN DEFAULT FALSE,
		components		[1] IMPLICIT SEQUENCE OF SEQUENCE
			{
			componentName		[0] IMPLICIT Identifier OPTIONAL,
			componentType		[1] TypeSpecification
			}
		},
*/

static ASN1_TBL_ENTRY struct_type2[]={
		{ASN1H_SEQ_TAG,				NOETT,NOHF,"",ASN1H_DONOT_DISPLAY, NULL},
		{(ASN1H_CTX | 0),			NOETT,NOHF,"componentName: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_typespec,NOHF,"typespecification", ASN1H_TBL_SET, type_spec},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY structure_type[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"packed: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_components,NOHF,"components",ASN1H_TBL_SET, struct_type2}, 
/*		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_components,NOHF,"components",ASN1H_TBL_SEQ, struct_type2}, */
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY type_spec[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),	&ett_obj,NOHF,"TypeName",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_array,NOHF,"Array",ASN1H_TBL_SEQ, array_type},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 2),	&ett_structure,NOHF,"structure",ASN1H_TBL_SET, structure_type},
		{(ASN1H_CTX | 3 ),			NOETT,NOHF,"Boolean",0, NULL},
		{(ASN1H_CTX | 4 ),			&ett_bitstring,NOHF,"BitString numbits=",ASN1H_MMS_TYPE_LENGTH,NULL },
		{(ASN1H_CTX | 5 ),			NOETT,NOHF,"SignedInteger precision=",ASN1H_UNSIGNED_DISPLAY,NULL },
		{(ASN1H_CTX | 6 ),			NOETT,NOHF,"UnsignedInteger precision=",ASN1H_UNSIGNED_DISPLAY,NULL },
		{(ASN1H_CTX | 7 | ASN1H_CONSTRUCT),	NOETT,NOHF,"FloatingPoint",ASN1H_TBL_SEQ,float_type1 },
		{(ASN1H_CTX | 8 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Real (depricated - should not be used)",0, NULL},
		{(ASN1H_CTX | 9),			&ett_octettstring,NOHF,"Octetstring numOctets=",ASN1H_MMS_TYPE_LENGTH,NULL },
		{(ASN1H_CTX | 10),			&ett_string,NOHF,"VisibleString numOctets=",ASN1H_MMS_TYPE_LENGTH,NULL },
		{(ASN1H_CTX | 11),			&ett_gentime,NOHF,"GeneralizedTime",0, NULL},
		{(ASN1H_CTX | 12),			&ett_bintime,NOHF,"BinaryTime SixOctets=",ASN1H_BOOLEAN_DISPLAY,NULL },
		{(ASN1H_CTX | 13),			&ett_bcd,NOHF,"BCD numOctets=",ASN1H_MMS_TYPE_LENGTH,NULL },
		{(ASN1H_CTX |  14),			&ett_octettstring,NOHF,"BOOLEAN Array ", ASN1H_MMS_TYPE_LENGTH,NULL},
		{(ASN1H_CTX | 15),			&ett_oid,NOHF,"ObjectID",0, NULL},
		{(ASN1H_CTX |  16),			&ett_unicode,NOHF,"Unicode",ASN1H_MMS_TYPE_LENGTH,NULL},
		{(ASN1H_CTX | 17),			&ett_utctime,NOHF,"UTCTime",0, NULL},
		{(ASN1H_CTX | 18),			&ett_unicode,NOHF,"Unicode numCharacters=",ASN1H_MMS_TYPE_LENGTH,NULL },
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
GetVariableAccessAttributes-Response ::= SEQUENCE
	{
	mmsDeletable		[0] IMPLICIT BOOLEAN,
	address			[1] Address OPTIONAL,
	typeSpecification	[2] TypeSpecification
	}
*/
static ASN1_TBL_ENTRY getVAA_resp[]={
		{(ASN1H_CTX | 0),			NOETT,NOHF,"MMSDeletable",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_addrspec,NULL,"Address",0,address_spec },
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),	&ett_typespec,NOHF,"TypeSpecification",ASN1H_TBL_SET, type_spec},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
DefineNamedVariable-Request ::= SEQUENCE
	{
	variableName		[0] ObjectName,
	address			[1] Address,
	typeSpecification	[2] TypeSpecification OPTIONAL
	}
*/

static ASN1_TBL_ENTRY defNameVar_req[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),&ett_obj,NOHF,"ObjectName",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),&ett_addrspec,NOHF,"Address",0,address_spec },
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),&ett_typespec,NOHF,"TypeSpecification",ASN1H_TBL_SET, type_spec},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
DeleteVariableAccess-Request ::= SEQUENCE
	{
	scopeOfDelete		[0] IMPLICIT INTEGER
		{
		specific		(0),
		aa-specific		(1),
		domain			(2),
		vmd			(3) 
		} DEFAULT specific,
	listOfName		[1] IMPLICIT SEQUENCE OF ObjectName OPTIONAL,
	domainName		[2] IMPLICIT Identifier OPTIONAL
	}

DeleteVariableAccess-Response ::= SEQUENCE
	{
	numberMatched	[0] IMPLICIT Unsigned32,
	numberDeleted	[1] IMPLICIT Unsigned32
	}
*/

static ASN1_TBL_ENTRY delVarAcc_resp[]={
		{(ASN1H_CTX | 0 ),NOETT,NOHF,"NumberMatched: ",ASN1H_INTEGER_DISPLAY, NULL},
		{(ASN1H_CTX | 1 ),NOETT,NOHF,"NumberDeleted: ",ASN1H_INTEGER_DISPLAY,NULL },
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

static ENUM_LOOKUP scope_of_delete1[4]= {
   {0,"Specific"},
   {1, "aa-specific"},
   {2, "domain"},
   {3, "vmd"}};

static ENUM_XLAT scope_of_delete= {4 , scope_of_delete1};

static ASN1_TBL_ENTRY delVarAcc_req[]={
		{(ASN1H_CTX | 0 ),NOETT,NOHF,"ScopeOfDelete",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&scope_of_delete},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),&ett_addrspec,NOHF,"Address",0,address_spec },
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),&ett_typespec,NOHF,"TypeSpecification",ASN1H_TBL_SET, type_spec},
		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
};

/*
Input-Request ::= SEQUENCE
	{
	operatorStationName	[0] IMPLICIT Identifier,
	echo			[1] IMPLICIT BOOLEAN DEFAULT TRUE,
	listOfPromptData	[2] IMPLICIT SEQUENCE OF VisibleString OPTIONAL,
	inputTimeOut		[3] IMPLICIT Unsigned32 OPTIONAL
	}


Input-Response ::= VisibleString	-- Input String
*/
static ASN1_TBL_ENTRY prompt_data[]={
		{(ASN1H_UNI | ASN1_VISSTR ),		NOETT,NOHF,"",ASN1H_VSTRING_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY input_req[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"OperatorStationName: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 1 ),			NOETT,NOHF,"Echo: ", ASN1H_BOOLEAN_DISPLAY,NULL },
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),	NOETT,NOHF,"PromptData",0, prompt_data},
		{(ASN1H_CTX | 3 ),			NOETT,NOHF,"InputTimeout",ASN1H_INTEGER_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
Output-Request ::= SEQUENCE
	{
	operatorStationName	[0] IMPLICIT Identifier,
	listOfOutputData	[1] IMPLICIT SEQUENCE OF VisibleString
	}


Output-Response ::= NULL
*/
static ASN1_TBL_ENTRY output_req[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"OperatorStationName: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),	NOETT,NOHF,"ListOfOutputData",0, prompt_data},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*******************************************************/
/*         DefineNamedVarlist Request				   */

/*
DefineNamedVariableList-Request ::= SEQUENCE
	{
	variableListName	ObjectName,
	listOfVariable		[0] IMPLICIT SEQUENCE OF SEQUENCE
		{
		variableSpecification 	VariableSpecification,
		alternateAccess		[5] IMPLICIT AlternateAccess OPTIONAL
		}
	}

*/

static ASN1_TBL_ENTRY def_varlist_req[]={
		{(ASN1H_UNI | ASN1H_CONSTRUCT | ASN1_SEQ),	NOETT,NOHF,NULL, ASN1H_TBL_SEQ ,NULL},
		{(ASN1H_CTX | 0),				NOETT,NOHF,"VMD Specific ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),		&ett_domainspec,NOHF,"Domain Specific ",ASN1H_TBL_SEQ,NULL},
		{(ASN1H_UNI | ASN1_VISSTR),			&ett_domainname,&hf_mms_domainname,"Domain Name: ",ASN1H_VSTRING_DISPLAY, NULL},
		{(ASN1H_UNI | ASN1_VISSTR),			&ett_item,&hf_mms_itemname,"Item Name: ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 2),				NOETT,NOHF,"AA-Specific ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 0),		&ett_varlist,NOHF,"List of Variable",ASN1H_TBL_SET,list_of_var},
		{ASN1H_END_OF_TABLE				,NOETT,NOHF,NULL,0,NULL}
};


/************************************************************************/
/*     Get Named Varlist Resp											*/

/*
GetNamedVariableListAttributes-Response ::= SEQUENCE
	{
	mmsDeletable		[0] IMPLICIT BOOLEAN,
	listOfVariable		[1] IMPLICIT SEQUENCE OF SEQUENCE 
		{
		variableSpecification 		VariableSpecification,
		alternateAccess			[5] IMPLICIT AlternateAccess OPTIONAL
		}
	}
*/

static ASN1_TBL_ENTRY get_varlist_attrib_resp[]={
		{(ASN1H_CTX | 0), 			NOETT,NOHF,"MMS Deletable: ",ASN1H_BOOLEAN_DISPLAY,NULL },
		{(ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_varlist,NOHF,"List of Variable",ASN1H_TBL_SET,list_of_var},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/************************************************************************/
/*     DeleteNamedVariableList Request									*/

/*
 DeleteNamedVariableList-Request ::= SEQUENCE
	{
	scopeOfDelete		[0] IMPLICIT INTEGER
		{
		specific			(0),
		aa-specific			(1),
		domain				(2),
		vmd				(3) 
		} DEFAULT specific,
	listOfVariableListName	[1] IMPLICIT SEQUENCE OF ObjectName OPTIONAL,
	domainName			[2] IMPLICIT Identifier OPTIONAL
	}
*/

static ASN1_TBL_ENTRY del_varlist_req[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"ScopeOfDelete",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry*)&scope_of_delete},
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 1), 	&ett_obj,NOHF,"Object Name",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CTX | 2),			&ett_domainname,&hf_mms_domainname,"Domain Name: ", ASN1H_VSTRING_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
GetDomainAttributes-Request ::= Identifier	-- Domain Name

DomainState ::= INTEGER
	{
	non-existent	(0),
	loading		(1),
	ready		(2),
	in-use		(3),
	complete	(4),
	incomplete	(5),
	d1		(7),
	d2		(8),
	d3		(9),
	d4		(10),
	d5		(11),
	d6		(12),
	d7		(13),
	d8		(14),
	d9		(15)
	}
*/
static ENUM_LOOKUP domainstate1[15]= {
   {0,"non-existent"},
   {1, "loading"},
   {2, "ready"},
   {3, "in-use"},
   {4, "complete"},
   {5, "incomplete"},
   {7, "d1"},
   {8, "d2"},
   {9, "d3"},
   {10, "d4"},
   {11, "d5"},
   {12, "d6"},
   {13, "d7"},
   {14, "d8"},
   {15, "d9"},
   };

static ENUM_XLAT domainState= {15 , domainstate1};

/*
GetDomainAttributes-Response ::= SEQUENCE
	{
	listOfCapabilities		[0] IMPLICIT SEQUENCE OF VisibleString,
	state				[1] IMPLICIT DomainState,
	mmsDeletable			[2] IMPLICIT BOOLEAN,
	sharable			[3] IMPLICIT BOOLEAN,
	listOfProgramInvocations	[4] IMPLICIT SEQUENCE OF Identifier, 	-- PI Names
	uploadInProgress		[5] IMPLICIT Integer8
	}
*/
static ASN1_TBL_ENTRY getDomAttr_resp[]={
		{(ASN1H_CTX | 0 |ASN1H_CONSTRUCT ),	NOETT,NOHF,"ListOfDomainCapabilities: ",0, prompt_data},
		{(ASN1H_CTX | 1 ),			NOETT,NOHF,"DomainState: ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry *)&domainState},
		{(ASN1H_CTX | 2 ),			NOETT,NOHF,"MMSDeletable: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 3 ),			NOETT,NOHF,"Shareable: ",ASN1H_BOOLEAN_DISPLAY, NULL},
		{(ASN1H_CTX | 4 | ASN1H_CONSTRUCT),	NOETT,NOHF,"ListOfProgramInvocations",0, prompt_data},
		{(ASN1H_CTX | 5 ),			NOETT,NOHF,"UploadInProgress: ",ASN1H_INTEGER_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/***********************************************************************/
/*   			read journal resp									   */

/*
EC-State ::= INTEGER
	{
	disabled		(0),
	idle			(1),
	active			(2)
	}
*/
static ENUM_LOOKUP EC_State_text[3]= {
   {0,"disabled"},
   {1, "idle"},
   {2, "active"},
   };

static ENUM_XLAT EC_State= {3 , EC_State_text};

/*
EntryContent ::= SEQUENCE
	{
	occurenceTime		[0] IMPLICIT TimeOfDay,
	additionalDetail	[1] JOU-Additional-Detail OPTIONAL,
		-- additionalDetail shall be omitted
		-- from abstract syntax defined in this standard
	entryForm		 CHOICE
		{
		data			[2] IMPLICIT SEQUENCE
			{
			event			[0] IMPLICIT SEQUENCE
				{
				eventConditionName	[0] ObjectName,
				currentState		[1] IMPLICIT EC-State
				} OPTIONAL,
			listOfVariables	[1] IMPLICIT SEQUENCE OF SEQUENCE
				{
				variableTag		[0] IMPLICIT VisibleString,
				valueSpecification	[1] Data
				} OPTIONAL
			},
		annotation		[3] IMPLICIT VisibleString
		}
	}
*/
static ASN1_TBL_ENTRY journalevent_data[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	NOETT,NOHF,"eventConditionName ",ASN1H_TBL_SEQ,object_name},
		{(ASN1H_CTX | 1 ),			NOETT,NOHF,"currentState: ",ASN1H_INTEGER_ENUM_DISPLAY, (struct asn1h_tbl_entry *)&EC_State},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY journalevent_vars[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"variableTag ",ASN1H_VSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	&ett_datatbl,NOHF,"ValueSpecification",ASN1H_TBL_SET, data_tbl},
		{ASN1H_END_OF_TABLE	,		NOETT,NOHF,NULL,0,NULL}
};
																			 
static ASN1_TBL_ENTRY journalentry_data[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	NOETT,NOHF,"event ",ASN1H_TBL_SEQ,journalevent_data},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	NOETT,NOHF,"listofVariables ",0,NULL},
		{ASN1H_SEQ_TAG,				NOETT,NOHF,"",ASN1H_TBL_SEQ,journalevent_vars},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY entryContent[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"Occurence Time: ",ASN1H_MMS_BTIME_DISPLAY,NULL},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Additional Detail ",0,NULL},
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Data ",ASN1H_TBL_SET,journalentry_data},
		{(ASN1H_CTX | 3 ),			NOETT,NOHF,"Annotation:  ",ASN1H_VSTRING_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/*
JournalEntry ::= SEQUENCE
	{
	entryIdentifier		[0] IMPLICIT OCTET STRING,
	originatingApplication		[1] ApplicationReference,
	entryContent			[2] IMPLICIT EntryContent
	}
*/

static ASN1_TBL_ENTRY journalEntry[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"EntryIdentifier: ",ASN1H_OCTETSTRING_DISPLAY,NULL},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT ),	NOETT,NOHF,"originatingApplication: ",0,NULL},
		{ASN1H_SEQ_TAG,				NOETT,NOHF,"",ASN1H_TBL_SEQ,applicationReference},
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT ),	NOETT,NOHF,"Entry Content ",ASN1H_TBL_SEQ,entryContent},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/*
ReadJournal-Response ::= SEQUENCE
	{
	listOfJournalEntry		[0] IMPLICIT SEQUENCE OF JournalEntry,
	moreFollows			[1] IMPLICIT BOOLEAN DEFAULT FALSE
	}
*/

static ASN1_TBL_ENTRY readjournal_resp[]={
		{(ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	NOETT,NOHF,"ListOfJournalEntry ",0,NULL},
		{ASN1H_SEQ_TAG,				NOETT,NOHF,"JournalEntry",ASN1H_TBL_SET,journalEntry},
		{(ASN1H_CTX | 1 ),			NOETT,NOHF,"More Follows: ",ASN1H_BOOLEAN_DISPLAY,NULL},

		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};



/***********************************************************************/
/*   			read journal req									   */


/*
ReadJournal-Request ::= SEQUENCE
	{
	journalName			[0] ObjectName,
	rangeStartSpecification	[1] CHOICE
		{
		startingTime			[0] IMPLICIT TimeOfDay,
		startingEntry			[1] IMPLICIT OCTET STRING
		} OPTIONAL,
	rangeStopSpecification	[2] CHOICE
		{
		endingTime			[0] IMPLICIT TimeOfDay,
		numberOfEntries		[1] IMPLICIT Integer32
		} OPTIONAL,
	listOfVariables		[4] IMPLICIT SEQUENCE OF VisibleString OPTIONAL,
	entryToStartAfter	[5] IMPLICIT SEQUENCE
		{
		timeSpecification		[0] IMPLICIT TimeOfDay,
		entrySpecification		[1] IMPLICIT OCTET STRING
		}
	}
*/
static ASN1_TBL_ENTRY rangeSpec[]={
		{(ASN1H_CTX | 0), 			NOETT,NOHF,"Time",ASN1H_MMS_BTIME_DISPLAY, NULL},
		{(ASN1H_CTX | 1), 			NOETT,NOHF,"Entry",ASN1H_OCTETSTRING_DISPLAY, NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY readjournal_req[]={
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 0),	&ett_obj,NOHF,"Object Name",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Range Start Specifictaion ",ASN1H_TBL_SEQ,rangeSpec},
		{(ASN1H_CTX | 2 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Range Stop Specifictaion ",ASN1H_TBL_SEQ,rangeSpec},
		{(ASN1H_CTX | 4 | ASN1H_CONSTRUCT),	&ett_varlist,NOHF,"List of Variables ",ASN1H_TBL_SET,domain_names},
		{(ASN1H_CTX | 5 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Entry to Start After ",ASN1H_TBL_SEQ,rangeSpec},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/***********************************************************************/
/*   			initialize journal req									   */

/*
InitializeJournal-Request ::= SEQUENCE
	{
	journalName			[0] ObjectName,
	limitSpecification		[1] IMPLICIT SEQUENCE
		{
		limitingTime			[0] IMPLICIT TimeOfDay,
		limitingEntry			[1] IMPLICIT OCTET STRING OPTIONAL
		} OPTIONAL
	}
*/
static ASN1_TBL_ENTRY initjournal_req[]={
		{(ASN1H_CONSTRUCT | ASN1H_CTX | 0), 	&ett_obj,NOHF,"Object Name",ASN1H_TBL_SEQ, object_name},
		{(ASN1H_CTX | 1 | ASN1H_CONSTRUCT),	NOETT,NOHF,"Limit Specifictaion ",ASN1H_TBL_SEQ,rangeSpec},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/***********************************************************************/
/*   			report journal status resp							   */

/*
ReportJournalStatus-Response ::= SEQUENCE
	{
	currentEntries		[0] IMPLICIT Unsigned32,
	mmsDeletable		[1] IMPLICIT  BOOLEAN
	}	
*/
static ASN1_TBL_ENTRY reportJournalStatus_resp[]={
		{(ASN1H_CTX | 0 ),			NOETT,NOHF,"currentEntries: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{(ASN1H_CTX | 1 ),			NOETT,NOHF,"MMS Deletable: ",ASN1H_BOOLEAN_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

/***********************************************************************/
/*   get_capability decoding										   */
/*
GetCapabilityList-Request ::= SEQUENCE {
	continueAfter	VisibleString OPTIONAL
	}

GetCapabilityList-Response ::= SEQUENCE {
	listOfCapabilities	[0] IMPLICIT SEQUENCE OF VisibleString,
	moreFollows		[1] IMPLICIT BOOLEAN DEFAULT TRUE
	}
*/

static ASN1_TBL_ENTRY getcap_string[]={
		{(ASN1H_UNI | ASN1_VISSTR),		NOETT,NOHF,NULL,	ASN1H_VSTRING_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE,			NOETT,NOHF,NULL,0,NULL}
};


static ASN1_TBL_ENTRY get_capability_resp[]={
		{(ASN1H_CTX | ASN1H_CONSTRUCT),		NOETT,NOHF,"ListofCapabilities ",ASN1H_TBL_SEQ,NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"MoreFollows ",ASN1H_BOOLEAN_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE,			NOETT,NOHF,NULL,0,NULL}
};

static ASN1_TBL_ENTRY getcap_req[]={
		{(ASN1H_UNI | ASN1_VISSTR),		NOETT,NOHF,"ContinueAfter: ",ASN1H_VSTRING_DISPLAY,NULL},
		{ASN1H_END_OF_TABLE,			NOETT,NOHF,NULL,0,NULL}
};



static ASN1_TBL_ENTRY mms_conf_resp[]={
		{ (ASN1H_UNI | ASN1_INT ),		NOETT,&hf_mms_invoke_id,"InvokeID: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 0 | ASN1H_CONSTRUCT),	&ett_service,NOHF,"Status",ASN1H_TBL_SEQ,status_resp},
		{ (ASN1H_CTX | 1 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetNameList", ASN1H_TBL_SEQ,getNameList_resp},
		{ (ASN1H_CTX | 2 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Identify", ASN1H_TBL_SEQ,identify_resp},
		{ (ASN1H_CTX | 3 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Rename (not currently decoded)", 0,NULL},
		{ (ASN1H_CTX | 4 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Read",0,read_resp},
		{ (ASN1H_CTX | 5 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Write",0,write_resp},
		{ (ASN1H_CTX | 6 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetVarAccessAttributes", ASN1H_TBL_SEQ,getVAA_resp},
		{ (ASN1H_CTX | 7 ), 			&ett_service,NOHF,"DefineNamedVariable", 0,NULL},
		{ (ASN1H_CTX | 8 ), 			&ett_service,NOHF,"DefineScatteredAccess (should not be in use)", 0,NULL},
		{ (ASN1H_CTX | 9 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetScatteredAccessAttributes (not decoded)",0,mms_conf_resp},
		{ (ASN1H_CTX | 10 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"DeleteVariableAccess", ASN1H_TBL_SEQ,delVarAcc_resp},
		{ (ASN1H_CTX | 11 ), 			&ett_service,NOHF,"DefineNamedVariableList", 0,NULL},
		{ (ASN1H_CTX | 12 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetNamedVariableListAttributes", ASN1H_TBL_SEQ,get_varlist_attrib_resp},
		{ (ASN1H_CTX | 13 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"DeleteNamedVariableList", ASN1H_TBL_SEQ,delVarAcc_resp},
		{ (ASN1H_CTX | 14 ), 			&ett_service,NOHF,"DefineNamedType", 0,NULL},
		{ (ASN1H_CTX | 15 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetNamedTypeAttributes (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 16 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"DeleteNamedType", ASN1H_TBL_SEQ,delVarAcc_resp},
		{ (ASN1H_CTX | 17 ), 			&ett_service,NOHF,"Input: ", ASN1H_VSTRING_DISPLAY,NULL},
		{ (ASN1H_CTX | 18 ), 			&ett_service,NOHF,"Output ", 0,NULL},
		{ (ASN1H_CTX | 19 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"TakeControl (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 20 ), 			&ett_service,NOHF,"RelinquishControl", 0,NULL},
		{ (ASN1H_CTX | 21 ), 			&ett_service,NOHF,"DefineSemaphore (not decoded) ", 0,NULL},
 		{ (ASN1H_CTX | 37 | ASN1H_CONSTRUCT),	&ett_service,NOHF,"GetDomainAttributes", ASN1H_TBL_SEQ,getDomAttr_resp},
		{ (ASN1H_CTX | 38), 			&ett_service,NOHF,"CreateProgramInvocation", 0,NULL},
		{ (ASN1H_CTX | 39), 			&ett_service,NOHF,"DeleteProgramInvocation", 0,NULL},
		{ (ASN1H_CTX | 40), 			&ett_service,NOHF,"Start", 0,NULL},
		{ (ASN1H_CTX | 41), 			&ett_service,NOHF,"Stop", 0,NULL},
		{ (ASN1H_CTX | 42), 			&ett_service,NOHF,"Reset", 0,NULL},
		{ (ASN1H_CTX | 43), 			&ett_service,NOHF,"Resume", 0,NULL},
		{ (ASN1H_CTX | 44), 			&ett_service,NOHF,"Kill", 0,NULL},
		{ (ASN1H_CTX | 45 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetProgramInvocationAttributes ", ASN1H_TBL_SEQ,getpi_attribs_resp},
		{ (ASN1H_CTX | 46 ), 			&ett_service,NOHF,"ObtainFile", 0,NULL},
		{ (ASN1H_CTX | 47 ), 			&ett_service,NOHF,"DefineEventCondition", 0,NULL},
		{ (ASN1H_CTX | 48 ), 			&ett_service,NOHF,"DeleteEventCondition", ASN1H_INTEGER_DISPLAY,NULL},
		{ (ASN1H_CTX | 49 |ASN1H_CONSTRUCT ),	&ett_service,NOHF,"GetEventConditionAttributes (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 50 |ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"ReportEventConditionStatus (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 51  ), 			&ett_service,NOHF,"AlterEventConditionMonitoring", 0,NULL},
		{ (ASN1H_CTX | 52  ), 			&ett_service,NOHF,"TriggerEvent", 0,NULL},
		{ (ASN1H_CTX | 53  ), 			&ett_service,NOHF,"DefineEventAction", 0,NULL},
		{ (ASN1H_CTX | 54  ), 			&ett_service,NOHF,"DeleteEventAction", ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 55 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetEventActionAttributes (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 56  ), 			&ett_service,NOHF,"ReportEventActionStatus", ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 57  ), 			&ett_service,NOHF,"DefineEventEnrollment", 0,NULL},
		{ (ASN1H_CTX | 58  ), 			&ett_service,NOHF,"DeleteEventEnrollment", ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 59 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"AlterEventEnrollment  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 60 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"ReportEventEnrollmentStatus  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 61 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetEventEnrollmentAttributes  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 62  ), 			&ett_service,NOHF,"AcknowledgeEventNotification", 0,NULL},
		{ (ASN1H_CTX | 63 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetAlarmSummary  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 64 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetAlarmEnrollmentSummary  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 65 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"ReadJournal", ASN1H_TBL_SEQ,readjournal_resp},
		{ (ASN1H_CTX | 66 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"WriteJournal", 0,NULL},
		{ (ASN1H_CTX | 67 ), 			&ett_service,NOHF,"InitializeJournal", 0,NULL},
		{ (ASN1H_CTX | 68 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"ReportJournalStatus", ASN1H_TBL_SEQ,reportJournalStatus_resp},
		{ (ASN1H_CTX | 69 ), 			&ett_service,NOHF,"CreateJournal", 0,NULL},
		{ (ASN1H_CTX | 70 ), 			&ett_service,NOHF,"DeleteJournal", 0,NULL},
		{ (ASN1H_CTX | 71 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetCapabilityList", ASN1H_TBL_SET, NULL},
		{ (ASN1H_CTX | 72 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"FileOpen", ASN1H_TBL_SEQ,file_open_resp},
		{ (ASN1H_CTX | 73 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"FileRead", ASN1H_TBL_SEQ,file_read_resp},
		{ (ASN1H_CTX | 74 ), 			&ett_service,NOHF,"FileClose", 0,NULL},
		{ (ASN1H_CTX | 75 ), 			&ett_service,NOHF,"FileRename", 0,NULL},
		{ (ASN1H_CTX | 76 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"FileDelete", 0,NULL},
		{ (ASN1H_CTX | 77 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"FileDirectory", ASN1H_TBL_SEQ,file_directory_resp},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
		};

static ASN1_TBL_ENTRY mms_conf_req[]={
		{ (ASN1H_UNI | ASN1_INT ),		NOETT,&hf_mms_invoke_id,"InvokeID: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX |  0), 			&ett_service,NOHF,"Status ExtendedStatus=", ASN1H_BOOLEAN_DISPLAY,NULL},	/*since status is just a BOOLEAN, no further decode */
		{ (ASN1H_CTX | ASN1H_CONSTRUCT |  1), 	&ett_service,NOHF,"GetNameList", ASN1H_TBL_SEQ,getNameList_req},		/*since status is just a BOOLEAN, no further decode */
		{ (ASN1H_CTX |  2), 			&ett_service,NOHF,"Identify", ASN1H_NULL_DISPLAY,NULL},			/*since status is just a BOOLEAN, no further decode */
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 3 ), 	&ett_service,NOHF,"Rename (not currently decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 4 ), 	&ett_service,NOHF,"Read",0,read_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 5 ), 	&ett_service,NOHF,"Write", ASN1H_TBL_SEQ,write_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 6 ),	&ett_service,NOHF,"GetVariableAccessAttributes",0,getVAA_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 7 ), 	&ett_service,NOHF,"DefineNamedVariable",0,defNameVar_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 8 ), 	&ett_service,NOHF,"ScatteredAccess (not currently decoded , should not be used)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 9 ), 	&ett_service,NOHF,"ScatteredAccessAttributes (not currently decoded , should not be used)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 10 ), 	&ett_service,NOHF,"DeleteVariableAccess", ASN1H_TBL_SEQ,delVarAcc_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 11 ), 	&ett_service,NOHF,"DefineNamedVariableList", ASN1H_TBL_SEQ,def_varlist_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 12 ), 	&ett_service,NOHF,"GetNamedVariableListAttributes", ASN1H_TBL_SEQ,object_name},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 13 ), 	&ett_service,NOHF,"DeleteNamedVariableList", ASN1H_TBL_SEQ,del_varlist_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 14 ), 	&ett_service,NOHF,"DefineNamedType (not decoded)",0, NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 15 ), 	&ett_service,NOHF,"GetNamedTypeAttributes (not decoded)",0, NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 16 ), 	&ett_service,NOHF,"DeleteNamedType (not decoded)",0, NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 17 ), 	&ett_service,NOHF,"Input", ASN1H_TBL_SEQ,input_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 18 ), 	&ett_service,NOHF,"Ouput", ASN1H_TBL_SEQ,output_req},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 19 ), 	&ett_service,NOHF,"TakeControl (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 20 ), 	&ett_service,NOHF,"RelinquishControl (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 21 ), 	&ett_service,NOHF,"DefineSemaphore (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 22 ), 	&ett_service,NOHF,"DeleteSemaphore (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 23 ), 	&ett_service,NOHF,"ReportSemaphoreStatus (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 24 ), 	&ett_service,NOHF,"ReportPoolSemaphoreStatus (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 25 ), 	&ett_service,NOHF,"ReportSemaphoreEntryStatus (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 26 ), 	&ett_service,NOHF,"InitiateDownloadSequence (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 27 ), 			&ett_service,NOHF,"DownloadSegment: ", ASN1H_VSTRING_DISPLAY,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 28 ), 	&ett_service,NOHF,"TerminateDownloadSequence (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 29 ), 	&ett_service,NOHF,"InitiateUploadSequence (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 30 ), 			&ett_service,NOHF,"UploadSegment: ", ASN1H_INTEGER_DISPLAY,NULL},
		{ (ASN1H_CTX | 31 ), 			&ett_service,NOHF,"TerminateUploadSegment: ", ASN1H_INTEGER_DISPLAY,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 32), 	&ett_service,NOHF,"RequestDomainDownload (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 33),	&ett_service,NOHF,"RequestDomainUpload (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 34), 	&ett_service,NOHF,"LoadDomainContent (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 35), 	&ett_service,NOHF,"StoreDomainContent (not decoded)", 0,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 36), 	&ett_service,NOHF,"DeleteDomain (not decoded)", 0,NULL},
		{ (ASN1H_CTX |  37 ), 			&ett_service,NOHF,"GetDomainAttributes Domain: ", ASN1H_VSTRING_DISPLAY,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 38), 	&ett_service,NOHF,"CreateProgramInvocation", ASN1H_TBL_SEQ,createpi_req},
		{ (ASN1H_CTX | 39), 			&ett_service,NOHF,"DeleteProgramInvocation: ", ASN1H_VSTRING_DISPLAY,NULL},
		{ (ASN1H_CTX | 40 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Start: ", ASN1H_TBL_SEQ,startpi_req},
		{ (ASN1H_CTX | 41 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Stop: ", ASN1H_TBL_SET,stoppi_req},
		{ (ASN1H_CTX | 42 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Resume: ", ASN1H_TBL_SEQ,startpi_req},
		{ (ASN1H_CTX | 43 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Reset: ", ASN1H_TBL_SET,stoppi_req},
		{ (ASN1H_CTX | 44 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"Kill: ", ASN1H_TBL_SET,stoppi_req},
		{ (ASN1H_CTX | 45 ), 			&ett_service,NOHF,"GetProgramInvocationAttributes: ", ASN1H_VSTRING_DISPLAY,NULL},
		{ (ASN1H_CTX | 46 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"ObtainFile ", ASN1H_TBL_SEQ,obtainfile_req},
		{ (ASN1H_CTX | 47 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"DefineEventCondition (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 48 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"DeleteEventCondition (not decoded)", ASN1H_INTEGER_DISPLAY,NULL},
		{ (ASN1H_CTX | 49 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"GetEventConditionAttributes (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 50 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"ReportEventConditionStatus (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 51 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"AlterEventConditionMonitoring (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 52 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"TriggerEvent (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 53 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"DefineEventAction (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 54 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"DeleteEventAction (not decoded)", ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 55 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetEventActionAttributes (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 56 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"ReportEventActionStatus (not decoded)", ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 57 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"DefineEventEnrollment (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 58 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"DeleteEventEnrollment (not decoded)", ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | 59 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"AlterEventEnrollment  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 60 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"ReportEventEnrollmentStatus  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 61 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetEventEnrollmentAttributes  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 62 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"AcknowledgeEventNotification (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 63 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetAlarmSummary  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 64 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetAlarmEnrollmentSummary  (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 65 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"ReadJournal", ASN1H_TBL_SEQ,readjournal_req},
		{ (ASN1H_CTX | 66 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"WriteJournal (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 67 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"InitializeJournal", ASN1H_TBL_SEQ,initjournal_req},
		{ (ASN1H_CTX | 68 | ASN1H_CTX ), 	&ett_service,NOHF,"ReportJournalStatus",ASN1H_TBL_SEQ, object_name},
		{ (ASN1H_CTX | 69 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"CreateJournal (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 70 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"DeleteJournal (not decoded)", 0,NULL},
		{ (ASN1H_CTX | 71 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"GetCapabilityList", ASN1H_TBL_SEQ, getcap_req},
		{ (ASN1H_CTX | 72 | ASN1H_CONSTRUCT), 	&ett_service,NOHF,"File Open", ASN1H_TBL_SEQ, file_open_req},
		{ (ASN1H_CTX | 73 ), 			&ett_service,NOHF,"File Read Request FRSMID = ", ASN1H_INTEGER_DISPLAY, NULL},
		{ (ASN1H_CTX | 74 ), 			&ett_service,NOHF,"File Close Request FRSMID = ", ASN1H_INTEGER_DISPLAY, NULL},
		{ (ASN1H_CTX | 75 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"File Rename  ", ASN1H_TBL_SEQ, file_rename_req},
		{ (ASN1H_CTX | 77 | ASN1H_CONSTRUCT ), 	&ett_service,NOHF,"File Directory  ", ASN1H_TBL_SEQ, file_directory_req},
 		{ASN1H_END_OF_TABLE	,NOETT,NOHF,NULL,0,NULL}
		};		 


/********************************************************************/
/*			Confired Error PDU										*/
/********************************************************************/

/*
Confirmed-ErrorPDU ::= SEQUENCE 
	{
	invokeID 		[0]	IMPLICIT Unsigned32,
	modifierPosition	[1]	IMPLICIT Unsigned32 OPTIONAL,
	serviceError		[2]	IMPLICIT ServiceError
	}
*/

static ASN1_TBL_ENTRY mms_conf_error[]={
		{(ASN1H_CTX | 0),			NOETT,&hf_mms_invoke_id,"Invoke ID: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{(ASN1H_CTX | 1),			NOETT,NOHF,"Modifier Position: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 2), 	&ett_service_error,NOHF,"Service Error: ", ASN1H_TBL_SEQ,mms_service_error},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};


/********************************************************************/
/*			Cancel Error PDU										*/
/********************************************************************/
/*
Cancel-ErrorPDU ::= SEQUENCE
	{
	originalInvokeID	[0] IMPLICIT Unsigned32,
	serviceError		[1] IMPLICIT ServiceError
	}
*/
static ASN1_TBL_ENTRY mms_cancel_err[]={
		{(ASN1H_CTX | 0),			NOETT,&hf_mms_invoke_id,"Invoke ID: ",ASN1H_UNSIGNED_DISPLAY,NULL},
		{ (ASN1H_CTX | ASN1H_CONSTRUCT | 1),	&ett_service_error,NOHF,"Service Error: ", ASN1H_TBL_SEQ,mms_service_error},
		{ASN1H_END_OF_TABLE			,NOETT,NOHF,NULL,0,NULL}
};

#define MMS_CONF_REQ_PDU	(ASN1H_CONSTRUCT | ASN1H_CTX | 0)
#define MMS_CONF_RESP_PDU	(ASN1H_CONSTRUCT | ASN1H_CTX | 1)
#define MMS_CONF_ERR_PDU	(ASN1H_CONSTRUCT | ASN1H_CTX | 2)
#define MMS_UNCONFIRMED_PDU  (ASN1H_CTX| ASN1H_CONSTRUCT | 3)
#define MMS_REJECT_PDU  (ASN1H_CTX| ASN1H_CONSTRUCT | 4)
#define MMS_CANCEL_REQ_PDU  (ASN1H_CTX | 5)
#define MMS_CANCEL_RESP_PDU  (ASN1H_CTX | 6)
#define MMS_CANCEL_ERR_PDU (ASN1H_CTX | ASN1H_CONSTRUCT | 7)
#define MMS_INITIATE_REQ_PDU (ASN1H_CTX | ASN1H_CONSTRUCT | 8)
#define MMS_INITIATE_RESP_PDU (ASN1H_CTX | ASN1H_CONSTRUCT | 9)
#define MMS_INITIATE_ERR_PDU (ASN1H_CTX | ASN1H_CONSTRUCT | 10)
#define MMS_CONCLUDE_REQ_PDU (ASN1H_CTX | 11)
#define MMS_CONCLUDE_RESP_PDU (ASN1H_CTX | 12)
#define MMS_CONCLUDE_ERROR_PDU (ASN1H_CTX | 13)

static const value_string mms_pdu_type_vals[] = {
  {0,   	"Conf Request"},
  {1,  		"Conf Response"},
  {2,   	"Conf Error"},
  {3,  		"Unconfirmed"},
  {4,  		"Reject"},
  {5,   	"Cancel Request"},
  {6,   	"Cancel Response"},
  {7,   	"Cancel Error"},
  {8,   	"Initiate Request"},
  {9,   	"Initiate Response"},
  {10,   	"Initiate Error"},
  {11,   	"Conclude Request"},
  {12,   	"Conclude Response"},
  {13,   	"Conclude Error"},
  {255,		"undefined PDU type"}
};

static void 
display_col_info(packet_info *pinfo, guchar *display_string)
{
	guchar *mytext;

	mytext = g_malloc(100+strlen(display_string));
	if (check_col(pinfo->cinfo, COL_INFO)) 
	{
		if (displayed_header == TRUE)
		{							 
			col_append_str(pinfo->cinfo, COL_INFO, display_string);
		}
		else
		{
			sprintf(mytext,"%s (Decoder Courtesy of Herbert Falk of SISCO)",display_string);
			col_set_str(pinfo->cinfo, COL_INFO, mytext);
			displayed_header = TRUE;
		}
	}

}

static guchar unknown_service_string[]="Unknown";


static void find_service_string(tvbuff_t *tvb,  gint offset, packet_info *pinfo)
{
	guchar *string_found = NULL;
	guint tag;
	ASN1_TBL_ENTRY *tbl1;
	int ret;
	ASN1_SCK asn1;
	gint def;
	guint total_length;
	gint myint;
	guchar *mytext;
	gint offset1 = offset;

	/*open the ASN1 tree, offset is tvb->offset */
	asn1_open(&asn1, tvb, offset);

	/*need to get the MMS PDU Type */
	asn1h_tag_get (&asn1, &tag);

	ret = asn1_length_decode (&asn1, &def, &total_length);
	if (def == FALSE)
	    asn1.offset++;

	switch(tag)
	{
		case MMS_INITIATE_REQ_PDU:
		  display_col_info(pinfo, "Initiate Request");
		  break;

		case MMS_INITIATE_RESP_PDU:
		  display_col_info(pinfo, "Initiate Response");
		  break;

		case MMS_INITIATE_ERR_PDU:
		  display_col_info(pinfo, "Initiate Error");
		  break;

		case MMS_CONF_REQ_PDU:
			/*need to skip over invokeID */
			asn1h_tag_get (&asn1, &tag);
			ret = asn1_length_decode (&asn1, &def, &total_length);
			asn1_int32_value_decode (&asn1, total_length, &myint);

			asn1h_tag_get (&asn1, &tag);
			tbl1 = mms_conf_req;
			while((string_found==NULL) && (tbl1->tag_string!=NULL))
			{
				if(tag == tbl1->tag)		/*then we have found the tag */
					string_found = tbl1->tag_string;
			       ++tbl1;
			}

			if (string_found ==NULL)
			       string_found = unknown_service_string;

			display_col_info(pinfo, string_found);
			break;

		case MMS_CONF_RESP_PDU:
			/*need to skip over invokeID */
			asn1h_tag_get (&asn1, &tag);
			ret = asn1_length_decode (&asn1, &def, &total_length);
			asn1_int32_value_decode (&asn1, total_length, &myint);
			
			asn1h_tag_get (&asn1, &tag);
			tbl1 = mms_conf_resp;
			while((string_found==NULL) && (tbl1->tag_string!=NULL))
			{
				if(tag == tbl1->tag)		/*then we have found the tag */
					string_found = tbl1->tag_string;
				++tbl1;
			}
			
			if(string_found ==NULL)
				string_found = unknown_service_string;
			
			display_col_info(pinfo, string_found);
			break;

		case MMS_CONF_ERR_PDU:
			/*need to skip over invokeID */
			asn1h_tag_get (&asn1, &tag);
			ret = asn1_length_decode (&asn1, &def, &total_length);
			asn1_int32_value_decode (&asn1, total_length, &myint);
			
			asn1h_tag_get (&asn1, &tag);
			tbl1 = mms_conf_resp;
			while((string_found==NULL) && (tbl1->tag_string!=NULL))
			{
				if(tag == tbl1->tag)		/*then we have found the tag */
					string_found = tbl1->tag_string;
				++tbl1;
			}
			
			if(string_found ==NULL)
				string_found = unknown_service_string;
			
			mytext = g_malloc(40 + strlen(string_found));
			display_col_info(pinfo, string_found);
			break;

		case MMS_REJECT_PDU:
			display_col_info(pinfo, "Reject");
			break;

		case MMS_CANCEL_REQ_PDU:
			display_col_info(pinfo, "Cancel Request");
			break;

		case MMS_CANCEL_RESP_PDU:
			display_col_info(pinfo, "Cancel Response");
			break;

		case MMS_CANCEL_ERR_PDU:
			display_col_info(pinfo, "Cancel Error");
			break;

		case MMS_CONCLUDE_REQ_PDU:
			display_col_info(pinfo, "Conclude Request");
			break;

		case MMS_CONCLUDE_RESP_PDU:
			display_col_info(pinfo, "Conclude Response");
			break;

		case MMS_CONCLUDE_ERROR_PDU:
			display_col_info(pinfo, "Conclude Error");
			break;

		case MMS_UNCONFIRMED_PDU:
			asn1h_tag_get (&asn1, &tag);
			tbl1 = mms_unconfirmed_pdu;
			while((string_found==NULL) && (tbl1->tag_string!=NULL))
			{
				if(tag == tbl1->tag)		/*then we have found the tag */
					string_found = tbl1->tag_string;
				++tbl1;
			}
			
			if(string_found ==NULL)
				string_found = unknown_service_string;
			
			display_col_info(pinfo, string_found);
			break;

		 default:
			break;
	}

	asn1_close (&asn1,&offset1);	
}

static void
dissect_mms_service(ASN1_SCK *asn1, ASN1_TBL_ENTRY *tbl_array, guchar *display_input, guint32 *total_len, guint32 init_offset1, proto_item *subtree, tvbuff_t *tvb, guint type_of_table, packet_info *pinfo)
{
	guint tag;
	guint total_length;
	gint invoke_id,service_id;
	int ret;
	gint def;
	ASN1_TBL_ENTRY *tbl1 = tbl_array;
	ASN1_SCK asn11 = *asn1;
	int i=0;

	/* need to skip over invokeID */
	asn1h_tag_get (asn1, &tag);
	ret = asn1h_length_decode (asn1, &def, &total_length);
	if (ret == ASN1_ERR_NOERROR) 
	{
		int offset;
		asn1_int32_value_decode (asn1, total_length, &invoke_id);
	
		offset = asn1->offset;
		asn1h_tag_get (asn1, &tag);
		ret = asn1h_length_decode (asn1, &def, &total_length);
		if (ret == ASN1_ERR_NOERROR) 
		{
 /* 		      asn1_int32_value_decode (asn1, total_length, &service_id);  */

	
			  while(tbl1->tag_string)
			  {
				if (tag == tbl1->tag)		/* then we have found the tag */
				{
					char s[80];
					service_id = i;
					proto_tree_add_uint_format (subtree, hf_mms_service_id, tvb, offset, asn1->offset, tag & 255,
								"%s (%u)",tbl1->tag_string,tag & 255);
					display_col_info(pinfo,": ");
					display_col_info(pinfo,tbl1->tag_string);
					sprintf(s," (InvokeID: %u)",invoke_id);
					display_col_info(pinfo,s);
					break;
				}
				++tbl1;
				++i;
			  }
			}
	}
	asn1h_do_parse (&asn11, tbl_array, display_input, total_len, init_offset1, subtree, tvb, type_of_table, global_mms_asn1_prefs_nodes_as_subtrees);
}

static void
dissect_mms(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
	int offset = 0;
	proto_item *ti;
	guint tag;
	int ret;
	gboolean def;
	ASN1_SCK asn1;
	guint total_length;
	gint32 myint;
	char display_text [100];
	gboolean enoughBytes;
	proto_item *data_item;

    if (tvb->length == 0)
	  return;

    if (global_mms_prefs_use_generic_decoder==TRUE && (generic_handle !=NULL))
      {
		call_dissector(generic_handle, tvb, pinfo, tree);
		return;
       } 

 
	if (check_col(pinfo->cinfo, COL_PROTOCOL))
		col_set_str(pinfo->cinfo, COL_PROTOCOL, "MMS");

  	if (check_col(pinfo->cinfo, COL_INFO))
  		col_clear(pinfo->cinfo, COL_INFO);

	if (!tree)
	{
		find_service_string(tvb,  offset , pinfo);
		return;
	}
   

	if (tree) 
	{
		/*open the ASN1 tree, offset is tvb->offset */
		asn1_open(&asn1, tvb, offset);
		
		ret = asn1h_tag_get (&asn1, &tag);
		
		ret = asn1_length_decode (&asn1, &def, &total_length);
		
		
		enoughBytes = tvb_bytes_exist(tvb, asn1.offset, total_length);
		
		if (enoughBytes == FALSE)
			return;

#if 0
		if (total_length < 0)			/*several MMS PDUs have lengths of NULL */
			return;
#endif

		ti = proto_tree_add_item(tree, proto_iso9506, tvb, 0, -1, FALSE);
		data_item = proto_item_add_subtree(ti, ett_iso9506); 
		proto_tree_add_uint_format (data_item, hf_mms_pdu_type, tvb, 0, asn1.offset, tag & 255,
						"%s (%u)",val_to_str(tag & 255,mms_pdu_type_vals,"unknown"),tag & 255);     
		display_col_info(pinfo,val_to_str(tag & 255,mms_pdu_type_vals,"unknown"));
		
		switch(tag)
		{
		case MMS_INITIATE_REQ_PDU:
		  asn1h_do_parse (&asn1, mms_initiate_req, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SEQ,global_mms_asn1_prefs_nodes_as_subtrees);
		  break;

		case MMS_INITIATE_RESP_PDU:
		  asn1h_do_parse (&asn1, mms_initiate_resp, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SEQ,global_mms_asn1_prefs_nodes_as_subtrees);
		  break;

		case MMS_INITIATE_ERR_PDU:
		  asn1h_do_parse (&asn1, mms_service_error, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SEQ,global_mms_asn1_prefs_nodes_as_subtrees);
		  break;

		case MMS_CONF_REQ_PDU:
		  dissect_mms_service (&asn1, mms_conf_req, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SEQ, pinfo);
		  break;

		case MMS_CONF_RESP_PDU:
		  dissect_mms_service (&asn1, mms_conf_resp, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET, pinfo);
		  break;

		case MMS_CONF_ERR_PDU:
		  dissect_mms_service (&asn1, mms_conf_error, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET, pinfo);
		  break;

		case MMS_REJECT_PDU:
		  asn1h_do_parse (&asn1, mms_reject, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET,global_mms_asn1_prefs_nodes_as_subtrees);
		  break;
 
		case MMS_CANCEL_REQ_PDU:
 		  asn1_int32_value_decode (&asn1, (gint )total_length, &myint);
		  sprintf(display_text,"MMS Cancel Request Invoke ID = %d", myint);
   		  proto_tree_add_text (data_item, tvb, 0, asn1.offset, display_text);
		  break;

		case MMS_CANCEL_RESP_PDU:
 		  asn1_int32_value_decode (&asn1, (gint)total_length, &myint);
		  sprintf(display_text,"MMS Cancel Response Invoke ID = %d", myint);
   		  proto_tree_add_text (data_item, tvb, 0, asn1.offset, display_text);
		  break;

		case MMS_CANCEL_ERR_PDU:
		  asn1h_do_parse (&asn1, mms_cancel_err, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET,global_mms_asn1_prefs_nodes_as_subtrees);
		  break;

		case MMS_CONCLUDE_REQ_PDU:
		  break;


		case MMS_CONCLUDE_RESP_PDU:
		  break;

		case MMS_CONCLUDE_ERROR_PDU:
		  asn1h_do_parse (&asn1, mms_service_error, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET,global_mms_asn1_prefs_nodes_as_subtrees);
  		  break;

		case MMS_UNCONFIRMED_PDU:
		  dissect_mms_service (&asn1, mms_unconfirmed_pdu, "", &total_length, asn1.offset, data_item, tvb, ASN1H_TBL_SET, pinfo);
  		  break;

		default:
		  break;
		}
		asn1_close (&asn1,&offset);
	}
}

/* added for ABBPLC profile (directly from RFC-1006/COTP) */
static gboolean
dissect_mms_abbplc_heur(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{		 
  ASN1_SCK asn1;
  guint tag;
  int ret;
  int offset = 0;

  if (tvb->length > 0)
   {
	/*need to check if the first bytes is actually MMS */
  	asn1_open(&asn1, tvb, offset);
		
  	ret = asn1h_tag_get (&asn1, &tag);
	switch(tag)
	  {
	  case MMS_INITIATE_REQ_PDU:
	  case MMS_INITIATE_RESP_PDU:
  	  case MMS_INITIATE_ERR_PDU:
		case MMS_CONF_REQ_PDU:
		case MMS_CONF_RESP_PDU:
		case MMS_CONF_ERR_PDU:
		case MMS_REJECT_PDU:
		case MMS_CANCEL_REQ_PDU:
		case MMS_CANCEL_RESP_PDU:
		case MMS_CANCEL_ERR_PDU:
		case MMS_CONCLUDE_REQ_PDU:
		case MMS_CONCLUDE_RESP_PDU:
		case MMS_CONCLUDE_ERROR_PDU:
		case MMS_UNCONFIRMED_PDU:
           dissect_mms(tvb, pinfo, tree);
		   return TRUE;
		   break;

      default:
	    return FALSE;
	    break;
      }
   	}
  return FALSE;
}

void
proto_register_iso9506(void)
{
	static hf_register_info hf[] =
	{
		/* hfields for packet-iso9506.c */	
		{ &hf_mms_pdu_type,
			{ "PDU Type", "mms.pdu_type", FT_UINT16, BASE_HEX, NULL, 0x0,"", HFILL}
		},
		{ &hf_mms_invoke_id,
			{ "InvokeID", "mms.invoke_id", FT_UINT24, BASE_DEC, NULL, 0x0,"", HFILL}
		},
		{ &hf_mms_service_id,
			{ "ServiceID", "mms.service_id", FT_UINT16, BASE_DEC, NULL, 0x0,"", HFILL}
		},
		
		/* hfields for asn1_helper.c */
		{ &hf_mms_version,
			{ "MMS Version Number", "mms.version", FT_INT24, BASE_DEC, NULL, 0x0,"", HFILL}
		},		
		{ &hf_mms_filename,
			{ "Filename", "mms.filename", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},
		{ &hf_mms_pin,
			{ "ProgramInvocationName ", "mms.pin", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},	
		{ &hf_mms_domainname,
			{ "DomainName", "mms.domainname", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},
		{ &hf_mms_itemname,
			{ "ItemName", "mms.itemname", FT_STRING, BASE_NONE, NULL, 0x0,"", HFILL}
		},

	};

	static gint *ett[] =
	{
		&ett_iso9506,
		&ett_service,
		&ett_varlist,
		&ett_vardesc,	
		&ett_varspec,	
		&ett_addrspec,
		&ett_typespec,
		&ett_altspec,	
		&ett_obj,
		&ett_filetab,
		&ett_filetab2,	
		&ett_fileattrtab,	
		&ett_domainname,
		&ett_domainspec,
		&ett_item,
		&ett_datatbl,
		&ett_array,
		&ett_structure,
		&ett_bitstring,
		&ett_octettstring,
		&ett_string,
		&ett_gentime,
		&ett_utctime,
		&ett_bintime,
		&ett_unicode,
		&ett_bcd,
		&ett_oid,
		&ett_components,
		&ett_init_detail,	
		&ett_init_cbb,		
		&ett_init_services,
		&ett_service_error,		
	};
	module_t   *iso9506_module;


	proto_iso9506 = proto_register_protocol(PROTO_STRING_ISO9506, "ISO/IEC 9506", "mms");

/*	proto_iso9506_abbplc = proto_register_protocol(PROTO_STRING_ISO9506_ABBPLC, "MMS ABBPLC", "mmsabb");*/

	proto_register_field_array(proto_iso9506, hf, array_length(hf));
	proto_register_subtree_array(ett, array_length(ett));

	iso9506_module = prefs_register_protocol(proto_iso9506, proto_reg_handoff_iso9506);

	prefs_register_bool_preference(iso9506_module, "asn1_nodes_as_subtrees", 
                                 "Show ASN1 nodes as subtrees", 
                                 "Show ASN1 nodes as subtrees",
                                 &global_mms_asn1_prefs_nodes_as_subtrees);

	prefs_register_bool_preference(iso9506_module, "use_generic_mms_decoder", 
                                 "Use Generic MMS Decoder", 
                                 "Use Generic MMS Decoder",
                                 &global_mms_prefs_use_generic_decoder);

 	register_dissector("mms", dissect_mms, proto_iso9506); 
/* 	register_dissector("abb", dissect_mms_abbplc, proto_iso9506_abbplc); */

}

void
proto_reg_handoff_iso9506(void)
{
#if 0	 /*original code updated for version 0.10.12 */
	dissector_handle_t iso9506_handle;
	iso9506_handle  = create_dissector_handle(dissect_mms,proto_iso9506);
	/* Register in acse oid table  */
	dissector_add_string("acse.application_context", "1.0.9506.2.3", iso9506_handle);
	dissector_add_string("acse.application_context", "iso.0.9506.2.3", iso9506_handle);
	dissector_add_string("acse.application_context", "1.0.9506.2.3 (iso.0.9506.2.3)", iso9506_handle); 
#endif
	dissector_handle_t iso9506_handle;
	static int iso9506_prefs_initialized = FALSE;

    if (!iso9506_prefs_initialized){
	  iso9506_handle = create_dissector_handle(dissect_mms, proto_iso9506);
	  iso9506_prefs_initialized = TRUE;
	}
	generic_handle = find_dissector("mms2");

	register_ber_oid_dissector("1.0.9506.2.3", dissect_mms, proto_iso9506,"MMS");
	register_ber_oid_dissector("1.0.9506.2.1", dissect_mms, proto_iso9506,"mms-abstract-syntax-version1(1)");

	/* add our session dissector to cotp dissector list 
	 * and cotp_is dissector list*/
    heur_dissector_add("cotp", dissect_mms_abbplc_heur, proto_iso9506);

/*	heur_dissector_add("cotp", dissect_mms_abbplc, proto_iso9506_abbplc);
	heur_dissector_add("cotp_is", dissect_mms_abbplc, proto_iso9506_abbplc);  */


}
